/*
 * Copyright Mesa3D Contributors
 * SPDX-License-Identifier: MIT
 *
 * Autogenerated file, do not edit
 */

#pragma once

#include "compiler/nir/nir.h"
#include "compiler/nir/nir_builder.h"

#include "util/u_printf.h"

#ifdef __cplusplus
extern "C" {
#endif
void
gfx35_write_3DSTATE_VERTEX_BUFFERS(nir_builder *b, nir_def *dst_ptr, nir_def *buffer_count);

void
gfx35_write_VERTEX_BUFFER_STATE(nir_builder *b, nir_def *dst_ptr, nir_def *mocs, nir_def *buffer_idx, nir_def *address, nir_def *size, nir_def *stride);

void
gfx35_write_3DPRIMITIVE(nir_builder *b, nir_def *dst_ptr, nir_def *is_predicated, nir_def *is_indexed, nir_def *uses_tbimr, nir_def *vertex_count_per_instance, nir_def *start_vertex_location, nir_def *instance_count, nir_def *start_instance_location, nir_def *base_vertex_location);

void
gfx35_write_3DPRIMITIVE_EXTENDED(nir_builder *b, nir_def *dst_ptr, nir_def *is_predicated, nir_def *is_indexed, nir_def *uses_tbimr, nir_def *vertex_count_per_instance, nir_def *start_vertex_location, nir_def *instance_count, nir_def *start_instance_location, nir_def *base_vertex_location, nir_def *param_base_vertex, nir_def *param_base_instance, nir_def *param_draw_id);

void
gfx35_write_MI_BATCH_BUFFER_START(nir_builder *b, nir_def *dst_ptr, nir_def *addr);

void
gfx35_write_draw(nir_builder *b, nir_def *dst_ptr, nir_def *indirect_ptr, nir_def *draw_id_ptr, nir_def *draw_id, nir_def *instance_multiplier, nir_def *is_indexed, nir_def *is_predicated, nir_def *uses_tbimr, nir_def *uses_base, nir_def *uses_drawid, nir_def *mocs);

void
gfx35_libanv_write_draw(nir_builder *b, nir_def *dst_base, nir_def *wa_insts_ptr, nir_def *indirect_base, nir_def *draw_id_base, nir_def *indirect_stride, nir_def *_draw_count, nir_def *draw_base, nir_def *instance_multiplier, nir_def *max_draw_count, nir_def *flags, nir_def *mocs, nir_def *_3dprim_size_B, nir_def *ring_count, nir_def *gen_addr, nir_def *end_addr, nir_def *item_idx);

void
gfx35_libiris_write_draw(nir_builder *b, nir_def *dst_base, nir_def *indirect_base, nir_def *draw_id_base, nir_def *indirect_stride, nir_def *indirect_draw_count, nir_def *draw_base, nir_def *max_draw_count, nir_def *flags, nir_def *ring_count, nir_def *gen_addr, nir_def *end_addr, nir_def *item_idx);

void
gfx35_libanv_query_copy(nir_builder *b, nir_def *destination_base, nir_def *destination_stride, nir_def *query_data, nir_def *first_query, nir_def *num_queries, nir_def *query_data_offset, nir_def *query_stride, nir_def *num_query_items, nir_def *copy_flags, nir_def *copy_item_idx);

void
gfx35_libanv_memcpy(nir_builder *b, nir_def *dst_base, nir_def *src_base, nir_def *num_dwords, nir_def *dword_offset);

void
gfx35_copy_data(nir_builder *b, nir_def *dst_ptr, nir_def *src_ptr, nir_def *size);

#ifdef __cplusplus
} /* extern C */
#endif
