#include "nv_push_cla0c0.h"

#include <stdio.h>

#include "util/u_math.h"





const char *P_PARSE_NVC1C0_MTHD(uint16_t idx) ATTRIBUTE_COLD;
void P_DUMP_NVC1C0_MTHD_DATA(FILE *fp, uint16_t idx, uint32_t data,
                                   const char *prefix) ATTRIBUTE_COLD;

const char*
P_PARSE_NVA0C0_MTHD(uint16_t idx)
{
    switch (idx) {

  
        
    case NVA0C0_SET_OBJECT:

        
    case NVA0C0_NO_OPERATION:

        
    case NVA0C0_SET_NOTIFY_A:

        
    case NVA0C0_SET_NOTIFY_B:

        
    case NVA0C0_NOTIFY:

        
    case NVA0C0_WAIT_FOR_IDLE:

        
    case NVA0C0_SET_GLOBAL_RENDER_ENABLE_A:

        
    case NVA0C0_SET_GLOBAL_RENDER_ENABLE_B:

        
    case NVA0C0_SET_GLOBAL_RENDER_ENABLE_C:

        
    case NVA0C0_SEND_GO_IDLE:

        
    case NVA0C0_PM_TRIGGER:

        
    case NVA0C0_PM_TRIGGER_WFI:

        
    case NVA0C0_SET_INSTRUMENTATION_METHOD_HEADER:

        
    case NVA0C0_SET_INSTRUMENTATION_METHOD_DATA:

        
    case NVA0C0_LINE_LENGTH_IN:

        
    case NVA0C0_LINE_COUNT:

        
    case NVA0C0_OFFSET_OUT:

        
    case NVA0C0_PITCH_OUT:

        
    case NVA0C0_SET_DST_BLOCK_SIZE:

        
    case NVA0C0_SET_DST_WIDTH:

        
    case NVA0C0_SET_DST_HEIGHT:

        
    case NVA0C0_SET_DST_DEPTH:

        
    case NVA0C0_SET_DST_LAYER:

        
    case NVA0C0_LAUNCH_DMA:

        
    case NVA0C0_LOAD_INLINE_DATA:

        
    case NVA0C0_SET_I2M_SEMAPHORE_A:

        
    case NVA0C0_SET_I2M_SEMAPHORE_B:

        
    case NVA0C0_SET_I2M_SEMAPHORE_C:

        
    case NVA0C0_SET_I2M_SPARE_NOOP00:

        
    case NVA0C0_SET_I2M_SPARE_NOOP01:

        
    case NVA0C0_SET_I2M_SPARE_NOOP02:

        
    case NVA0C0_SET_I2M_SPARE_NOOP03:

        
    case NVA0C0_PERFMON_TRANSFER:

        
    case NVA0C0_SET_SHADER_SHARED_MEMORY_WINDOW:

        
    case NVA0C0_INVALIDATE_SHADER_CACHES:

        
    case NVA0C0_SET_CWD_CONTROL:

        
    case NVA0C0_INVALIDATE_TEXTURE_HEADER_CACHE_NO_WFI:

        
    case NVA0C0_SET_CWD_REF_COUNTER:

        
    case NVA0C0_INVALIDATE_CONSTANT_BUFFER_CACHE_B:

        
    case NVA0C0_INVALIDATE_CONSTANT_BUFFER_CACHE_C:

        
    case NVA0C0_SET_COMPUTE_CLASS_VERSION:

        
    case NVA0C0_CHECK_COMPUTE_CLASS_VERSION:

        
    case NVA0C0_SET_QMD_VERSION:

        
    case NVA0C0_CHECK_QMD_VERSION:

        
    case NVA0C0_SET_CWD_SLOT_COUNT:

        
    case NVA0C0_SEND_PCAS_A:

        
    case NVA0C0_SEND_PCAS_B:

        
    case NVA0C0_SEND_SIGNALING_PCAS_B:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_NON_THROTTLED_A:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_NON_THROTTLED_B:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_NON_THROTTLED_C:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_THROTTLED_A:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_THROTTLED_B:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_THROTTLED_C:

        
    case NVA0C0_SET_SPA_VERSION:

        
    case NVA0C0_SET_FALCON00:

        
    case NVA0C0_SET_FALCON01:

        
    case NVA0C0_SET_FALCON02:

        
    case NVA0C0_SET_FALCON03:

        
    case NVA0C0_SET_FALCON04:

        
    case NVA0C0_SET_FALCON05:

        
    case NVA0C0_SET_FALCON06:

        
    case NVA0C0_SET_FALCON07:

        
    case NVA0C0_SET_FALCON08:

        
    case NVA0C0_SET_FALCON09:

        
    case NVA0C0_SET_FALCON10:

        
    case NVA0C0_SET_FALCON11:

        
    case NVA0C0_SET_FALCON12:

        
    case NVA0C0_SET_FALCON13:

        
    case NVA0C0_SET_FALCON14:

        
    case NVA0C0_SET_FALCON15:

        
    case NVA0C0_SET_FALCON16:

        
    case NVA0C0_SET_FALCON17:

        
    case NVA0C0_SET_FALCON18:

        
    case NVA0C0_SET_FALCON19:

        
    case NVA0C0_SET_FALCON20:

        
    case NVA0C0_SET_FALCON21:

        
    case NVA0C0_SET_FALCON22:

        
    case NVA0C0_SET_FALCON23:

        
    case NVA0C0_SET_FALCON24:

        
    case NVA0C0_SET_FALCON25:

        
    case NVA0C0_SET_FALCON26:

        
    case NVA0C0_SET_FALCON27:

        
    case NVA0C0_SET_FALCON28:

        
    case NVA0C0_SET_FALCON29:

        
    case NVA0C0_SET_FALCON30:

        
    case NVA0C0_SET_FALCON31:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_WINDOW:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_B:

        
    case NVA0C0_SET_SHADER_CACHE_CONTROL:

        
    case NVA0C0_SET_SM_TIMEOUT_INTERVAL:

        
    case NVA0C0_SET_SPARE_NOOP12:

        
    case NVA0C0_SET_SPARE_NOOP13:

        
    case NVA0C0_SET_SPARE_NOOP14:

        
    case NVA0C0_SET_SPARE_NOOP15:

        
    case NVA0C0_SET_SPARE_NOOP00:

        
    case NVA0C0_SET_SPARE_NOOP01:

        
    case NVA0C0_SET_SPARE_NOOP02:

        
    case NVA0C0_SET_SPARE_NOOP03:

        
    case NVA0C0_SET_SPARE_NOOP04:

        
    case NVA0C0_SET_SPARE_NOOP05:

        
    case NVA0C0_SET_SPARE_NOOP06:

        
    case NVA0C0_SET_SPARE_NOOP07:

        
    case NVA0C0_SET_SPARE_NOOP08:

        
    case NVA0C0_SET_SPARE_NOOP09:

        
    case NVA0C0_SET_SPARE_NOOP10:

        
    case NVA0C0_SET_SPARE_NOOP11:

        
    case NVA0C0_INVALIDATE_TEXTURE_DATA_CACHE_NO_WFI:

        
    case NVA0C0_INVALIDATE_SAMPLER_CACHE:

        
    case NVA0C0_INVALIDATE_TEXTURE_HEADER_CACHE:

        
    case NVA0C0_INVALIDATE_TEXTURE_DATA_CACHE:

        
    case NVA0C0_INVALIDATE_SAMPLER_CACHE_NO_WFI:

        
    case NVA0C0_SET_SHADER_EXCEPTIONS:

        
    case NVA0C0_SET_RENDER_ENABLE_A:

        
    case NVA0C0_SET_RENDER_ENABLE_B:

        
    case NVA0C0_SET_RENDER_ENABLE_C:

        
    case NVA0C0_SET_TEX_SAMPLER_POOL_B:

        
    case NVA0C0_SET_TEX_SAMPLER_POOL_C:

        
    case NVA0C0_SET_TEX_HEADER_POOL_B:

        
    case NVA0C0_SET_TEX_HEADER_POOL_C:

        
    case NVA0C0_SET_PROGRAM_REGION_B:

        
    case NVA0C0_INVALIDATE_SHADER_CACHES_NO_WFI:

        
    case NVA0C0_SET_RENDER_ENABLE_OVERRIDE:

        
    case NVA0C0_PIPE_NOP:

        
    case NVA0C0_SET_SPARE00:

        
    case NVA0C0_SET_SPARE01:

        
    case NVA0C0_SET_SPARE02:

        
    case NVA0C0_SET_SPARE03:

        
    case NVA0C0_SET_REPORT_SEMAPHORE_A:

        
    case NVA0C0_SET_REPORT_SEMAPHORE_B:

        
    case NVA0C0_SET_REPORT_SEMAPHORE_C:

        
    case NVA0C0_SET_REPORT_SEMAPHORE_D:

        
    case NVA0C0_SET_BINDLESS_TEXTURE:

        
    case NVA0C0_SET_TRAP_HANDLER:

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(0):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(1):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(2):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(3):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(4):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(5):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(6):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(7):

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(0):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(1):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(2):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(3):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(4):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(5):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(6):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(7):

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(0):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(1):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(2):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(3):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(4):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(5):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(6):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(7):

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(0):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(1):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(2):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(3):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(4):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(5):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(6):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(7):

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_TRAP_CONTROL:

        
    case NVA0C0_SET_MME_SHADOW_SCRATCH(0):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(1):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(2):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(3):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(4):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(5):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(6):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(7):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(8):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(9):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(10):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(11):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(12):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(13):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(14):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(15):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(16):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(17):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(18):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(19):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(20):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(21):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(22):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(23):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(24):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(25):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(26):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(27):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(28):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(29):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(30):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(31):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(32):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(33):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(34):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(35):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(36):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(37):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(38):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(39):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(40):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(41):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(42):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(43):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(44):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(45):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(46):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(47):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(48):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(49):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(50):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(51):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(52):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(53):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(54):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(55):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(56):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(57):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(58):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(59):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(60):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(61):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(62):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(63):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(64):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(65):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(66):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(67):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(68):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(69):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(70):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(71):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(72):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(73):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(74):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(75):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(76):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(77):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(78):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(79):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(80):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(81):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(82):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(83):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(84):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(85):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(86):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(87):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(88):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(89):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(90):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(91):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(92):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(93):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(94):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(95):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(96):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(97):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(98):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(99):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(100):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(101):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(102):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(103):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(104):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(105):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(106):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(107):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(108):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(109):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(110):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(111):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(112):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(113):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(114):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(115):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(116):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(117):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(118):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(119):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(120):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(121):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(122):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(123):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(124):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(125):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(126):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(127):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(128):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(129):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(130):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(131):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(132):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(133):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(134):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(135):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(136):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(137):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(138):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(139):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(140):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(141):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(142):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(143):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(144):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(145):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(146):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(147):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(148):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(149):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(150):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(151):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(152):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(153):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(154):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(155):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(156):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(157):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(158):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(159):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(160):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(161):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(162):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(163):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(164):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(165):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(166):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(167):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(168):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(169):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(170):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(171):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(172):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(173):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(174):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(175):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(176):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(177):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(178):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(179):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(180):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(181):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(182):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(183):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(184):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(185):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(186):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(187):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(188):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(189):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(190):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(191):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(192):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(193):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(194):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(195):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(196):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(197):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(198):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(199):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(200):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(201):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(202):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(203):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(204):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(205):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(206):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(207):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(208):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(209):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(210):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(211):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(212):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(213):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(214):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(215):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(216):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(217):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(218):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(219):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(220):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(221):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(222):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(223):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(224):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(225):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(226):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(227):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(228):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(229):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(230):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(231):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(232):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(233):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(234):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(235):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(236):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(237):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(238):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(239):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(240):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(241):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(242):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(243):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(244):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(245):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(246):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(247):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(248):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(249):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(250):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(251):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(252):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(253):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(254):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(255):


    return P_PARSE_NVC1C0_MTHD(idx);

                                                                    case NVA0C0_OFFSET_OUT_UPPER:
        return "NVA0C0_OFFSET_OUT_UPPER";
                                case NVA0C0_SET_DST_ORIGIN_BYTES_X:
        return "NVA0C0_SET_DST_ORIGIN_BYTES_X";
    case NVA0C0_SET_DST_ORIGIN_SAMPLES_Y:
        return "NVA0C0_SET_DST_ORIGIN_SAMPLES_Y";
                                                                case NVA0C0_INVALIDATE_CONSTANT_BUFFER_CACHE_A:
        return "NVA0C0_INVALIDATE_CONSTANT_BUFFER_CACHE_A";
                                                                                                                                                                                                            case NVA0C0_SET_SHADER_LOCAL_MEMORY_A:
        return "NVA0C0_SET_SHADER_LOCAL_MEMORY_A";
                                                                                                                    case NVA0C0_SET_TEX_SAMPLER_POOL_A:
        return "NVA0C0_SET_TEX_SAMPLER_POOL_A";
            case NVA0C0_SET_TEX_HEADER_POOL_A:
        return "NVA0C0_SET_TEX_HEADER_POOL_A";
            case NVA0C0_SET_PROGRAM_REGION_A:
        return "NVA0C0_SET_PROGRAM_REGION_A";
        case NVA0C0_SET_SHADER_CONTROL:
        return "NVA0C0_SET_SHADER_CONTROL";
                                                                            
    default:
        return "unknown method";
    }
}

void
P_DUMP_NVA0C0_MTHD_DATA(FILE *fp, uint16_t idx, uint32_t data,
                         const char *prefix)
{
    UNUSED uint32_t parsed;
    switch (idx) {

  
        
    case NVA0C0_SET_OBJECT:

        
    case NVA0C0_NO_OPERATION:

        
    case NVA0C0_SET_NOTIFY_A:

        
    case NVA0C0_SET_NOTIFY_B:

        
    case NVA0C0_NOTIFY:

        
    case NVA0C0_WAIT_FOR_IDLE:

        
    case NVA0C0_SET_GLOBAL_RENDER_ENABLE_A:

        
    case NVA0C0_SET_GLOBAL_RENDER_ENABLE_B:

        
    case NVA0C0_SET_GLOBAL_RENDER_ENABLE_C:

        
    case NVA0C0_SEND_GO_IDLE:

        
    case NVA0C0_PM_TRIGGER:

        
    case NVA0C0_PM_TRIGGER_WFI:

        
    case NVA0C0_SET_INSTRUMENTATION_METHOD_HEADER:

        
    case NVA0C0_SET_INSTRUMENTATION_METHOD_DATA:

        
    case NVA0C0_LINE_LENGTH_IN:

        
    case NVA0C0_LINE_COUNT:

        
    case NVA0C0_OFFSET_OUT:

        
    case NVA0C0_PITCH_OUT:

        
    case NVA0C0_SET_DST_BLOCK_SIZE:

        
    case NVA0C0_SET_DST_WIDTH:

        
    case NVA0C0_SET_DST_HEIGHT:

        
    case NVA0C0_SET_DST_DEPTH:

        
    case NVA0C0_SET_DST_LAYER:

        
    case NVA0C0_LAUNCH_DMA:

        
    case NVA0C0_LOAD_INLINE_DATA:

        
    case NVA0C0_SET_I2M_SEMAPHORE_A:

        
    case NVA0C0_SET_I2M_SEMAPHORE_B:

        
    case NVA0C0_SET_I2M_SEMAPHORE_C:

        
    case NVA0C0_SET_I2M_SPARE_NOOP00:

        
    case NVA0C0_SET_I2M_SPARE_NOOP01:

        
    case NVA0C0_SET_I2M_SPARE_NOOP02:

        
    case NVA0C0_SET_I2M_SPARE_NOOP03:

        
    case NVA0C0_PERFMON_TRANSFER:

        
    case NVA0C0_SET_SHADER_SHARED_MEMORY_WINDOW:

        
    case NVA0C0_INVALIDATE_SHADER_CACHES:

        
    case NVA0C0_SET_CWD_CONTROL:

        
    case NVA0C0_INVALIDATE_TEXTURE_HEADER_CACHE_NO_WFI:

        
    case NVA0C0_SET_CWD_REF_COUNTER:

        
    case NVA0C0_INVALIDATE_CONSTANT_BUFFER_CACHE_B:

        
    case NVA0C0_INVALIDATE_CONSTANT_BUFFER_CACHE_C:

        
    case NVA0C0_SET_COMPUTE_CLASS_VERSION:

        
    case NVA0C0_CHECK_COMPUTE_CLASS_VERSION:

        
    case NVA0C0_SET_QMD_VERSION:

        
    case NVA0C0_CHECK_QMD_VERSION:

        
    case NVA0C0_SET_CWD_SLOT_COUNT:

        
    case NVA0C0_SEND_PCAS_A:

        
    case NVA0C0_SEND_PCAS_B:

        
    case NVA0C0_SEND_SIGNALING_PCAS_B:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_NON_THROTTLED_A:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_NON_THROTTLED_B:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_NON_THROTTLED_C:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_THROTTLED_A:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_THROTTLED_B:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_THROTTLED_C:

        
    case NVA0C0_SET_SPA_VERSION:

        
    case NVA0C0_SET_FALCON00:

        
    case NVA0C0_SET_FALCON01:

        
    case NVA0C0_SET_FALCON02:

        
    case NVA0C0_SET_FALCON03:

        
    case NVA0C0_SET_FALCON04:

        
    case NVA0C0_SET_FALCON05:

        
    case NVA0C0_SET_FALCON06:

        
    case NVA0C0_SET_FALCON07:

        
    case NVA0C0_SET_FALCON08:

        
    case NVA0C0_SET_FALCON09:

        
    case NVA0C0_SET_FALCON10:

        
    case NVA0C0_SET_FALCON11:

        
    case NVA0C0_SET_FALCON12:

        
    case NVA0C0_SET_FALCON13:

        
    case NVA0C0_SET_FALCON14:

        
    case NVA0C0_SET_FALCON15:

        
    case NVA0C0_SET_FALCON16:

        
    case NVA0C0_SET_FALCON17:

        
    case NVA0C0_SET_FALCON18:

        
    case NVA0C0_SET_FALCON19:

        
    case NVA0C0_SET_FALCON20:

        
    case NVA0C0_SET_FALCON21:

        
    case NVA0C0_SET_FALCON22:

        
    case NVA0C0_SET_FALCON23:

        
    case NVA0C0_SET_FALCON24:

        
    case NVA0C0_SET_FALCON25:

        
    case NVA0C0_SET_FALCON26:

        
    case NVA0C0_SET_FALCON27:

        
    case NVA0C0_SET_FALCON28:

        
    case NVA0C0_SET_FALCON29:

        
    case NVA0C0_SET_FALCON30:

        
    case NVA0C0_SET_FALCON31:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_WINDOW:

        
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_B:

        
    case NVA0C0_SET_SHADER_CACHE_CONTROL:

        
    case NVA0C0_SET_SM_TIMEOUT_INTERVAL:

        
    case NVA0C0_SET_SPARE_NOOP12:

        
    case NVA0C0_SET_SPARE_NOOP13:

        
    case NVA0C0_SET_SPARE_NOOP14:

        
    case NVA0C0_SET_SPARE_NOOP15:

        
    case NVA0C0_SET_SPARE_NOOP00:

        
    case NVA0C0_SET_SPARE_NOOP01:

        
    case NVA0C0_SET_SPARE_NOOP02:

        
    case NVA0C0_SET_SPARE_NOOP03:

        
    case NVA0C0_SET_SPARE_NOOP04:

        
    case NVA0C0_SET_SPARE_NOOP05:

        
    case NVA0C0_SET_SPARE_NOOP06:

        
    case NVA0C0_SET_SPARE_NOOP07:

        
    case NVA0C0_SET_SPARE_NOOP08:

        
    case NVA0C0_SET_SPARE_NOOP09:

        
    case NVA0C0_SET_SPARE_NOOP10:

        
    case NVA0C0_SET_SPARE_NOOP11:

        
    case NVA0C0_INVALIDATE_TEXTURE_DATA_CACHE_NO_WFI:

        
    case NVA0C0_INVALIDATE_SAMPLER_CACHE:

        
    case NVA0C0_INVALIDATE_TEXTURE_HEADER_CACHE:

        
    case NVA0C0_INVALIDATE_TEXTURE_DATA_CACHE:

        
    case NVA0C0_INVALIDATE_SAMPLER_CACHE_NO_WFI:

        
    case NVA0C0_SET_SHADER_EXCEPTIONS:

        
    case NVA0C0_SET_RENDER_ENABLE_A:

        
    case NVA0C0_SET_RENDER_ENABLE_B:

        
    case NVA0C0_SET_RENDER_ENABLE_C:

        
    case NVA0C0_SET_TEX_SAMPLER_POOL_B:

        
    case NVA0C0_SET_TEX_SAMPLER_POOL_C:

        
    case NVA0C0_SET_TEX_HEADER_POOL_B:

        
    case NVA0C0_SET_TEX_HEADER_POOL_C:

        
    case NVA0C0_SET_PROGRAM_REGION_B:

        
    case NVA0C0_INVALIDATE_SHADER_CACHES_NO_WFI:

        
    case NVA0C0_SET_RENDER_ENABLE_OVERRIDE:

        
    case NVA0C0_PIPE_NOP:

        
    case NVA0C0_SET_SPARE00:

        
    case NVA0C0_SET_SPARE01:

        
    case NVA0C0_SET_SPARE02:

        
    case NVA0C0_SET_SPARE03:

        
    case NVA0C0_SET_REPORT_SEMAPHORE_A:

        
    case NVA0C0_SET_REPORT_SEMAPHORE_B:

        
    case NVA0C0_SET_REPORT_SEMAPHORE_C:

        
    case NVA0C0_SET_REPORT_SEMAPHORE_D:

        
    case NVA0C0_SET_BINDLESS_TEXTURE:

        
    case NVA0C0_SET_TRAP_HANDLER:

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(0):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(1):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(2):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(3):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(4):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(5):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(6):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_VALUE(7):

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(0):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(1):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(2):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(3):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(4):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(5):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(6):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_EVENT(7):

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(0):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(1):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(2):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(3):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(4):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(5):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(6):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_A(7):

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(0):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(1):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(2):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(3):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(4):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(5):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(6):
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_CONTROL_B(7):

        
    case NVA0C0_SET_SHADER_PERFORMANCE_COUNTER_TRAP_CONTROL:

        
    case NVA0C0_SET_MME_SHADOW_SCRATCH(0):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(1):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(2):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(3):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(4):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(5):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(6):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(7):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(8):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(9):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(10):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(11):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(12):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(13):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(14):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(15):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(16):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(17):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(18):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(19):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(20):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(21):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(22):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(23):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(24):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(25):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(26):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(27):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(28):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(29):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(30):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(31):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(32):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(33):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(34):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(35):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(36):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(37):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(38):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(39):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(40):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(41):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(42):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(43):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(44):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(45):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(46):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(47):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(48):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(49):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(50):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(51):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(52):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(53):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(54):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(55):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(56):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(57):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(58):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(59):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(60):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(61):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(62):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(63):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(64):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(65):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(66):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(67):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(68):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(69):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(70):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(71):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(72):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(73):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(74):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(75):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(76):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(77):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(78):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(79):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(80):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(81):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(82):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(83):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(84):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(85):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(86):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(87):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(88):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(89):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(90):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(91):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(92):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(93):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(94):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(95):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(96):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(97):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(98):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(99):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(100):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(101):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(102):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(103):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(104):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(105):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(106):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(107):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(108):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(109):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(110):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(111):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(112):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(113):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(114):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(115):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(116):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(117):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(118):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(119):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(120):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(121):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(122):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(123):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(124):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(125):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(126):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(127):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(128):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(129):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(130):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(131):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(132):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(133):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(134):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(135):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(136):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(137):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(138):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(139):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(140):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(141):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(142):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(143):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(144):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(145):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(146):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(147):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(148):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(149):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(150):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(151):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(152):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(153):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(154):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(155):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(156):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(157):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(158):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(159):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(160):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(161):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(162):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(163):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(164):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(165):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(166):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(167):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(168):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(169):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(170):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(171):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(172):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(173):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(174):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(175):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(176):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(177):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(178):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(179):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(180):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(181):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(182):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(183):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(184):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(185):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(186):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(187):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(188):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(189):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(190):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(191):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(192):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(193):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(194):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(195):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(196):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(197):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(198):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(199):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(200):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(201):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(202):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(203):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(204):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(205):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(206):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(207):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(208):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(209):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(210):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(211):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(212):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(213):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(214):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(215):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(216):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(217):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(218):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(219):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(220):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(221):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(222):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(223):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(224):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(225):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(226):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(227):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(228):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(229):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(230):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(231):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(232):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(233):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(234):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(235):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(236):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(237):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(238):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(239):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(240):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(241):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(242):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(243):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(244):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(245):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(246):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(247):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(248):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(249):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(250):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(251):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(252):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(253):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(254):
    case NVA0C0_SET_MME_SHADOW_SCRATCH(255):


    P_DUMP_NVC1C0_MTHD_DATA(fp, idx, data, prefix);
    break;

                                                                  
    case NVA0C0_OFFSET_OUT_UPPER:

    
        parsed = (data >> 0) & ((1u << 8) - 1);
        fprintf(fp, "%s.VALUE = ", prefix);
        fprintf(fp, "(0x%x)\n", parsed);
        break;
                              
    case NVA0C0_SET_DST_ORIGIN_BYTES_X:

    
        parsed = (data >> 0) & ((1u << 20) - 1);
        fprintf(fp, "%s.V = ", prefix);
        fprintf(fp, "(0x%x)\n", parsed);
        break;
  
    case NVA0C0_SET_DST_ORIGIN_SAMPLES_Y:

    
        parsed = (data >> 0) & ((1u << 16) - 1);
        fprintf(fp, "%s.V = ", prefix);
        fprintf(fp, "(0x%x)\n", parsed);
        break;
                                                              
    case NVA0C0_INVALIDATE_CONSTANT_BUFFER_CACHE_A:

    
        parsed = (data >> 0) & ((1u << 8) - 1);
        fprintf(fp, "%s.ADDRESS_UPPER = ", prefix);
        fprintf(fp, "(0x%x)\n", parsed);
        break;
                                                                                                                                                                                                          
    case NVA0C0_SET_SHADER_LOCAL_MEMORY_A:

    
        parsed = (data >> 0) & ((1u << 8) - 1);
        fprintf(fp, "%s.ADDRESS_UPPER = ", prefix);
        fprintf(fp, "(0x%x)\n", parsed);
        break;
                                                                                                                  
    case NVA0C0_SET_TEX_SAMPLER_POOL_A:

    
        parsed = (data >> 0) & ((1u << 8) - 1);
        fprintf(fp, "%s.OFFSET_UPPER = ", prefix);
        fprintf(fp, "(0x%x)\n", parsed);
        break;
          
    case NVA0C0_SET_TEX_HEADER_POOL_A:

    
        parsed = (data >> 0) & ((1u << 8) - 1);
        fprintf(fp, "%s.OFFSET_UPPER = ", prefix);
        fprintf(fp, "(0x%x)\n", parsed);
        break;
          
    case NVA0C0_SET_PROGRAM_REGION_A:

    
        parsed = (data >> 0) & ((1u << 8) - 1);
        fprintf(fp, "%s.ADDRESS_UPPER = ", prefix);
        fprintf(fp, "(0x%x)\n", parsed);
        break;
      
    case NVA0C0_SET_SHADER_CONTROL:

    
        parsed = (data >> 0) & ((1u << 1) - 1);
        fprintf(fp, "%s.DEFAULT_PARTIAL = ", prefix);
        switch (parsed) {
        case NVA0C0_SET_SHADER_CONTROL_DEFAULT_PARTIAL_ZERO:
            fprintf(fp, "ZERO\n");
            break;
        case NVA0C0_SET_SHADER_CONTROL_DEFAULT_PARTIAL_INFINITY:
            fprintf(fp, "INFINITY\n");
            break;
        default:
            fprintf(fp, "0x%x\n", parsed);
            break;
        }
        break;
                                                                                default:
        fprintf(fp, "%s.VALUE = 0x%x\n", prefix, data);
        break;
    }
}
