/*
 * Copyright © 2014 Broadcom
 * Copyright (C) 2012 Rob Clark <robclark@freedesktop.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef U_DRM_H
#define U_DRM_H

#include <stdint.h>
#include <stdbool.h>

/* Does the u64 array contain the listed u64? */

static inline bool
util_array_contains_u64(uint64_t needle, const uint64_t *haystack, unsigned count)
{
   unsigned i;

   for (i = 0; i < count; i++) {
      if (haystack[i] == needle)
         return true;
   }

   return false;
}

/* Given a list of DRM modifiers and a desired modifier, returns whether the
 * modifier is found */

static inline bool
drm_find_modifier(uint64_t modifier, const uint64_t *modifiers, unsigned count)
{
   return util_array_contains_u64(modifier, modifiers, count);
}

#endif
