
#ifndef __NOUVEAU_DRM_PUBLIC_H__
#define __NOUVEAU_DRM_PUBLIC_H__

struct pipe_screen;
struct pipe_screen_config;
struct renderonly;

struct pipe_screen *nouveau_drm_screen_create(int drmFD);

#endif
