/*
 * Copyright Mesa3D Contributors
 * SPDX-License-Identifier: MIT
 *
 * Autogenerated file, do not edit
 */

#include "compiler/nir/nir.h"
#include "compiler/nir/nir_builder.h"

#include "util/u_printf.h"

#ifdef __cplusplus
extern "C" {
#endif
#ifdef __cplusplus
} /* extern C */
#endif
namespace {
   struct vtn_bindgen_dummy {
      vtn_bindgen_dummy() {
         /* Format strings:
          *
          */
const uint32_t printf_0_blob[1] = {
    0x00000000,
};
         u_printf_singleton_init_or_ref();
         u_printf_singleton_add_serialized((const void*)printf_0_blob, sizeof(printf_0_blob));
      }

      ~vtn_bindgen_dummy() {
         u_printf_singleton_decref();
      }
   };

   static vtn_bindgen_dummy vtn_bindgen_dummy_instance;
}
