/*
 * Copyright © 2014 Rob Clark <robclark@freedesktop.org>
 * SPDX-License-Identifier: MIT
 *
 * Authors:
 *    Rob Clark <robclark@freedesktop.org>
 */

#ifndef FD4_GMEM_H_
#define FD4_GMEM_H_

#include "pipe/p_context.h"

void fd4_gmem_init(struct pipe_context *pctx);

#endif /* FD4_GMEM_H_ */
