/* This file is automatically generated. DO NOT EDIT.
 *
 * Copyright 2026 Advanced Micro Devices, Inc.
 * SPDX-License-Identifier: MIT
 */

#ifndef AMD_CP_IB_PARSER_GFX11
#define AMD_CP_IB_PARSER_GFX11

#include "ac_debug.h"

/* Print the packet and use assertions to validate its content. */
void
amd_cp_print_packet_gfx11(struct ac_ib_parser *ib, unsigned opcode, unsigned pkt_count_field);

#endif
