/*
 * Copyright 2023 Intel Corporation
 * SPDX-License-Identifier: MIT
 */

/* This is an automatically generated file. */

#include "glsl_types.h"
#include "util/glheader.h"

const char glsl_type_builtin_names[] =
   "INVALID\0"
   "error\0"
   "void\0"
   "bool\0"
   "bvec2\0"
   "bvec3\0"
   "bvec4\0"
   "bvec5\0"
   "bvec8\0"
   "bvec16\0"
   "int\0"
   "ivec2\0"
   "ivec3\0"
   "ivec4\0"
   "ivec5\0"
   "ivec8\0"
   "ivec16\0"
   "uint\0"
   "uvec2\0"
   "uvec3\0"
   "uvec4\0"
   "uvec5\0"
   "uvec8\0"
   "uvec16\0"
   "float\0"
   "vec2\0"
   "vec3\0"
   "vec4\0"
   "vec5\0"
   "vec8\0"
   "vec16\0"
   "float16_t\0"
   "f16vec2\0"
   "f16vec3\0"
   "f16vec4\0"
   "f16vec5\0"
   "f16vec8\0"
   "f16vec16\0"
   "double\0"
   "dvec2\0"
   "dvec3\0"
   "dvec4\0"
   "dvec5\0"
   "dvec8\0"
   "dvec16\0"
   "int64_t\0"
   "i64vec2\0"
   "i64vec3\0"
   "i64vec4\0"
   "i64vec5\0"
   "i64vec8\0"
   "i64vec16\0"
   "uint64_t\0"
   "u64vec2\0"
   "u64vec3\0"
   "u64vec4\0"
   "u64vec5\0"
   "u64vec8\0"
   "u64vec16\0"
   "int16_t\0"
   "i16vec2\0"
   "i16vec3\0"
   "i16vec4\0"
   "i16vec5\0"
   "i16vec8\0"
   "i16vec16\0"
   "uint16_t\0"
   "u16vec2\0"
   "u16vec3\0"
   "u16vec4\0"
   "u16vec5\0"
   "u16vec8\0"
   "u16vec16\0"
   "int8_t\0"
   "i8vec2\0"
   "i8vec3\0"
   "i8vec4\0"
   "i8vec5\0"
   "i8vec8\0"
   "i8vec16\0"
   "uint8_t\0"
   "u8vec2\0"
   "u8vec3\0"
   "u8vec4\0"
   "u8vec5\0"
   "u8vec8\0"
   "u8vec16\0"
   "bfloat16_t\0"
   "bf16vec2\0"
   "bf16vec3\0"
   "bf16vec4\0"
   "bf16vec5\0"
   "bf16vec8\0"
   "bf16vec16\0"
   "e4m3fn_t\0"
   "e4m3fnvec2\0"
   "e4m3fnvec3\0"
   "e4m3fnvec4\0"
   "e4m3fnvec5\0"
   "e4m3fnvec8\0"
   "e4m3fnvec16\0"
   "e5m2_t\0"
   "e5m2vec2\0"
   "e5m2vec3\0"
   "e5m2vec4\0"
   "e5m2vec5\0"
   "e5m2vec8\0"
   "e5m2vec16\0"
   "mat2\0"
   "mat3\0"
   "mat4\0"
   "mat2x3\0"
   "mat2x4\0"
   "mat3x2\0"
   "mat3x4\0"
   "mat4x2\0"
   "mat4x3\0"
   "f16mat2\0"
   "f16mat3\0"
   "f16mat4\0"
   "f16mat2x3\0"
   "f16mat2x4\0"
   "f16mat3x2\0"
   "f16mat3x4\0"
   "f16mat4x2\0"
   "f16mat4x3\0"
   "dmat2\0"
   "dmat3\0"
   "dmat4\0"
   "dmat2x3\0"
   "dmat2x4\0"
   "dmat3x2\0"
   "dmat3x4\0"
   "dmat4x2\0"
   "dmat4x3\0"
   "atomic_uint\0"
   "sampler\0"
   "sampler1D\0"
   "sampler2D\0"
   "sampler3D\0"
   "samplerCube\0"
   "sampler1DArray\0"
   "sampler2DArray\0"
   "samplerCubeArray\0"
   "sampler2DRect\0"
   "samplerBuffer\0"
   "sampler2DMS\0"
   "sampler2DMSArray\0"
   "isampler1D\0"
   "isampler2D\0"
   "isampler3D\0"
   "isamplerCube\0"
   "isampler1DArray\0"
   "isampler2DArray\0"
   "isamplerCubeArray\0"
   "isampler2DRect\0"
   "isamplerBuffer\0"
   "isampler2DMS\0"
   "isampler2DMSArray\0"
   "usampler1D\0"
   "usampler2D\0"
   "usampler3D\0"
   "usamplerCube\0"
   "usampler1DArray\0"
   "usampler2DArray\0"
   "usamplerCubeArray\0"
   "usampler2DRect\0"
   "usamplerBuffer\0"
   "usampler2DMS\0"
   "usampler2DMSArray\0"
   "samplerShadow\0"
   "sampler1DShadow\0"
   "sampler2DShadow\0"
   "samplerCubeShadow\0"
   "sampler1DArrayShadow\0"
   "sampler2DArrayShadow\0"
   "samplerCubeArrayShadow\0"
   "sampler2DRectShadow\0"
   "samplerExternalOES\0"
   "texture1D\0"
   "texture2D\0"
   "texture3D\0"
   "textureCube\0"
   "texture1DArray\0"
   "texture2DArray\0"
   "textureCubeArray\0"
   "texture2DRect\0"
   "textureBuffer\0"
   "texture2DMS\0"
   "texture2DMSArray\0"
   "itexture1D\0"
   "itexture2D\0"
   "itexture3D\0"
   "itextureCube\0"
   "itexture1DArray\0"
   "itexture2DArray\0"
   "itextureCubeArray\0"
   "itexture2DRect\0"
   "itextureBuffer\0"
   "itexture2DMS\0"
   "itexture2DMSArray\0"
   "utexture1D\0"
   "utexture2D\0"
   "utexture3D\0"
   "utextureCube\0"
   "utexture1DArray\0"
   "utexture2DArray\0"
   "utextureCubeArray\0"
   "utexture2DRect\0"
   "utextureBuffer\0"
   "utexture2DMS\0"
   "utexture2DMSArray\0"
   "textureExternalOES\0"
   "vtexture1D\0"
   "vtexture2D\0"
   "vtexture3D\0"
   "vtexture2DMS\0"
   "vtexture2DMSArray\0"
   "vtexture1DArray\0"
   "vtexture2DArray\0"
   "vtextureBuffer\0"
   "image1D\0"
   "image2D\0"
   "image3D\0"
   "image2DRect\0"
   "imageCube\0"
   "imageBuffer\0"
   "image1DArray\0"
   "image2DArray\0"
   "imageCubeArray\0"
   "image2DMS\0"
   "image2DMSArray\0"
   "iimage1D\0"
   "iimage2D\0"
   "iimage3D\0"
   "iimage2DRect\0"
   "iimageCube\0"
   "iimageBuffer\0"
   "iimage1DArray\0"
   "iimage2DArray\0"
   "iimageCubeArray\0"
   "iimage2DMS\0"
   "iimage2DMSArray\0"
   "uimage1D\0"
   "uimage2D\0"
   "uimage3D\0"
   "uimage2DRect\0"
   "uimageCube\0"
   "uimageBuffer\0"
   "uimage1DArray\0"
   "uimage2DArray\0"
   "uimageCubeArray\0"
   "uimage2DMS\0"
   "uimage2DMSArray\0"
   "i64image1D\0"
   "i64image2D\0"
   "i64image3D\0"
   "i64image2DRect\0"
   "i64imageCube\0"
   "i64imageBuffer\0"
   "i64image1DArray\0"
   "i64image2DArray\0"
   "i64imageCubeArray\0"
   "i64image2DMS\0"
   "i64image2DMSArray\0"
   "u64image1D\0"
   "u64image2D\0"
   "u64image3D\0"
   "u64image2DRect\0"
   "u64imageCube\0"
   "u64imageBuffer\0"
   "u64image1DArray\0"
   "u64image2DArray\0"
   "u64imageCubeArray\0"
   "u64image2DMS\0"
   "u64image2DMSArray\0"
   "vbuffer\0"
   "vimage1D\0"
   "vimage2D\0"
   "vimage3D\0"
   "vimage2DMS\0"
   "vimage2DMSArray\0"
   "vimage1DArray\0"
   "vimage2DArray\0"
   "subpassInput\0"
   "subpassInputMS\0"
   "isubpassInput\0"
   "isubpassInputMS\0"
   "usubpassInput\0"
   "usubpassInputMS\0"
   "textureSubpassInput\0"
   "textureSubpassInputMS\0"
   "itextureSubpassInput\0"
   "itextureSubpassInputMS\0"
   "utextureSubpassInput\0"
   "utextureSubpassInputMS\0"
;

const struct glsl_type glsl_type_builtin_error = {
                    .gl_type = GL_INVALID_ENUM,
          .base_type = GLSL_TYPE_ERROR,
          .vector_elements = 0,
          .matrix_columns = 0,
          .name_id = 8,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_void = {
                    .gl_type = GL_INVALID_ENUM,
          .base_type = GLSL_TYPE_VOID,
          .vector_elements = 0,
          .matrix_columns = 0,
          .name_id = 14,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bool = {
                    .gl_type = GL_BOOL,
          .base_type = GLSL_TYPE_BOOL,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 19,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bvec2 = {
                    .gl_type = GL_BOOL_VEC2,
          .base_type = GLSL_TYPE_BOOL,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 24,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bvec3 = {
                    .gl_type = GL_BOOL_VEC3,
          .base_type = GLSL_TYPE_BOOL,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 30,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bvec4 = {
                    .gl_type = GL_BOOL_VEC4,
          .base_type = GLSL_TYPE_BOOL,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 36,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bvec5 = {
                              .base_type = GLSL_TYPE_BOOL,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 42,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bvec8 = {
                              .base_type = GLSL_TYPE_BOOL,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 48,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bvec16 = {
                              .base_type = GLSL_TYPE_BOOL,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 54,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_int = {
                    .gl_type = GL_INT,
          .base_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 61,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_ivec2 = {
                    .gl_type = GL_INT_VEC2,
          .base_type = GLSL_TYPE_INT,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 65,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_ivec3 = {
                    .gl_type = GL_INT_VEC3,
          .base_type = GLSL_TYPE_INT,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 71,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_ivec4 = {
                    .gl_type = GL_INT_VEC4,
          .base_type = GLSL_TYPE_INT,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 77,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_ivec5 = {
                              .base_type = GLSL_TYPE_INT,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 83,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_ivec8 = {
                              .base_type = GLSL_TYPE_INT,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 89,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_ivec16 = {
                              .base_type = GLSL_TYPE_INT,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 95,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uint = {
                    .gl_type = GL_UNSIGNED_INT,
          .base_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 102,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uvec2 = {
                    .gl_type = GL_UNSIGNED_INT_VEC2,
          .base_type = GLSL_TYPE_UINT,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 107,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uvec3 = {
                    .gl_type = GL_UNSIGNED_INT_VEC3,
          .base_type = GLSL_TYPE_UINT,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 113,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uvec4 = {
                    .gl_type = GL_UNSIGNED_INT_VEC4,
          .base_type = GLSL_TYPE_UINT,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 119,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uvec5 = {
                              .base_type = GLSL_TYPE_UINT,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 125,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uvec8 = {
                              .base_type = GLSL_TYPE_UINT,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 131,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uvec16 = {
                              .base_type = GLSL_TYPE_UINT,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 137,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_float = {
                    .gl_type = GL_FLOAT,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 144,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vec2 = {
                    .gl_type = GL_FLOAT_VEC2,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 150,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vec3 = {
                    .gl_type = GL_FLOAT_VEC3,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 155,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vec4 = {
                    .gl_type = GL_FLOAT_VEC4,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 160,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vec5 = {
                              .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 165,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vec8 = {
                              .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 170,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vec16 = {
                              .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 175,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_float16_t = {
                    .gl_type = GL_FLOAT16_NV,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 181,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16vec2 = {
                    .gl_type = GL_FLOAT16_VEC2_NV,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 191,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16vec3 = {
                    .gl_type = GL_FLOAT16_VEC3_NV,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 199,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16vec4 = {
                    .gl_type = GL_FLOAT16_VEC4_NV,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 207,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16vec5 = {
                              .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 215,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16vec8 = {
                              .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 223,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16vec16 = {
                              .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 231,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_double = {
                    .gl_type = GL_DOUBLE,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 240,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dvec2 = {
                    .gl_type = GL_DOUBLE_VEC2,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 247,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dvec3 = {
                    .gl_type = GL_DOUBLE_VEC3,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 253,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dvec4 = {
                    .gl_type = GL_DOUBLE_VEC4,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 259,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dvec5 = {
                              .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 265,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dvec8 = {
                              .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 271,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dvec16 = {
                              .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 277,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_int64_t = {
                    .gl_type = GL_INT64_ARB,
          .base_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 284,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64vec2 = {
                    .gl_type = GL_INT64_VEC2_ARB,
          .base_type = GLSL_TYPE_INT64,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 292,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64vec3 = {
                    .gl_type = GL_INT64_VEC3_ARB,
          .base_type = GLSL_TYPE_INT64,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 300,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64vec4 = {
                    .gl_type = GL_INT64_VEC4_ARB,
          .base_type = GLSL_TYPE_INT64,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 308,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64vec5 = {
                              .base_type = GLSL_TYPE_INT64,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 316,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64vec8 = {
                              .base_type = GLSL_TYPE_INT64,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 324,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64vec16 = {
                              .base_type = GLSL_TYPE_INT64,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 332,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uint64_t = {
                    .gl_type = GL_UNSIGNED_INT64_ARB,
          .base_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 341,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64vec2 = {
                    .gl_type = GL_UNSIGNED_INT64_VEC2_ARB,
          .base_type = GLSL_TYPE_UINT64,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 350,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64vec3 = {
                    .gl_type = GL_UNSIGNED_INT64_VEC3_ARB,
          .base_type = GLSL_TYPE_UINT64,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 358,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64vec4 = {
                    .gl_type = GL_UNSIGNED_INT64_VEC4_ARB,
          .base_type = GLSL_TYPE_UINT64,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 366,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64vec5 = {
                              .base_type = GLSL_TYPE_UINT64,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 374,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64vec8 = {
                              .base_type = GLSL_TYPE_UINT64,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 382,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64vec16 = {
                              .base_type = GLSL_TYPE_UINT64,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 390,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_int16_t = {
                    .gl_type = GL_INT16_NV,
          .base_type = GLSL_TYPE_INT16,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 399,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i16vec2 = {
                    .gl_type = GL_INT16_VEC2_NV,
          .base_type = GLSL_TYPE_INT16,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 407,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i16vec3 = {
                    .gl_type = GL_INT16_VEC3_NV,
          .base_type = GLSL_TYPE_INT16,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 415,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i16vec4 = {
                    .gl_type = GL_INT16_VEC4_NV,
          .base_type = GLSL_TYPE_INT16,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 423,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i16vec5 = {
                              .base_type = GLSL_TYPE_INT16,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 431,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i16vec8 = {
                              .base_type = GLSL_TYPE_INT16,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 439,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i16vec16 = {
                              .base_type = GLSL_TYPE_INT16,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 447,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uint16_t = {
                    .gl_type = GL_UNSIGNED_INT16_NV,
          .base_type = GLSL_TYPE_UINT16,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 456,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u16vec2 = {
                    .gl_type = GL_UNSIGNED_INT16_VEC2_NV,
          .base_type = GLSL_TYPE_UINT16,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 465,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u16vec3 = {
                    .gl_type = GL_UNSIGNED_INT16_VEC3_NV,
          .base_type = GLSL_TYPE_UINT16,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 473,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u16vec4 = {
                    .gl_type = GL_UNSIGNED_INT16_VEC4_NV,
          .base_type = GLSL_TYPE_UINT16,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 481,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u16vec5 = {
                              .base_type = GLSL_TYPE_UINT16,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 489,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u16vec8 = {
                              .base_type = GLSL_TYPE_UINT16,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 497,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u16vec16 = {
                              .base_type = GLSL_TYPE_UINT16,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 505,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_int8_t = {
                    .gl_type = GL_INT8_NV,
          .base_type = GLSL_TYPE_INT8,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 514,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i8vec2 = {
                    .gl_type = GL_INT8_VEC2_NV,
          .base_type = GLSL_TYPE_INT8,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 521,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i8vec3 = {
                    .gl_type = GL_INT8_VEC3_NV,
          .base_type = GLSL_TYPE_INT8,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 528,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i8vec4 = {
                    .gl_type = GL_INT8_VEC4_NV,
          .base_type = GLSL_TYPE_INT8,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 535,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i8vec5 = {
                              .base_type = GLSL_TYPE_INT8,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 542,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i8vec8 = {
                              .base_type = GLSL_TYPE_INT8,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 549,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i8vec16 = {
                              .base_type = GLSL_TYPE_INT8,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 556,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uint8_t = {
                    .gl_type = GL_UNSIGNED_INT8_NV,
          .base_type = GLSL_TYPE_UINT8,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 564,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u8vec2 = {
                    .gl_type = GL_UNSIGNED_INT8_VEC2_NV,
          .base_type = GLSL_TYPE_UINT8,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 572,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u8vec3 = {
                    .gl_type = GL_UNSIGNED_INT8_VEC3_NV,
          .base_type = GLSL_TYPE_UINT8,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 579,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u8vec4 = {
                    .gl_type = GL_UNSIGNED_INT8_VEC4_NV,
          .base_type = GLSL_TYPE_UINT8,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 586,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u8vec5 = {
                              .base_type = GLSL_TYPE_UINT8,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 593,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u8vec8 = {
                              .base_type = GLSL_TYPE_UINT8,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 600,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u8vec16 = {
                              .base_type = GLSL_TYPE_UINT8,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 607,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bfloat16_t = {
                              .base_type = GLSL_TYPE_BFLOAT16,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 615,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bf16vec2 = {
                              .base_type = GLSL_TYPE_BFLOAT16,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 626,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bf16vec3 = {
                              .base_type = GLSL_TYPE_BFLOAT16,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 635,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bf16vec4 = {
                              .base_type = GLSL_TYPE_BFLOAT16,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 644,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bf16vec5 = {
                              .base_type = GLSL_TYPE_BFLOAT16,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 653,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bf16vec8 = {
                              .base_type = GLSL_TYPE_BFLOAT16,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 662,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_bf16vec16 = {
                              .base_type = GLSL_TYPE_BFLOAT16,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 671,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e4m3fn_t = {
                              .base_type = GLSL_TYPE_FLOAT_E4M3FN,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 681,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e4m3fnvec2 = {
                              .base_type = GLSL_TYPE_FLOAT_E4M3FN,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 690,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e4m3fnvec3 = {
                              .base_type = GLSL_TYPE_FLOAT_E4M3FN,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 701,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e4m3fnvec4 = {
                              .base_type = GLSL_TYPE_FLOAT_E4M3FN,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 712,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e4m3fnvec5 = {
                              .base_type = GLSL_TYPE_FLOAT_E4M3FN,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 723,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e4m3fnvec8 = {
                              .base_type = GLSL_TYPE_FLOAT_E4M3FN,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 734,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e4m3fnvec16 = {
                              .base_type = GLSL_TYPE_FLOAT_E4M3FN,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 745,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e5m2_t = {
                              .base_type = GLSL_TYPE_FLOAT_E5M2,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 757,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e5m2vec2 = {
                              .base_type = GLSL_TYPE_FLOAT_E5M2,
          .vector_elements = 2,
          .matrix_columns = 1,
          .name_id = 764,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e5m2vec3 = {
                              .base_type = GLSL_TYPE_FLOAT_E5M2,
          .vector_elements = 3,
          .matrix_columns = 1,
          .name_id = 773,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e5m2vec4 = {
                              .base_type = GLSL_TYPE_FLOAT_E5M2,
          .vector_elements = 4,
          .matrix_columns = 1,
          .name_id = 782,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e5m2vec5 = {
                              .base_type = GLSL_TYPE_FLOAT_E5M2,
          .vector_elements = 5,
          .matrix_columns = 1,
          .name_id = 791,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e5m2vec8 = {
                              .base_type = GLSL_TYPE_FLOAT_E5M2,
          .vector_elements = 8,
          .matrix_columns = 1,
          .name_id = 800,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_e5m2vec16 = {
                              .base_type = GLSL_TYPE_FLOAT_E5M2,
          .vector_elements = 16,
          .matrix_columns = 1,
          .name_id = 809,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_mat2 = {
                    .gl_type = GL_FLOAT_MAT2,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 2,
          .matrix_columns = 2,
          .name_id = 819,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_mat3 = {
                    .gl_type = GL_FLOAT_MAT3,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 3,
          .matrix_columns = 3,
          .name_id = 824,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_mat4 = {
                    .gl_type = GL_FLOAT_MAT4,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 4,
          .matrix_columns = 4,
          .name_id = 829,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_mat2x3 = {
                    .gl_type = GL_FLOAT_MAT2x3,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 3,
          .matrix_columns = 2,
          .name_id = 834,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_mat2x4 = {
                    .gl_type = GL_FLOAT_MAT2x4,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 4,
          .matrix_columns = 2,
          .name_id = 841,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_mat3x2 = {
                    .gl_type = GL_FLOAT_MAT3x2,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 2,
          .matrix_columns = 3,
          .name_id = 848,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_mat3x4 = {
                    .gl_type = GL_FLOAT_MAT3x4,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 4,
          .matrix_columns = 3,
          .name_id = 855,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_mat4x2 = {
                    .gl_type = GL_FLOAT_MAT4x2,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 2,
          .matrix_columns = 4,
          .name_id = 862,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_mat4x3 = {
                    .gl_type = GL_FLOAT_MAT4x3,
          .base_type = GLSL_TYPE_FLOAT,
          .vector_elements = 3,
          .matrix_columns = 4,
          .name_id = 869,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16mat2 = {
                    .gl_type = GL_FLOAT16_MAT2_AMD,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 2,
          .matrix_columns = 2,
          .name_id = 876,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16mat3 = {
                    .gl_type = GL_FLOAT16_MAT3_AMD,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 3,
          .matrix_columns = 3,
          .name_id = 884,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16mat4 = {
                    .gl_type = GL_FLOAT16_MAT4_AMD,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 4,
          .matrix_columns = 4,
          .name_id = 892,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16mat2x3 = {
                    .gl_type = GL_FLOAT16_MAT2x3_AMD,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 3,
          .matrix_columns = 2,
          .name_id = 900,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16mat2x4 = {
                    .gl_type = GL_FLOAT16_MAT2x4_AMD,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 4,
          .matrix_columns = 2,
          .name_id = 910,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16mat3x2 = {
                    .gl_type = GL_FLOAT16_MAT3x2_AMD,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 2,
          .matrix_columns = 3,
          .name_id = 920,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16mat3x4 = {
                    .gl_type = GL_FLOAT16_MAT3x4_AMD,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 4,
          .matrix_columns = 3,
          .name_id = 930,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16mat4x2 = {
                    .gl_type = GL_FLOAT16_MAT4x2_AMD,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 2,
          .matrix_columns = 4,
          .name_id = 940,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_f16mat4x3 = {
                    .gl_type = GL_FLOAT16_MAT4x3_AMD,
          .base_type = GLSL_TYPE_FLOAT16,
          .vector_elements = 3,
          .matrix_columns = 4,
          .name_id = 950,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dmat2 = {
                    .gl_type = GL_DOUBLE_MAT2,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 2,
          .matrix_columns = 2,
          .name_id = 960,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dmat3 = {
                    .gl_type = GL_DOUBLE_MAT3,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 3,
          .matrix_columns = 3,
          .name_id = 966,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dmat4 = {
                    .gl_type = GL_DOUBLE_MAT4,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 4,
          .matrix_columns = 4,
          .name_id = 972,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dmat2x3 = {
                    .gl_type = GL_DOUBLE_MAT2x3,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 3,
          .matrix_columns = 2,
          .name_id = 978,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dmat2x4 = {
                    .gl_type = GL_DOUBLE_MAT2x4,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 4,
          .matrix_columns = 2,
          .name_id = 986,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dmat3x2 = {
                    .gl_type = GL_DOUBLE_MAT3x2,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 2,
          .matrix_columns = 3,
          .name_id = 994,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dmat3x4 = {
                    .gl_type = GL_DOUBLE_MAT3x4,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 4,
          .matrix_columns = 3,
          .name_id = 1002,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dmat4x2 = {
                    .gl_type = GL_DOUBLE_MAT4x2,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 2,
          .matrix_columns = 4,
          .name_id = 1010,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_dmat4x3 = {
                    .gl_type = GL_DOUBLE_MAT4x3,
          .base_type = GLSL_TYPE_DOUBLE,
          .vector_elements = 3,
          .matrix_columns = 4,
          .name_id = 1018,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_atomic_uint = {
                    .gl_type = GL_UNSIGNED_INT_ATOMIC_COUNTER,
          .base_type = GLSL_TYPE_ATOMIC_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1026,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler = {
                    .gl_type = GL_SAMPLER_1D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1038,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler1D = {
                    .gl_type = GL_SAMPLER_1D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1046,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler2D = {
                    .gl_type = GL_SAMPLER_2D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1056,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler3D = {
                    .gl_type = GL_SAMPLER_3D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1066,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_samplerCube = {
                    .gl_type = GL_SAMPLER_CUBE,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1076,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler1DArray = {
                    .gl_type = GL_SAMPLER_1D_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1088,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler2DArray = {
                    .gl_type = GL_SAMPLER_2D_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1103,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_samplerCubeArray = {
                    .gl_type = GL_SAMPLER_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1118,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler2DRect = {
                    .gl_type = GL_SAMPLER_2D_RECT,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1135,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_samplerBuffer = {
                    .gl_type = GL_SAMPLER_BUFFER,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1149,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler2DMS = {
                    .gl_type = GL_SAMPLER_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1163,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler2DMSArray = {
                    .gl_type = GL_SAMPLER_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1175,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isampler1D = {
                    .gl_type = GL_INT_SAMPLER_1D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1192,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isampler2D = {
                    .gl_type = GL_INT_SAMPLER_2D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1203,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isampler3D = {
                    .gl_type = GL_INT_SAMPLER_3D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1214,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isamplerCube = {
                    .gl_type = GL_INT_SAMPLER_CUBE,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1225,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isampler1DArray = {
                    .gl_type = GL_INT_SAMPLER_1D_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1238,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isampler2DArray = {
                    .gl_type = GL_INT_SAMPLER_2D_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1254,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isamplerCubeArray = {
                    .gl_type = GL_INT_SAMPLER_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1270,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isampler2DRect = {
                    .gl_type = GL_INT_SAMPLER_2D_RECT,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1288,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isamplerBuffer = {
                    .gl_type = GL_INT_SAMPLER_BUFFER,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1303,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isampler2DMS = {
                    .gl_type = GL_INT_SAMPLER_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1318,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isampler2DMSArray = {
                    .gl_type = GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1331,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usampler1D = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_1D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1349,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usampler2D = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1360,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usampler3D = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_3D,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1371,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usamplerCube = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_CUBE,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1382,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usampler1DArray = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_1D_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1395,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usampler2DArray = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1411,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usamplerCubeArray = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1427,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usampler2DRect = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D_RECT,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1445,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usamplerBuffer = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_BUFFER,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1460,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usampler2DMS = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1475,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usampler2DMSArray = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1488,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_samplerShadow = {
                    .gl_type = GL_SAMPLER_1D_SHADOW,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 1,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1506,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler1DShadow = {
                    .gl_type = GL_SAMPLER_1D_SHADOW,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 1,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1520,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler2DShadow = {
                    .gl_type = GL_SAMPLER_2D_SHADOW,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 1,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1536,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_samplerCubeShadow = {
                    .gl_type = GL_SAMPLER_CUBE_SHADOW,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 1,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1552,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler1DArrayShadow = {
                    .gl_type = GL_SAMPLER_1D_ARRAY_SHADOW,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 1,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1570,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler2DArrayShadow = {
                    .gl_type = GL_SAMPLER_2D_ARRAY_SHADOW,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 1,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1591,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_samplerCubeArrayShadow = {
                    .gl_type = GL_SAMPLER_CUBE_MAP_ARRAY_SHADOW,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 1,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1612,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_sampler2DRectShadow = {
                    .gl_type = GL_SAMPLER_2D_RECT_SHADOW,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 1,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1635,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_samplerExternalOES = {
                    .gl_type = GL_SAMPLER_EXTERNAL_OES,
          .base_type = GLSL_TYPE_SAMPLER,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_EXTERNAL,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1655,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_texture1D = {
                    .gl_type = GL_SAMPLER_1D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1674,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_texture2D = {
                    .gl_type = GL_SAMPLER_2D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1684,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_texture3D = {
                    .gl_type = GL_SAMPLER_3D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1694,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_textureCube = {
                    .gl_type = GL_SAMPLER_CUBE,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1704,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_texture1DArray = {
                    .gl_type = GL_SAMPLER_1D_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1716,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_texture2DArray = {
                    .gl_type = GL_SAMPLER_2D_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1731,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_textureCubeArray = {
                    .gl_type = GL_SAMPLER_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1746,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_texture2DRect = {
                    .gl_type = GL_SAMPLER_2D_RECT,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1763,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_textureBuffer = {
                    .gl_type = GL_SAMPLER_BUFFER,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1777,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_texture2DMS = {
                    .gl_type = GL_SAMPLER_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1791,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_texture2DMSArray = {
                    .gl_type = GL_SAMPLER_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1803,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itexture1D = {
                    .gl_type = GL_INT_SAMPLER_1D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1820,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itexture2D = {
                    .gl_type = GL_INT_SAMPLER_2D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1831,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itexture3D = {
                    .gl_type = GL_INT_SAMPLER_3D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1842,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itextureCube = {
                    .gl_type = GL_INT_SAMPLER_CUBE,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1853,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itexture1DArray = {
                    .gl_type = GL_INT_SAMPLER_1D_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1866,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itexture2DArray = {
                    .gl_type = GL_INT_SAMPLER_2D_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1882,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itextureCubeArray = {
                    .gl_type = GL_INT_SAMPLER_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1898,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itexture2DRect = {
                    .gl_type = GL_INT_SAMPLER_2D_RECT,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1916,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itextureBuffer = {
                    .gl_type = GL_INT_SAMPLER_BUFFER,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1931,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itexture2DMS = {
                    .gl_type = GL_INT_SAMPLER_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1946,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itexture2DMSArray = {
                    .gl_type = GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1959,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utexture1D = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_1D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1977,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utexture2D = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1988,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utexture3D = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_3D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 1999,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utextureCube = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_CUBE,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2010,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utexture1DArray = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_1D_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2023,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utexture2DArray = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2039,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utextureCubeArray = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2055,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utexture2DRect = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D_RECT,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2073,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utextureBuffer = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_BUFFER,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2088,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utexture2DMS = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2103,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utexture2DMSArray = {
                    .gl_type = GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2116,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_textureExternalOES = {
                    .gl_type = GL_SAMPLER_EXTERNAL_OES,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_EXTERNAL,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2134,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vtexture1D = {
                    .gl_type = GL_SAMPLER_1D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2153,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vtexture2D = {
                    .gl_type = GL_SAMPLER_2D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2164,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vtexture3D = {
                    .gl_type = GL_SAMPLER_3D,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2175,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vtexture2DMS = {
                    .gl_type = GL_SAMPLER_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2186,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vtexture2DMSArray = {
                    .gl_type = GL_SAMPLER_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2199,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vtexture1DArray = {
                    .gl_type = GL_SAMPLER_1D_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2217,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vtexture2DArray = {
                    .gl_type = GL_SAMPLER_2D_ARRAY,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2233,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vtextureBuffer = {
                    .gl_type = GL_SAMPLER_BUFFER,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2249,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_image1D = {
                    .gl_type = GL_IMAGE_1D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2264,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_image2D = {
                    .gl_type = GL_IMAGE_2D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2272,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_image3D = {
                    .gl_type = GL_IMAGE_3D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2280,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_image2DRect = {
                    .gl_type = GL_IMAGE_2D_RECT,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2288,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_imageCube = {
                    .gl_type = GL_IMAGE_CUBE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2300,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_imageBuffer = {
                    .gl_type = GL_IMAGE_BUFFER,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2310,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_image1DArray = {
                    .gl_type = GL_IMAGE_1D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2322,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_image2DArray = {
                    .gl_type = GL_IMAGE_2D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2335,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_imageCubeArray = {
                    .gl_type = GL_IMAGE_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2348,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_image2DMS = {
                    .gl_type = GL_IMAGE_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2363,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_image2DMSArray = {
                    .gl_type = GL_IMAGE_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2373,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimage1D = {
                    .gl_type = GL_INT_IMAGE_1D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2388,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimage2D = {
                    .gl_type = GL_INT_IMAGE_2D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2397,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimage3D = {
                    .gl_type = GL_INT_IMAGE_3D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2406,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimage2DRect = {
                    .gl_type = GL_INT_IMAGE_2D_RECT,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2415,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimageCube = {
                    .gl_type = GL_INT_IMAGE_CUBE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2428,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimageBuffer = {
                    .gl_type = GL_INT_IMAGE_BUFFER,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2439,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimage1DArray = {
                    .gl_type = GL_INT_IMAGE_1D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2452,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimage2DArray = {
                    .gl_type = GL_INT_IMAGE_2D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2466,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimageCubeArray = {
                    .gl_type = GL_INT_IMAGE_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2480,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimage2DMS = {
                    .gl_type = GL_INT_IMAGE_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2496,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_iimage2DMSArray = {
                    .gl_type = GL_INT_IMAGE_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2507,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimage1D = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_1D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2523,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimage2D = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2532,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimage3D = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_3D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2541,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimage2DRect = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D_RECT,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2550,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimageCube = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_CUBE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2563,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimageBuffer = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_BUFFER,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2574,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimage1DArray = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_1D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2587,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimage2DArray = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2601,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimageCubeArray = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2615,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimage2DMS = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2631,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_uimage2DMSArray = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2642,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64image1D = {
                    .gl_type = GL_INT_IMAGE_1D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2658,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64image2D = {
                    .gl_type = GL_INT_IMAGE_2D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2669,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64image3D = {
                    .gl_type = GL_INT_IMAGE_3D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2680,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64image2DRect = {
                    .gl_type = GL_INT_IMAGE_2D_RECT,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2691,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64imageCube = {
                    .gl_type = GL_INT_IMAGE_CUBE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2706,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64imageBuffer = {
                    .gl_type = GL_INT_IMAGE_BUFFER,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2719,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64image1DArray = {
                    .gl_type = GL_INT_IMAGE_1D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2734,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64image2DArray = {
                    .gl_type = GL_INT_IMAGE_2D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2750,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64imageCubeArray = {
                    .gl_type = GL_INT_IMAGE_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2766,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64image2DMS = {
                    .gl_type = GL_INT_IMAGE_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2784,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_i64image2DMSArray = {
                    .gl_type = GL_INT_IMAGE_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_INT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2797,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64image1D = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_1D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2815,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64image2D = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2826,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64image3D = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_3D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2837,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64image2DRect = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D_RECT,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_RECT,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2848,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64imageCube = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_CUBE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2863,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64imageBuffer = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_BUFFER,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2876,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64image1DArray = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_1D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2891,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64image2DArray = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2907,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64imageCubeArray = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_CUBE_MAP_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_CUBE,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2923,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64image2DMS = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2941,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_u64image2DMSArray = {
                    .gl_type = GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_UINT64,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2954,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vbuffer = {
                    .gl_type = GL_IMAGE_BUFFER,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_BUF,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2972,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vimage1D = {
                    .gl_type = GL_IMAGE_1D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2980,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vimage2D = {
                    .gl_type = GL_IMAGE_2D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2989,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vimage3D = {
                    .gl_type = GL_IMAGE_3D,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_3D,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 2998,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vimage2DMS = {
                    .gl_type = GL_IMAGE_2D_MULTISAMPLE,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3007,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vimage2DMSArray = {
                    .gl_type = GL_IMAGE_2D_MULTISAMPLE_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_MS,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3018,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vimage1DArray = {
                    .gl_type = GL_IMAGE_1D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_1D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3034,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_vimage2DArray = {
                    .gl_type = GL_IMAGE_2D_ARRAY,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_2D,
          .sampler_shadow = 0,
          .sampler_array = 1,
          .sampled_type = GLSL_TYPE_VOID,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3048,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_subpassInput = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3062,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_subpassInputMS = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3075,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isubpassInput = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3090,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_isubpassInputMS = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3104,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usubpassInput = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3120,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_usubpassInputMS = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_IMAGE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3134,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_textureSubpassInput = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3150,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_textureSubpassInputMS = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_FLOAT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3170,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itextureSubpassInput = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3192,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_itextureSubpassInputMS = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_INT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3213,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utextureSubpassInput = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3236,
          .has_builtin_name = 1,
};

const struct glsl_type glsl_type_builtin_utextureSubpassInputMS = {
                    .gl_type = 0,
          .base_type = GLSL_TYPE_TEXTURE,
          .sampler_dimensionality = GLSL_SAMPLER_DIM_SUBPASS_MS,
          .sampler_shadow = 0,
          .sampler_array = 0,
          .sampled_type = GLSL_TYPE_UINT,
          .vector_elements = 1,
          .matrix_columns = 1,
          .name_id = 3257,
          .has_builtin_name = 1,
};

