/*
 * Copyright Collabora Ltd
 * SPDX-License-Identifier: MIT
 *
 * Autogenerated file, do not edit
 */

#include <stdint.h>

#pragma once
#include "util/macros.h"

enum libpan_shaders_program {
    LIBPAN_SHADERS_NUM_PROGRAMS,
};



extern const uint32_t *libpan_shaders_default_v4[LIBPAN_SHADERS_NUM_PROGRAMS];
