
/* Copyright © 2015-2021 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/* This file generated from vk_entrypoints_gen.py, don't edit directly. */

/* This file is the only place we rely on undefined symbols to fall back to
 * NULL. Other files use regular symbol declarations.
 * See also comments on VK_ENTRY_WEAK.
 */
#define VK_ENTRY_USE_WEAK 1
#include "lvp_entrypoints.h"

/* Weak aliases for all potential implementations. These will resolve to
 * NULL if they're not defined, which lets the resolve_entrypoint() function
 * either pick the correct entry point.
 *
 * MSVC uses different decorated names for 32-bit versus 64-bit. Declare
 * all argument sizes for 32-bit because computing the actual size would be
 * difficult.
 */




#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateInstance@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateInstance=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateInstance=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyInstance@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyInstance=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyInstance=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDevices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumeratePhysicalDevices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumeratePhysicalDevices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetInstanceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetInstanceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetInstanceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceVersion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumerateInstanceVersion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumerateInstanceVersion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceLayerProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumerateInstanceLayerProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumerateInstanceLayerProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateInstanceExtensionProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumerateInstanceExtensionProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumerateInstanceExtensionProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAndroidSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateAndroidSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateAndroidSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSurfaceOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateSurfaceOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateSurfaceOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayPlaneSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDisplayPlaneSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDisplayPlaneSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroySurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroySurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_VI_NN
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateViSurfaceNN@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateViSurfaceNN=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateViSurfaceNN=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_VI_NN
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWaylandSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateWaylandSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateWaylandSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateUbmSurfaceSEC@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateUbmSurfaceSEC=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateUbmSurfaceSEC=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateWin32SurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateWin32SurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateWin32SurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXlibSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateXlibSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateXlibSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateXcbSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateXcbSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateXcbSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDirectFBSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDirectFBSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDirectFBSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImagePipeSurfaceFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateImagePipeSurfaceFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateImagePipeSurfaceFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_GGP
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateStreamDescriptorSurfaceGGP@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateStreamDescriptorSurfaceGGP=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateStreamDescriptorSurfaceGGP=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_GGP
#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateScreenSurfaceQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateScreenSurfaceQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateScreenSurfaceQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugReportCallbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDebugReportCallbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDebugReportCallbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugReportCallbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyDebugReportCallbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyDebugReportCallbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugReportMessageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DebugReportMessageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DebugReportMessageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroups@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumeratePhysicalDeviceGroups=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumeratePhysicalDeviceGroups=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceGroupsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumeratePhysicalDeviceGroupsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumeratePhysicalDeviceGroupsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_IOS_MVK
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIOSSurfaceMVK@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateIOSSurfaceMVK=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateIOSSurfaceMVK=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_IOS_MVK
#ifdef VK_USE_PLATFORM_MACOS_MVK
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMacOSSurfaceMVK@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateMacOSSurfaceMVK=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateMacOSSurfaceMVK=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_MACOS_MVK
#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMetalSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateMetalSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateMetalSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDebugUtilsMessengerEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDebugUtilsMessengerEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDebugUtilsMessengerEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDebugUtilsMessengerEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyDebugUtilsMessengerEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyDebugUtilsMessengerEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SubmitDebugUtilsMessageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SubmitDebugUtilsMessageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SubmitDebugUtilsMessageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateHeadlessSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateHeadlessSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateHeadlessSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetExternalComputeQueueDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetExternalComputeQueueDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetExternalComputeQueueDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_instance_entrypoint_table lvp_instance_entrypoints = {
    .CreateInstance = lvp_CreateInstance,
    .DestroyInstance = lvp_DestroyInstance,
    .EnumeratePhysicalDevices = lvp_EnumeratePhysicalDevices,
    .GetInstanceProcAddr = lvp_GetInstanceProcAddr,
    .EnumerateInstanceVersion = lvp_EnumerateInstanceVersion,
    .EnumerateInstanceLayerProperties = lvp_EnumerateInstanceLayerProperties,
    .EnumerateInstanceExtensionProperties = lvp_EnumerateInstanceExtensionProperties,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .CreateAndroidSurfaceKHR = lvp_CreateAndroidSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateAndroidSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_OHOS
    .CreateSurfaceOHOS = lvp_CreateSurfaceOHOS,
#elif defined(_MSC_VER)
    .CreateSurfaceOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .CreateDisplayPlaneSurfaceKHR = lvp_CreateDisplayPlaneSurfaceKHR,
    .DestroySurfaceKHR = lvp_DestroySurfaceKHR,
#ifdef VK_USE_PLATFORM_VI_NN
    .CreateViSurfaceNN = lvp_CreateViSurfaceNN,
#elif defined(_MSC_VER)
    .CreateViSurfaceNN = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_VI_NN
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
    .CreateWaylandSurfaceKHR = lvp_CreateWaylandSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateWaylandSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
    .CreateUbmSurfaceSEC = lvp_CreateUbmSurfaceSEC,
#elif defined(_MSC_VER)
    .CreateUbmSurfaceSEC = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .CreateWin32SurfaceKHR = lvp_CreateWin32SurfaceKHR,
#elif defined(_MSC_VER)
    .CreateWin32SurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
    .CreateXlibSurfaceKHR = lvp_CreateXlibSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateXlibSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
    .CreateXcbSurfaceKHR = lvp_CreateXcbSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateXcbSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
    .CreateDirectFBSurfaceEXT = lvp_CreateDirectFBSurfaceEXT,
#elif defined(_MSC_VER)
    .CreateDirectFBSurfaceEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateImagePipeSurfaceFUCHSIA = lvp_CreateImagePipeSurfaceFUCHSIA,
#elif defined(_MSC_VER)
    .CreateImagePipeSurfaceFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_GGP
    .CreateStreamDescriptorSurfaceGGP = lvp_CreateStreamDescriptorSurfaceGGP,
#elif defined(_MSC_VER)
    .CreateStreamDescriptorSurfaceGGP = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_GGP
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .CreateScreenSurfaceQNX = lvp_CreateScreenSurfaceQNX,
#elif defined(_MSC_VER)
    .CreateScreenSurfaceQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CreateDebugReportCallbackEXT = lvp_CreateDebugReportCallbackEXT,
    .DestroyDebugReportCallbackEXT = lvp_DestroyDebugReportCallbackEXT,
    .DebugReportMessageEXT = lvp_DebugReportMessageEXT,
    .EnumeratePhysicalDeviceGroups = lvp_EnumeratePhysicalDeviceGroups,
    .EnumeratePhysicalDeviceGroupsKHR = lvp_EnumeratePhysicalDeviceGroupsKHR,
#ifdef VK_USE_PLATFORM_IOS_MVK
    .CreateIOSSurfaceMVK = lvp_CreateIOSSurfaceMVK,
#elif defined(_MSC_VER)
    .CreateIOSSurfaceMVK = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_IOS_MVK
#ifdef VK_USE_PLATFORM_MACOS_MVK
    .CreateMacOSSurfaceMVK = lvp_CreateMacOSSurfaceMVK,
#elif defined(_MSC_VER)
    .CreateMacOSSurfaceMVK = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_MACOS_MVK
#ifdef VK_USE_PLATFORM_METAL_EXT
    .CreateMetalSurfaceEXT = lvp_CreateMetalSurfaceEXT,
#elif defined(_MSC_VER)
    .CreateMetalSurfaceEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CreateDebugUtilsMessengerEXT = lvp_CreateDebugUtilsMessengerEXT,
    .DestroyDebugUtilsMessengerEXT = lvp_DestroyDebugUtilsMessengerEXT,
    .SubmitDebugUtilsMessageEXT = lvp_SubmitDebugUtilsMessageEXT,
    .CreateHeadlessSurfaceEXT = lvp_CreateHeadlessSurfaceEXT,
    .GetExternalComputeQueueDataNV = lvp_GetExternalComputeQueueDataNV,
};



#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceQueueFamilyProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceQueueFamilyProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceMemoryProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceMemoryProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceImageFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceImageFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceLayerProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumerateDeviceLayerProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumerateDeviceLayerProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumerateDeviceExtensionProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumerateDeviceExtensionProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumerateDeviceExtensionProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSparseImageFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSparseImageFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceDisplayPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceDisplayPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneSupportedDisplaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDisplayPlaneSupportedDisplaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDisplayPlaneSupportedDisplaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDisplayModePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDisplayModePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDisplayModeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDisplayModeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDisplayModeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDisplayPlaneCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDisplayPlaneCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSurfaceSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSurfaceSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormatsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSurfaceFormatsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSurfaceFormatsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WAYLAND_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWaylandPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceWaylandPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceWaylandPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceUbmPresentationSupportSEC@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceUbmPresentationSupportSEC=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceUbmPresentationSupportSEC=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceWin32PresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceWin32PresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceWin32PresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXlibPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceXlibPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceXlibPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceXcbPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceXcbPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceXcbPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceScreenPresentationSupportQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceScreenPresentationSupportQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceScreenPresentationSupportQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceFeatures2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceFeatures2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFeatures2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceFeatures2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceFeatures2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceImageFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceImageFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceImageFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceImageFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceImageFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceQueueFamilyProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceQueueFamilyProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceQueueFamilyProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceQueueFamilyProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceMemoryProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceMemoryProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMemoryProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceMemoryProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceMemoryProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSparseImageFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSparseImageFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceExternalBufferProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceExternalBufferProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalBufferPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceExternalBufferPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceExternalBufferPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphoreProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceExternalSemaphoreProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceExternalSemaphoreProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFenceProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceExternalFenceProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceExternalFenceProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalFencePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceExternalFencePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceExternalFencePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ReleaseDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ReleaseDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireXlibDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AcquireXlibDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AcquireXlibDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRandROutputDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetRandROutputDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetRandROutputDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireWinrtDisplayNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AcquireWinrtDisplayNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AcquireWinrtDisplayNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetWinrtDisplayNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetWinrtDisplayNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetWinrtDisplayNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSurfaceCapabilities2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSurfaceCapabilities2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDevicePresentRectanglesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDevicePresentRectanglesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDevicePresentRectanglesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceMultisamplePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceMultisamplePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceMultisamplePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceCapabilities2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSurfaceCapabilities2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSurfaceCapabilities2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfaceFormats2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSurfaceFormats2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSurfaceFormats2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceDisplayProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceDisplayProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayModeProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDisplayModeProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDisplayModeProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDisplayPlaneCapabilities2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDisplayPlaneCapabilities2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDisplayPlaneCapabilities2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceToolProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceToolProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceToolPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceToolPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceToolPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceFragmentShadingRatesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceFragmentShadingRatesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceFragmentShadingRatesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceVideoCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceVideoCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoFormatPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceVideoFormatPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceVideoFormatPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireDrmDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AcquireDrmDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AcquireDrmDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDrmDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDrmDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDrmDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceExternalTensorPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceExternalTensorPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceExternalTensorPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceDescriptorSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceDescriptorSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceDescriptorSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_physical_device_entrypoint_table lvp_physical_device_entrypoints = {
    .GetPhysicalDeviceProperties = lvp_GetPhysicalDeviceProperties,
    .GetPhysicalDeviceQueueFamilyProperties = lvp_GetPhysicalDeviceQueueFamilyProperties,
    .GetPhysicalDeviceMemoryProperties = lvp_GetPhysicalDeviceMemoryProperties,
    .GetPhysicalDeviceFeatures = lvp_GetPhysicalDeviceFeatures,
    .GetPhysicalDeviceFormatProperties = lvp_GetPhysicalDeviceFormatProperties,
    .GetPhysicalDeviceImageFormatProperties = lvp_GetPhysicalDeviceImageFormatProperties,
    .CreateDevice = lvp_CreateDevice,
    .EnumerateDeviceLayerProperties = lvp_EnumerateDeviceLayerProperties,
    .EnumerateDeviceExtensionProperties = lvp_EnumerateDeviceExtensionProperties,
    .GetPhysicalDeviceSparseImageFormatProperties = lvp_GetPhysicalDeviceSparseImageFormatProperties,
    .GetPhysicalDeviceDisplayPropertiesKHR = lvp_GetPhysicalDeviceDisplayPropertiesKHR,
    .GetPhysicalDeviceDisplayPlanePropertiesKHR = lvp_GetPhysicalDeviceDisplayPlanePropertiesKHR,
    .GetDisplayPlaneSupportedDisplaysKHR = lvp_GetDisplayPlaneSupportedDisplaysKHR,
    .GetDisplayModePropertiesKHR = lvp_GetDisplayModePropertiesKHR,
    .CreateDisplayModeKHR = lvp_CreateDisplayModeKHR,
    .GetDisplayPlaneCapabilitiesKHR = lvp_GetDisplayPlaneCapabilitiesKHR,
    .GetPhysicalDeviceSurfaceSupportKHR = lvp_GetPhysicalDeviceSurfaceSupportKHR,
    .GetPhysicalDeviceSurfaceCapabilitiesKHR = lvp_GetPhysicalDeviceSurfaceCapabilitiesKHR,
    .GetPhysicalDeviceSurfaceFormatsKHR = lvp_GetPhysicalDeviceSurfaceFormatsKHR,
    .GetPhysicalDeviceSurfacePresentModesKHR = lvp_GetPhysicalDeviceSurfacePresentModesKHR,
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
    .GetPhysicalDeviceWaylandPresentationSupportKHR = lvp_GetPhysicalDeviceWaylandPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceWaylandPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
    .GetPhysicalDeviceUbmPresentationSupportSEC = lvp_GetPhysicalDeviceUbmPresentationSupportSEC,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceUbmPresentationSupportSEC = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceWin32PresentationSupportKHR = lvp_GetPhysicalDeviceWin32PresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceWin32PresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
    .GetPhysicalDeviceXlibPresentationSupportKHR = lvp_GetPhysicalDeviceXlibPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceXlibPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
    .GetPhysicalDeviceXcbPresentationSupportKHR = lvp_GetPhysicalDeviceXcbPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceXcbPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
    .GetPhysicalDeviceDirectFBPresentationSupportEXT = lvp_GetPhysicalDeviceDirectFBPresentationSupportEXT,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceDirectFBPresentationSupportEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetPhysicalDeviceScreenPresentationSupportQNX = lvp_GetPhysicalDeviceScreenPresentationSupportQNX,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceScreenPresentationSupportQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .GetPhysicalDeviceExternalImageFormatPropertiesNV = lvp_GetPhysicalDeviceExternalImageFormatPropertiesNV,
    .GetPhysicalDeviceFeatures2 = lvp_GetPhysicalDeviceFeatures2,
    .GetPhysicalDeviceFeatures2KHR = lvp_GetPhysicalDeviceFeatures2KHR,
    .GetPhysicalDeviceProperties2 = lvp_GetPhysicalDeviceProperties2,
    .GetPhysicalDeviceProperties2KHR = lvp_GetPhysicalDeviceProperties2KHR,
    .GetPhysicalDeviceFormatProperties2 = lvp_GetPhysicalDeviceFormatProperties2,
    .GetPhysicalDeviceFormatProperties2KHR = lvp_GetPhysicalDeviceFormatProperties2KHR,
    .GetPhysicalDeviceImageFormatProperties2 = lvp_GetPhysicalDeviceImageFormatProperties2,
    .GetPhysicalDeviceImageFormatProperties2KHR = lvp_GetPhysicalDeviceImageFormatProperties2KHR,
    .GetPhysicalDeviceQueueFamilyProperties2 = lvp_GetPhysicalDeviceQueueFamilyProperties2,
    .GetPhysicalDeviceQueueFamilyProperties2KHR = lvp_GetPhysicalDeviceQueueFamilyProperties2KHR,
    .GetPhysicalDeviceMemoryProperties2 = lvp_GetPhysicalDeviceMemoryProperties2,
    .GetPhysicalDeviceMemoryProperties2KHR = lvp_GetPhysicalDeviceMemoryProperties2KHR,
    .GetPhysicalDeviceSparseImageFormatProperties2 = lvp_GetPhysicalDeviceSparseImageFormatProperties2,
    .GetPhysicalDeviceSparseImageFormatProperties2KHR = lvp_GetPhysicalDeviceSparseImageFormatProperties2KHR,
    .GetPhysicalDeviceExternalBufferProperties = lvp_GetPhysicalDeviceExternalBufferProperties,
    .GetPhysicalDeviceExternalBufferPropertiesKHR = lvp_GetPhysicalDeviceExternalBufferPropertiesKHR,
    .GetPhysicalDeviceExternalSemaphoreProperties = lvp_GetPhysicalDeviceExternalSemaphoreProperties,
    .GetPhysicalDeviceExternalSemaphorePropertiesKHR = lvp_GetPhysicalDeviceExternalSemaphorePropertiesKHR,
    .GetPhysicalDeviceExternalFenceProperties = lvp_GetPhysicalDeviceExternalFenceProperties,
    .GetPhysicalDeviceExternalFencePropertiesKHR = lvp_GetPhysicalDeviceExternalFencePropertiesKHR,
    .ReleaseDisplayEXT = lvp_ReleaseDisplayEXT,
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
    .AcquireXlibDisplayEXT = lvp_AcquireXlibDisplayEXT,
#elif defined(_MSC_VER)
    .AcquireXlibDisplayEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
    .GetRandROutputDisplayEXT = lvp_GetRandROutputDisplayEXT,
#elif defined(_MSC_VER)
    .GetRandROutputDisplayEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireWinrtDisplayNV = lvp_AcquireWinrtDisplayNV,
#elif defined(_MSC_VER)
    .AcquireWinrtDisplayNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetWinrtDisplayNV = lvp_GetWinrtDisplayNV,
#elif defined(_MSC_VER)
    .GetWinrtDisplayNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceSurfaceCapabilities2EXT = lvp_GetPhysicalDeviceSurfaceCapabilities2EXT,
    .GetPhysicalDevicePresentRectanglesKHR = lvp_GetPhysicalDevicePresentRectanglesKHR,
    .GetPhysicalDeviceMultisamplePropertiesEXT = lvp_GetPhysicalDeviceMultisamplePropertiesEXT,
    .GetPhysicalDeviceSurfaceCapabilities2KHR = lvp_GetPhysicalDeviceSurfaceCapabilities2KHR,
    .GetPhysicalDeviceSurfaceFormats2KHR = lvp_GetPhysicalDeviceSurfaceFormats2KHR,
    .GetPhysicalDeviceDisplayProperties2KHR = lvp_GetPhysicalDeviceDisplayProperties2KHR,
    .GetPhysicalDeviceDisplayPlaneProperties2KHR = lvp_GetPhysicalDeviceDisplayPlaneProperties2KHR,
    .GetDisplayModeProperties2KHR = lvp_GetDisplayModeProperties2KHR,
    .GetDisplayPlaneCapabilities2KHR = lvp_GetDisplayPlaneCapabilities2KHR,
    .GetPhysicalDeviceCalibrateableTimeDomainsKHR = lvp_GetPhysicalDeviceCalibrateableTimeDomainsKHR,
    .GetPhysicalDeviceCalibrateableTimeDomainsEXT = lvp_GetPhysicalDeviceCalibrateableTimeDomainsEXT,
    .GetPhysicalDeviceCooperativeMatrixPropertiesNV = lvp_GetPhysicalDeviceCooperativeMatrixPropertiesNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceSurfacePresentModes2EXT = lvp_GetPhysicalDeviceSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR = lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR,
    .GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR = lvp_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR,
    .GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV = lvp_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV,
    .GetPhysicalDeviceToolProperties = lvp_GetPhysicalDeviceToolProperties,
    .GetPhysicalDeviceToolPropertiesEXT = lvp_GetPhysicalDeviceToolPropertiesEXT,
    .GetPhysicalDeviceFragmentShadingRatesKHR = lvp_GetPhysicalDeviceFragmentShadingRatesKHR,
    .GetPhysicalDeviceVideoCapabilitiesKHR = lvp_GetPhysicalDeviceVideoCapabilitiesKHR,
    .GetPhysicalDeviceVideoFormatPropertiesKHR = lvp_GetPhysicalDeviceVideoFormatPropertiesKHR,
    .GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR = lvp_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR,
    .AcquireDrmDisplayEXT = lvp_AcquireDrmDisplayEXT,
    .GetDrmDisplayEXT = lvp_GetDrmDisplayEXT,
    .GetPhysicalDeviceOpticalFlowImageFormatsNV = lvp_GetPhysicalDeviceOpticalFlowImageFormatsNV,
    .GetPhysicalDeviceCooperativeMatrixPropertiesKHR = lvp_GetPhysicalDeviceCooperativeMatrixPropertiesKHR,
    .GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV = lvp_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV,
    .GetPhysicalDeviceCooperativeVectorPropertiesNV = lvp_GetPhysicalDeviceCooperativeVectorPropertiesNV,
    .EnumeratePhysicalDeviceShaderInstrumentationMetricsARM = lvp_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM,
    .GetPhysicalDeviceExternalTensorPropertiesARM = lvp_GetPhysicalDeviceExternalTensorPropertiesARM,
    .GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM = lvp_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM,
    .GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM = lvp_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM,
    .EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM = lvp_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM,
    .GetPhysicalDeviceDescriptorSizeEXT = lvp_GetPhysicalDeviceDescriptorSizeEXT,
    .GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM = lvp_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM,
};



#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_lvp_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:lvp_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#lvp_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_device_entrypoint_table lvp_device_entrypoints = {
    .GetDeviceProcAddr = lvp_GetDeviceProcAddr,
    .DestroyDevice = lvp_DestroyDevice,
    .GetDeviceQueue = lvp_GetDeviceQueue,
    .QueueSubmit = lvp_QueueSubmit,
    .QueueWaitIdle = lvp_QueueWaitIdle,
    .DeviceWaitIdle = lvp_DeviceWaitIdle,
    .AllocateMemory = lvp_AllocateMemory,
    .FreeMemory = lvp_FreeMemory,
    .MapMemory = lvp_MapMemory,
    .UnmapMemory = lvp_UnmapMemory,
    .FlushMappedMemoryRanges = lvp_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = lvp_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = lvp_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = lvp_GetBufferMemoryRequirements,
    .BindBufferMemory = lvp_BindBufferMemory,
    .GetImageMemoryRequirements = lvp_GetImageMemoryRequirements,
    .BindImageMemory = lvp_BindImageMemory,
    .GetImageSparseMemoryRequirements = lvp_GetImageSparseMemoryRequirements,
    .QueueBindSparse = lvp_QueueBindSparse,
    .CreateFence = lvp_CreateFence,
    .DestroyFence = lvp_DestroyFence,
    .ResetFences = lvp_ResetFences,
    .GetFenceStatus = lvp_GetFenceStatus,
    .WaitForFences = lvp_WaitForFences,
    .CreateSemaphore = lvp_CreateSemaphore,
    .DestroySemaphore = lvp_DestroySemaphore,
    .CreateEvent = lvp_CreateEvent,
    .DestroyEvent = lvp_DestroyEvent,
    .GetEventStatus = lvp_GetEventStatus,
    .SetEvent = lvp_SetEvent,
    .ResetEvent = lvp_ResetEvent,
    .CreateQueryPool = lvp_CreateQueryPool,
    .DestroyQueryPool = lvp_DestroyQueryPool,
    .GetQueryPoolResults = lvp_GetQueryPoolResults,
    .ResetQueryPool = lvp_ResetQueryPool,
    .ResetQueryPoolEXT = lvp_ResetQueryPoolEXT,
    .CreateBuffer = lvp_CreateBuffer,
    .DestroyBuffer = lvp_DestroyBuffer,
    .CreateBufferView = lvp_CreateBufferView,
    .DestroyBufferView = lvp_DestroyBufferView,
    .CreateImage = lvp_CreateImage,
    .DestroyImage = lvp_DestroyImage,
    .GetImageSubresourceLayout = lvp_GetImageSubresourceLayout,
    .CreateImageView = lvp_CreateImageView,
    .DestroyImageView = lvp_DestroyImageView,
    .CreateShaderModule = lvp_CreateShaderModule,
    .DestroyShaderModule = lvp_DestroyShaderModule,
    .CreatePipelineCache = lvp_CreatePipelineCache,
    .DestroyPipelineCache = lvp_DestroyPipelineCache,
    .GetPipelineCacheData = lvp_GetPipelineCacheData,
    .MergePipelineCaches = lvp_MergePipelineCaches,
    .CreatePipelineBinariesKHR = lvp_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = lvp_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = lvp_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = lvp_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = lvp_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = lvp_CreateGraphicsPipelines,
    .CreateComputePipelines = lvp_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = lvp_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = lvp_DestroyPipeline,
    .CreatePipelineLayout = lvp_CreatePipelineLayout,
    .DestroyPipelineLayout = lvp_DestroyPipelineLayout,
    .CreateSampler = lvp_CreateSampler,
    .DestroySampler = lvp_DestroySampler,
    .CreateDescriptorSetLayout = lvp_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = lvp_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = lvp_CreateDescriptorPool,
    .DestroyDescriptorPool = lvp_DestroyDescriptorPool,
    .ResetDescriptorPool = lvp_ResetDescriptorPool,
    .AllocateDescriptorSets = lvp_AllocateDescriptorSets,
    .FreeDescriptorSets = lvp_FreeDescriptorSets,
    .UpdateDescriptorSets = lvp_UpdateDescriptorSets,
    .CreateFramebuffer = lvp_CreateFramebuffer,
    .DestroyFramebuffer = lvp_DestroyFramebuffer,
    .CreateRenderPass = lvp_CreateRenderPass,
    .DestroyRenderPass = lvp_DestroyRenderPass,
    .GetRenderAreaGranularity = lvp_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = lvp_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = lvp_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = lvp_CreateCommandPool,
    .DestroyCommandPool = lvp_DestroyCommandPool,
    .ResetCommandPool = lvp_ResetCommandPool,
    .AllocateCommandBuffers = lvp_AllocateCommandBuffers,
    .FreeCommandBuffers = lvp_FreeCommandBuffers,
    .BeginCommandBuffer = lvp_BeginCommandBuffer,
    .EndCommandBuffer = lvp_EndCommandBuffer,
    .ResetCommandBuffer = lvp_ResetCommandBuffer,
    .CmdBindPipeline = lvp_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = lvp_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = lvp_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = lvp_CmdSetViewport,
    .CmdSetScissor = lvp_CmdSetScissor,
    .CmdSetLineWidth = lvp_CmdSetLineWidth,
    .CmdSetDepthBias = lvp_CmdSetDepthBias,
    .CmdSetBlendConstants = lvp_CmdSetBlendConstants,
    .CmdSetDepthBounds = lvp_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = lvp_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = lvp_CmdSetStencilWriteMask,
    .CmdSetStencilReference = lvp_CmdSetStencilReference,
    .CmdBindDescriptorSets = lvp_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = lvp_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = lvp_CmdBindVertexBuffers,
    .CmdDraw = lvp_CmdDraw,
    .CmdDrawIndexed = lvp_CmdDrawIndexed,
    .CmdDrawMultiEXT = lvp_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = lvp_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = lvp_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = lvp_CmdDrawIndexedIndirect,
    .CmdDispatch = lvp_CmdDispatch,
    .CmdDispatchIndirect = lvp_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = lvp_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = lvp_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = lvp_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = lvp_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = lvp_CmdCopyBuffer,
    .CmdCopyImage = lvp_CmdCopyImage,
    .CmdBlitImage = lvp_CmdBlitImage,
    .CmdCopyBufferToImage = lvp_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = lvp_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = lvp_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = lvp_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = lvp_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = lvp_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = lvp_CmdUpdateBuffer,
    .CmdFillBuffer = lvp_CmdFillBuffer,
    .CmdClearColorImage = lvp_CmdClearColorImage,
    .CmdClearDepthStencilImage = lvp_CmdClearDepthStencilImage,
    .CmdClearAttachments = lvp_CmdClearAttachments,
    .CmdResolveImage = lvp_CmdResolveImage,
    .CmdSetEvent = lvp_CmdSetEvent,
    .CmdResetEvent = lvp_CmdResetEvent,
    .CmdWaitEvents = lvp_CmdWaitEvents,
    .CmdPipelineBarrier = lvp_CmdPipelineBarrier,
    .CmdBeginQuery = lvp_CmdBeginQuery,
    .CmdEndQuery = lvp_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = lvp_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = lvp_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = lvp_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = lvp_CmdResetQueryPool,
    .CmdWriteTimestamp = lvp_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = lvp_CmdCopyQueryPoolResults,
    .CmdPushConstants = lvp_CmdPushConstants,
    .CmdBeginRenderPass = lvp_CmdBeginRenderPass,
    .CmdNextSubpass = lvp_CmdNextSubpass,
    .CmdEndRenderPass = lvp_CmdEndRenderPass,
    .CmdExecuteCommands = lvp_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = lvp_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = lvp_CreateSwapchainKHR,
    .DestroySwapchainKHR = lvp_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = lvp_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = lvp_AcquireNextImageKHR,
    .QueuePresentKHR = lvp_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = lvp_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = lvp_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = lvp_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = lvp_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = lvp_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = lvp_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = lvp_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = lvp_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = lvp_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = lvp_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = lvp_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = lvp_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = lvp_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = lvp_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = lvp_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = lvp_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = lvp_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = lvp_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = lvp_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = lvp_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = lvp_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = lvp_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = lvp_CmdPushDescriptorSetKHR,
    .TrimCommandPool = lvp_TrimCommandPool,
    .TrimCommandPoolKHR = lvp_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = lvp_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = lvp_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = lvp_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = lvp_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = lvp_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = lvp_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = lvp_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = lvp_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = lvp_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = lvp_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = lvp_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = lvp_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = lvp_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = lvp_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = lvp_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = lvp_GetFenceFdKHR,
    .ImportFenceFdKHR = lvp_ImportFenceFdKHR,
    .DisplayPowerControlEXT = lvp_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = lvp_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = lvp_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = lvp_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = lvp_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = lvp_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = lvp_BindBufferMemory2,
    .BindBufferMemory2KHR = lvp_BindBufferMemory2KHR,
    .BindImageMemory2 = lvp_BindImageMemory2,
    .BindImageMemory2KHR = lvp_BindImageMemory2KHR,
    .CmdSetDeviceMask = lvp_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = lvp_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = lvp_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = lvp_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = lvp_AcquireNextImage2KHR,
    .CmdDispatchBase = lvp_CmdDispatchBase,
    .CmdDispatchBaseKHR = lvp_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = lvp_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = lvp_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = lvp_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = lvp_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = lvp_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = lvp_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = lvp_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = lvp_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = lvp_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = lvp_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = lvp_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = lvp_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = lvp_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = lvp_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = lvp_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = lvp_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = lvp_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = lvp_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = lvp_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = lvp_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = lvp_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = lvp_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = lvp_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = lvp_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = lvp_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = lvp_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = lvp_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = lvp_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = lvp_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = lvp_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = lvp_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = lvp_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = lvp_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = lvp_GetDeviceQueue2,
    .CreateValidationCacheEXT = lvp_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = lvp_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = lvp_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = lvp_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = lvp_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = lvp_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = lvp_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = lvp_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = lvp_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = lvp_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = lvp_GetShaderInfoAMD,
    .SetLocalDimmingAMD = lvp_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = lvp_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = lvp_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = lvp_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = lvp_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = lvp_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = lvp_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = lvp_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = lvp_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = lvp_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = lvp_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = lvp_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = lvp_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = lvp_CreateRenderPass2,
    .CreateRenderPass2KHR = lvp_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = lvp_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = lvp_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = lvp_CmdNextSubpass2,
    .CmdNextSubpass2KHR = lvp_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = lvp_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = lvp_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = lvp_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = lvp_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = lvp_WaitSemaphores,
    .WaitSemaphoresKHR = lvp_WaitSemaphoresKHR,
    .SignalSemaphore = lvp_SignalSemaphore,
    .SignalSemaphoreKHR = lvp_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = lvp_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = lvp_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = lvp_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = lvp_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = lvp_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = lvp_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = lvp_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = lvp_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = lvp_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = lvp_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = lvp_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = lvp_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = lvp_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = lvp_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = lvp_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = lvp_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = lvp_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = lvp_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = lvp_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = lvp_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = lvp_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = lvp_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = lvp_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = lvp_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = lvp_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = lvp_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = lvp_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = lvp_CompileDeferredNV,
    .CreateAccelerationStructureNV = lvp_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = lvp_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = lvp_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = lvp_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = lvp_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = lvp_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = lvp_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = lvp_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = lvp_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = lvp_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = lvp_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = lvp_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = lvp_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = lvp_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = lvp_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = lvp_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = lvp_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = lvp_CmdTraceRaysKHR,
    .CmdTraceRaysNV = lvp_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = lvp_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = lvp_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = lvp_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = lvp_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = lvp_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = lvp_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = lvp_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = lvp_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = lvp_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = lvp_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = lvp_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = lvp_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = lvp_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = lvp_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = lvp_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = lvp_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = lvp_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = lvp_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = lvp_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = lvp_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = lvp_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = lvp_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = lvp_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = lvp_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = lvp_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = lvp_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = lvp_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = lvp_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = lvp_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = lvp_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = lvp_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = lvp_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = lvp_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = lvp_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = lvp_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = lvp_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = lvp_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = lvp_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = lvp_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = lvp_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = lvp_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = lvp_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = lvp_CmdSetLineStipple,
    .CmdSetLineStippleKHR = lvp_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = lvp_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = lvp_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = lvp_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = lvp_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = lvp_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = lvp_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = lvp_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = lvp_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = lvp_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = lvp_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = lvp_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = lvp_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = lvp_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = lvp_AntiLagUpdateAMD,
    .CmdSetCullMode = lvp_CmdSetCullMode,
    .CmdSetCullModeEXT = lvp_CmdSetCullModeEXT,
    .CmdSetFrontFace = lvp_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = lvp_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = lvp_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = lvp_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = lvp_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = lvp_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = lvp_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = lvp_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = lvp_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = lvp_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = lvp_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = lvp_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = lvp_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = lvp_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = lvp_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = lvp_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = lvp_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = lvp_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = lvp_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = lvp_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = lvp_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = lvp_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = lvp_CmdSetStencilOp,
    .CmdSetStencilOpEXT = lvp_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = lvp_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = lvp_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = lvp_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = lvp_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = lvp_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = lvp_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = lvp_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = lvp_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = lvp_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = lvp_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = lvp_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = lvp_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = lvp_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = lvp_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = lvp_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = lvp_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = lvp_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = lvp_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = lvp_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = lvp_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = lvp_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = lvp_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = lvp_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = lvp_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = lvp_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = lvp_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = lvp_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = lvp_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = lvp_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = lvp_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = lvp_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = lvp_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = lvp_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = lvp_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = lvp_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = lvp_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = lvp_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = lvp_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = lvp_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = lvp_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = lvp_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = lvp_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = lvp_DestroyPrivateDataSlotEXT,
    .SetPrivateData = lvp_SetPrivateData,
    .SetPrivateDataEXT = lvp_SetPrivateDataEXT,
    .GetPrivateData = lvp_GetPrivateData,
    .GetPrivateDataEXT = lvp_GetPrivateDataEXT,
    .CmdCopyBuffer2 = lvp_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = lvp_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = lvp_CmdCopyImage2,
    .CmdCopyImage2KHR = lvp_CmdCopyImage2KHR,
    .CmdBlitImage2 = lvp_CmdBlitImage2,
    .CmdBlitImage2KHR = lvp_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = lvp_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = lvp_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = lvp_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = lvp_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = lvp_CmdResolveImage2,
    .CmdResolveImage2KHR = lvp_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = lvp_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = lvp_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = lvp_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = lvp_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = lvp_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = lvp_CmdSetEvent2,
    .CmdSetEvent2KHR = lvp_CmdSetEvent2KHR,
    .CmdResetEvent2 = lvp_CmdResetEvent2,
    .CmdResetEvent2KHR = lvp_CmdResetEvent2KHR,
    .CmdWaitEvents2 = lvp_CmdWaitEvents2,
    .CmdWaitEvents2KHR = lvp_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = lvp_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = lvp_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = lvp_QueueSubmit2,
    .QueueSubmit2KHR = lvp_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = lvp_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = lvp_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = lvp_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = lvp_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = lvp_CopyMemoryToImage,
    .CopyMemoryToImageEXT = lvp_CopyMemoryToImageEXT,
    .CopyImageToMemory = lvp_CopyImageToMemory,
    .CopyImageToMemoryEXT = lvp_CopyImageToMemoryEXT,
    .CopyImageToImage = lvp_CopyImageToImage,
    .CopyImageToImageEXT = lvp_CopyImageToImageEXT,
    .TransitionImageLayout = lvp_TransitionImageLayout,
    .TransitionImageLayoutEXT = lvp_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = lvp_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = lvp_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = lvp_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = lvp_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = lvp_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = lvp_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = lvp_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = lvp_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = lvp_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = lvp_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = lvp_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = lvp_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = lvp_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = lvp_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = lvp_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = lvp_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = lvp_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = lvp_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = lvp_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = lvp_CreateCuModuleNVX,
    .CreateCuFunctionNVX = lvp_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = lvp_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = lvp_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = lvp_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = lvp_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = lvp_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = lvp_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = lvp_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = lvp_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = lvp_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = lvp_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = lvp_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = lvp_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = lvp_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = lvp_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = lvp_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = lvp_WaitForPresent2KHR,
    .WaitForPresentKHR = lvp_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = lvp_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = lvp_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = lvp_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = lvp_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = lvp_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = lvp_CmdBeginRendering,
    .CmdBeginRenderingKHR = lvp_CmdBeginRenderingKHR,
    .CmdEndRendering = lvp_CmdEndRendering,
    .CmdEndRendering2KHR = lvp_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = lvp_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = lvp_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = lvp_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = lvp_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = lvp_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = lvp_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = lvp_BuildMicromapsEXT,
    .DestroyMicromapEXT = lvp_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = lvp_CmdCopyMicromapEXT,
    .CopyMicromapEXT = lvp_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = lvp_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = lvp_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = lvp_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = lvp_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = lvp_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = lvp_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = lvp_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = lvp_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = lvp_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = lvp_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = lvp_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = lvp_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = lvp_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = lvp_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = lvp_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = lvp_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = lvp_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = lvp_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = lvp_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = lvp_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = lvp_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = lvp_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = lvp_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = lvp_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = lvp_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = lvp_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = lvp_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = lvp_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = lvp_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = lvp_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = lvp_MapMemory2,
    .MapMemory2KHR = lvp_MapMemory2KHR,
    .UnmapMemory2 = lvp_UnmapMemory2,
    .UnmapMemory2KHR = lvp_UnmapMemory2KHR,
    .CreateShadersEXT = lvp_CreateShadersEXT,
    .DestroyShaderEXT = lvp_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = lvp_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = lvp_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = lvp_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = lvp_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = lvp_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = lvp_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = lvp_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = lvp_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = lvp_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = lvp_CmdPushConstants2,
    .CmdPushConstants2KHR = lvp_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = lvp_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = lvp_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = lvp_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = lvp_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = lvp_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = lvp_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = lvp_SetLatencySleepModeNV,
    .LatencySleepNV = lvp_LatencySleepNV,
    .SetLatencyMarkerNV = lvp_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = lvp_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = lvp_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = lvp_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = lvp_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = lvp_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = lvp_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = lvp_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = lvp_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = lvp_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = lvp_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = lvp_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = lvp_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = lvp_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = lvp_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = lvp_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = lvp_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = lvp_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = lvp_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = lvp_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = lvp_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = lvp_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = lvp_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = lvp_CreateTensorARM,
    .DestroyTensorARM = lvp_DestroyTensorARM,
    .CreateTensorViewARM = lvp_CreateTensorViewARM,
    .DestroyTensorViewARM = lvp_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = lvp_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = lvp_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = lvp_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = lvp_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = lvp_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = lvp_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = lvp_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = lvp_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = lvp_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = lvp_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = lvp_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = lvp_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = lvp_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = lvp_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = lvp_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = lvp_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = lvp_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = lvp_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = lvp_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = lvp_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = lvp_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = lvp_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = lvp_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = lvp_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = lvp_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = lvp_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = lvp_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = lvp_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = lvp_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = lvp_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = lvp_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = lvp_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = lvp_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = lvp_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = lvp_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = lvp_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = lvp_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = lvp_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = lvp_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = lvp_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = lvp_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = lvp_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = lvp_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = lvp_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = lvp_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = lvp_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = lvp_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = lvp_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = lvp_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = lvp_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = lvp_CmdSetDispatchParametersARM,
};


