/*
 * Copyright (c) 2018 Valve Corporation
 *
 * SPDX-License-Identifier: MIT
 *
 * This file was generated by aco_opcodes_cpp.py
 */

#include "aco_ir.h"

namespace aco {



extern const aco::Info instr_info = {
   {
      50,
      -1,
      82,
      57,
      89,
      49,
      81,
      -1,
      61,
      93,
      62,
      94,
      64,
      96,
      63,
      95,
      60,
      92,
      58,
      90,
      -1,
      -1,
      -1,
      -1,
      55,
      87,
      53,
      85,
      51,
      83,
      48,
      80,
      56,
      88,
      54,
      86,
      59,
      91,
      -1,
      -1,
      12,
      13,
      15,
      14,
      -1,
      -1,
      -1,
      -1,
      -1,
      0,
      1,
      2,
      3,
      9,
      -1,
      -1,
      -1,
      -1,
      11,
      8,
      -1,
      -1,
      10,
      24,
      -1,
      28,
      29,
      31,
      30,
      -1,
      -1,
      -1,
      -1,
      -1,
      4,
      5,
      6,
      7,
      -1,
      26,
      -1,
      113,
      112,
      -1,
      -1,
      -1,
      32,
      96,
      -1,
      128,
      192,
      0,
      64,
      9,
      73,
      41,
      105,
      137,
      201,
      62,
      -1,
      -1,
      -1,
      -1,
      16,
      80,
      17,
      81,
      48,
      112,
      49,
      113,
      -1,
      -1,
      253,
      126,
      61,
      36,
      100,
      132,
      196,
      4,
      68,
      29,
      25,
      27,
      28,
      24,
      26,
      35,
      99,
      131,
      195,
      3,
      67,
      19,
      83,
      6,
      70,
      51,
      115,
      38,
      102,
      40,
      104,
      147,
      211,
      134,
      198,
      136,
      200,
      8,
      72,
      18,
      82,
      5,
      69,
      50,
      114,
      37,
      101,
      39,
      103,
      146,
      210,
      133,
      197,
      135,
      199,
      7,
      71,
      12,
      76,
      44,
      108,
      20,
      10,
      74,
      42,
      106,
      138,
      202,
      63,
      -1,
      -1,
      -1,
      -1,
      -1,
      55,
      119,
      56,
      120,
      -1,
      255,
      54,
      118,
      254,
      59,
      57,
      -1,
      -1,
      60,
      -1,
      -1,
      58,
      -1,
      -1,
      34,
      98,
      130,
      194,
      2,
      66,
      -1,
      -1,
      -1,
      33,
      97,
      129,
      193,
      1,
      65,
      53,
      52,
      14,
      78,
      15,
      79,
      -1,
      223,
      31,
      -1,
      13,
      77,
      30,
      -1,
      222,
      141,
      205,
      46,
      110,
      47,
      111,
      45,
      109,
      11,
      75,
      43,
      107,
      139,
      203,
      0,
      50,
      -1,
      82,
      57,
      89,
      49,
      81,
      -1,
      -1,
      61,
      93,
      62,
      94,
      64,
      96,
      63,
      95,
      60,
      92,
      58,
      90,
      -1,
      -1,
      55,
      87,
      53,
      85,
      51,
      83,
      48,
      80,
      56,
      88,
      54,
      86,
      59,
      91,
      12,
      13,
      15,
      14,
      9,
      -1,
      -1,
      -1,
      -1,
      11,
      8,
      -1,
      -1,
      10,
      24,
      -1,
      28,
      29,
      31,
      30,
      26,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      17,
      -1,
      24,
      16,
      28,
      29,
      31,
      30,
      27,
      25,
      -1,
      -1,
      -1,
      22,
      20,
      18,
      15,
      23,
      21,
      26,
      -1,
      -1,
      -1,
      -1,
      64,
      69,
      70,
      86,
      85,
      72,
      77,
      78,
      94,
      93,
      73,
      89,
      76,
      92,
      79,
      95,
      88,
      65,
      81,
      68,
      84,
      71,
      87,
      80,
      96,
      14,
      0,
      1,
      4,
      5,
      2,
      3,
      -1,
      32,
      37,
      38,
      54,
      53,
      40,
      45,
      46,
      62,
      61,
      106,
      107,
      111,
      110,
      41,
      57,
      42,
      43,
      -1,
      59,
      -1,
      -1,
      58,
      -1,
      44,
      60,
      47,
      63,
      56,
      104,
      105,
      109,
      108,
      33,
      49,
      34,
      35,
      -1,
      51,
      -1,
      -1,
      50,
      -1,
      36,
      52,
      39,
      55,
      48,
      8,
      9,
      11,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      52,
      44,
      -1,
      -1,
      2,
      0,
      -1,
      4,
      15,
      14,
      15,
      -1,
      36,
      -1,
      -1,
      -1,
      -1,
      20,
      21,
      -1,
      39,
      -1,
      -1,
      34,
      35,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      13,
      14,
      15,
      16,
      40,
      42,
      39,
      41,
      36,
      37,
      12,
      14,
      13,
      15,
      -1,
      27,
      28,
      29,
      30,
      2,
      11,
      12,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      8,
      12,
      9,
      -1,
      10,
      11,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      23,
      26,
      25,
      24,
      9,
      8,
      43,
      17,
      50,
      4,
      5,
      7,
      6,
      -1,
      -1,
      -1,
      5,
      6,
      2,
      -1,
      -1,
      0,
      6,
      -1,
      -1,
      -1,
      3,
      9,
      -1,
      -1,
      2,
      8,
      -1,
      -1,
      5,
      11,
      -1,
      -1,
      1,
      7,
      -1,
      -1,
      -1,
      4,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      3,
      9,
      6,
      12,
      5,
      11,
      8,
      14,
      4,
      10,
      7,
      13,
      -1,
      10,
      11,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      31,
      29,
      -1,
      -1,
      21,
      -1,
      -1,
      1,
      -1,
      -1,
      17,
      18,
      19,
      20,
      23,
      21,
      22,
      24,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      31,
      18,
      -1,
      19,
      20,
      -1,
      0,
      4,
      1,
      -1,
      2,
      3,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      30,
      31,
      32,
      33,
      -1,
      -1,
      8,
      9,
      -1,
      -1,
      -1,
      30,
      -1,
      -1,
      6,
      7,
      -1,
      -1,
      3,
      4,
      53,
      0,
      48,
      49,
      46,
      47,
      -1,
      -1,
      -1,
      -1,
      -1,
      38,
      -1,
      16,
      24,
      25,
      -1,
      41,
      0,
      26,
      27,
      -1,
      42,
      7,
      8,
      16,
      17,
      -1,
      37,
      -1,
      -1,
      22,
      23,
      -1,
      40,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      44,
      45,
      34,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      16,
      -1,
      -1,
      17,
      -1,
      -1,
      -1,
      -1,
      13,
      11,
      32,
      15,
      19,
      21,
      16,
      26,
      25,
      -1,
      14,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      3,
      1,
      -1,
      5,
      -1,
      -1,
      33,
      18,
      -1,
      -1,
      22,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      12,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      9,
      10,
      28,
      29,
      -1,
      43,
      18,
      19,
      -1,
      38,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      0,
      1,
      2,
      3,
      -1,
      -1,
      -1,
      -1,
      4,
      5,
      6,
      7,
      -1,
      37,
      -1,
      -1,
      3,
      -1,
      356,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      40,
      334,
      335,
      -1,
      27,
      -1,
      23,
      355,
      -1,
      -1,
      24,
      -1,
      290,
      329,
      328,
      330,
      286,
      56,
      -1,
      34,
      24,
      65,
      -1,
      136,
      168,
      -1,
      2,
      34,
      -1,
      130,
      162,
      -1,
      194,
      226,
      -1,
      0,
      32,
      -1,
      128,
      160,
      -1,
      192,
      224,
      -1,
      6,
      38,
      -1,
      134,
      166,
      -1,
      198,
      230,
      -1,
      4,
      36,
      -1,
      132,
      164,
      -1,
      196,
      228,
      -1,
      3,
      35,
      -1,
      131,
      163,
      -1,
      195,
      227,
      -1,
      5,
      37,
      -1,
      133,
      165,
      -1,
      197,
      229,
      -1,
      1,
      33,
      -1,
      129,
      161,
      -1,
      193,
      225,
      -1,
      13,
      45,
      -1,
      9,
      41,
      -1,
      11,
      43,
      -1,
      12,
      44,
      -1,
      10,
      42,
      -1,
      14,
      46,
      -1,
      7,
      39,
      -1,
      15,
      47,
      -1,
      135,
      167,
      -1,
      199,
      231,
      -1,
      8,
      40,
      -1,
      152,
      184,
      -1,
      18,
      50,
      -1,
      146,
      178,
      -1,
      210,
      242,
      -1,
      16,
      48,
      -1,
      144,
      176,
      -1,
      208,
      240,
      -1,
      22,
      54,
      -1,
      150,
      182,
      -1,
      214,
      246,
      -1,
      20,
      52,
      -1,
      148,
      180,
      -1,
      212,
      244,
      -1,
      19,
      51,
      -1,
      147,
      179,
      -1,
      211,
      243,
      -1,
      21,
      53,
      -1,
      149,
      181,
      -1,
      213,
      245,
      -1,
      17,
      49,
      -1,
      145,
      177,
      -1,
      209,
      241,
      -1,
      29,
      61,
      -1,
      25,
      57,
      -1,
      27,
      59,
      -1,
      28,
      60,
      -1,
      26,
      58,
      -1,
      30,
      62,
      -1,
      23,
      55,
      -1,
      31,
      63,
      -1,
      151,
      183,
      -1,
      215,
      247,
      -1,
      24,
      56,
      -1,
      0,
      -1,
      54,
      324,
      327,
      325,
      326,
      10,
      -1,
      -1,
      -1,
      11,
      15,
      -1,
      5,
      6,
      17,
      18,
      19,
      20,
      16,
      4,
      22,
      13,
      -1,
      8,
      3,
      -1,
      -1,
      -1,
      14,
      -1,
      -1,
      -1,
      -1,
      -1,
      305,
      -1,
      304,
      350,
      300,
      -1,
      301,
      -1,
      302,
      47,
      -1,
      12,
      -1,
      -1,
      -1,
      7,
      21,
      -1,
      -1,
      351,
      352,
      -1,
      367,
      368,
      365,
      366,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      37,
      70,
      59,
      57,
      58,
      -1,
      36,
      26,
      -1,
      331,
      332,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      32,
      62,
      -1,
      63,
      60,
      -1,
      64,
      61,
      2,
      -1,
      -1,
      -1,
      0,
      -1,
      -1,
      -1,
      -1,
      1,
      -1,
      -1,
      -1,
      -1,
      -1,
      299,
      360,
      333,
      38,
      -1,
      39,
      69,
      -1,
      25,
      353,
      -1,
      -1,
      -1,
      26,
      -1,
      -1,
      21,
      354,
      -1,
      -1,
      22,
      -1,
      -1,
      31,
      6,
      -1,
      321,
      -1,
      -1,
      322,
      375,
      -1,
      320,
      -1,
      -1,
      -1,
      -1,
      323,
      374,
      -1,
      33,
      -1,
      32,
      -1,
      340,
      -1,
      341,
      -1,
      342,
      -1,
      16,
      -1,
      359,
      -1,
      -1,
      18,
      14,
      -1,
      -1,
      20,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      36,
      -1,
      291,
      -1,
      343,
      -1,
      344,
      -1,
      345,
      -1,
      337,
      -1,
      338,
      -1,
      339,
      -1,
      15,
      -1,
      358,
      -1,
      -1,
      17,
      13,
      -1,
      -1,
      19,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      1,
      66,
      67,
      -1,
      68,
      371,
      373,
      369,
      -1,
      8,
      -1,
      357,
      364,
      10,
      362,
      12,
      9,
      7,
      363,
      -1,
      -1,
      361,
      11,
      336,
      0,
      -1,
      55,
      -1,
      -1,
      28,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      370,
      40,
      48,
      -1,
      42,
      47,
      43,
      41,
      2,
      1,
      -1,
      -1,
      35,
      25,
      44,
      50,
      -1,
      46,
      49,
      45,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      347,
      348,
      349,
      346,
      -1,
      -1,
      -1,
      53,
      -1,
      51,
      52,
      38,
      -1,
      -1,
      4,
      -1,
      -1,
      -1,
      -1,
      -1,
      41,
      42,
      39,
      -1,
      -1,
      5,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      372,
      -1,
      33,
      23,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      2,
      -1,
      -1,
      -1,
      -1,
      -1,
      29,
   },
   {
      66,
      -1,
      98,
      72,
      104,
      65,
      97,
      -1,
      76,
      108,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      75,
      107,
      73,
      105,
      -1,
      -1,
      -1,
      -1,
      70,
      102,
      68,
      100,
      67,
      99,
      64,
      96,
      71,
      103,
      69,
      101,
      74,
      106,
      -1,
      -1,
      20,
      21,
      22,
      23,
      38,
      8,
      9,
      10,
      11,
      0,
      1,
      2,
      3,
      17,
      34,
      35,
      36,
      37,
      19,
      16,
      32,
      33,
      18,
      24,
      25,
      28,
      29,
      30,
      31,
      39,
      12,
      13,
      14,
      15,
      4,
      5,
      6,
      7,
      61,
      26,
      27,
      62,
      63,
      21,
      -1,
      53,
      32,
      96,
      149,
      128,
      192,
      0,
      64,
      9,
      73,
      41,
      105,
      137,
      201,
      190,
      63,
      -1,
      -1,
      -1,
      16,
      80,
      17,
      81,
      48,
      112,
      49,
      113,
      -1,
      -1,
      -1,
      126,
      189,
      36,
      100,
      132,
      196,
      4,
      68,
      157,
      153,
      155,
      156,
      152,
      154,
      35,
      99,
      131,
      195,
      3,
      67,
      19,
      83,
      6,
      70,
      51,
      115,
      38,
      102,
      40,
      104,
      147,
      211,
      134,
      198,
      136,
      200,
      8,
      72,
      18,
      82,
      5,
      69,
      50,
      114,
      37,
      101,
      39,
      103,
      146,
      210,
      133,
      197,
      135,
      199,
      7,
      71,
      12,
      76,
      44,
      108,
      20,
      10,
      74,
      42,
      106,
      138,
      202,
      191,
      62,
      -1,
      -1,
      -1,
      -1,
      55,
      119,
      56,
      120,
      182,
      255,
      54,
      118,
      254,
      59,
      57,
      88,
      89,
      60,
      90,
      91,
      58,
      86,
      87,
      34,
      98,
      130,
      194,
      2,
      66,
      -1,
      -1,
      -1,
      33,
      97,
      129,
      193,
      1,
      65,
      61,
      52,
      14,
      78,
      15,
      79,
      29,
      223,
      31,
      85,
      13,
      77,
      30,
      84,
      222,
      141,
      205,
      46,
      110,
      47,
      111,
      45,
      109,
      11,
      75,
      43,
      107,
      139,
      203,
      0,
      66,
      -1,
      98,
      72,
      104,
      65,
      97,
      -1,
      -1,
      76,
      108,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      75,
      107,
      73,
      105,
      -1,
      -1,
      70,
      102,
      68,
      100,
      67,
      99,
      64,
      96,
      71,
      103,
      69,
      101,
      74,
      106,
      20,
      21,
      22,
      23,
      17,
      34,
      35,
      36,
      37,
      19,
      16,
      32,
      33,
      18,
      24,
      25,
      28,
      29,
      30,
      31,
      26,
      27,
      66,
      -1,
      98,
      72,
      104,
      65,
      97,
      -1,
      -1,
      76,
      108,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      75,
      107,
      73,
      105,
      -1,
      -1,
      -1,
      70,
      102,
      68,
      100,
      67,
      99,
      64,
      96,
      71,
      103,
      69,
      101,
      74,
      106,
      -1,
      20,
      -1,
      21,
      22,
      23,
      17,
      34,
      35,
      36,
      37,
      19,
      -1,
      -1,
      16,
      32,
      33,
      18,
      24,
      25,
      28,
      -1,
      29,
      30,
      31,
      26,
      27,
      -1,
      -1,
      18,
      -1,
      24,
      17,
      28,
      -1,
      -1,
      -1,
      27,
      25,
      -1,
      -1,
      -1,
      22,
      20,
      19,
      16,
      23,
      21,
      26,
      -1,
      -1,
      -1,
      -1,
      64,
      69,
      70,
      86,
      85,
      72,
      77,
      78,
      94,
      93,
      73,
      89,
      76,
      92,
      79,
      95,
      88,
      65,
      81,
      68,
      84,
      71,
      87,
      80,
      96,
      14,
      0,
      1,
      4,
      5,
      2,
      3,
      -1,
      32,
      37,
      38,
      54,
      53,
      40,
      45,
      46,
      62,
      61,
      106,
      107,
      111,
      110,
      41,
      57,
      42,
      43,
      -1,
      59,
      -1,
      -1,
      58,
      -1,
      44,
      60,
      47,
      63,
      56,
      104,
      105,
      109,
      108,
      33,
      49,
      34,
      35,
      -1,
      51,
      -1,
      -1,
      50,
      -1,
      36,
      52,
      39,
      55,
      48,
      8,
      9,
      11,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      48,
      42,
      -1,
      -1,
      2,
      0,
      -1,
      4,
      14,
      12,
      13,
      -1,
      32,
      -1,
      51,
      -1,
      53,
      18,
      19,
      -1,
      35,
      -1,
      54,
      32,
      33,
      38,
      39,
      130,
      162,
      136,
      168,
      129,
      161,
      140,
      172,
      139,
      171,
      137,
      169,
      134,
      166,
      132,
      164,
      131,
      163,
      128,
      160,
      135,
      167,
      133,
      165,
      138,
      170,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      10,
      11,
      12,
      13,
      38,
      40,
      37,
      39,
      34,
      35,
      12,
      14,
      13,
      15,
      55,
      24,
      25,
      26,
      27,
      2,
      8,
      9,
      66,
      98,
      72,
      104,
      65,
      97,
      76,
      108,
      75,
      107,
      73,
      105,
      70,
      102,
      68,
      100,
      67,
      99,
      64,
      96,
      71,
      103,
      69,
      101,
      74,
      106,
      8,
      12,
      9,
      -1,
      10,
      11,
      -1,
      -1,
      -1,
      -1,
      -1,
      24,
      25,
      26,
      21,
      23,
      26,
      25,
      24,
      9,
      8,
      41,
      16,
      46,
      4,
      5,
      7,
      6,
      -1,
      -1,
      -1,
      2,
      3,
      1,
      -1,
      -1,
      0,
      6,
      18,
      -1,
      -1,
      3,
      9,
      -1,
      -1,
      2,
      8,
      -1,
      -1,
      5,
      11,
      -1,
      -1,
      1,
      7,
      19,
      -1,
      -1,
      4,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      2,
      8,
      5,
      11,
      4,
      10,
      7,
      13,
      3,
      9,
      6,
      12,
      -1,
      10,
      11,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      40,
      41,
      32,
      34,
      33,
      35,
      21,
      -1,
      -1,
      1,
      30,
      27,
      14,
      15,
      16,
      17,
      20,
      18,
      19,
      21,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      28,
      17,
      31,
      19,
      20,
      -1,
      0,
      4,
      1,
      -1,
      2,
      3,
      -1,
      -1,
      -1,
      -1,
      46,
      47,
      48,
      49,
      28,
      29,
      30,
      31,
      -1,
      -1,
      8,
      9,
      -1,
      -1,
      37,
      36,
      -1,
      -1,
      6,
      7,
      -1,
      -1,
      0,
      1,
      -1,
      0,
      44,
      45,
      42,
      43,
      -1,
      -1,
      -1,
      45,
      44,
      36,
      -1,
      15,
      22,
      23,
      -1,
      37,
      0,
      24,
      25,
      -1,
      38,
      4,
      5,
      14,
      15,
      -1,
      33,
      -1,
      52,
      20,
      21,
      -1,
      36,
      52,
      -1,
      51,
      50,
      -1,
      -1,
      -1,
      -1,
      40,
      41,
      31,
      43,
      -1,
      -1,
      -1,
      5,
      6,
      7,
      21,
      22,
      23,
      16,
      -1,
      -1,
      17,
      50,
      29,
      28,
      17,
      13,
      11,
      29,
      15,
      18,
      20,
      16,
      23,
      22,
      -1,
      14,
      -1,
      16,
      17,
      18,
      -1,
      -1,
      3,
      1,
      -1,
      5,
      -1,
      -1,
      30,
      18,
      -1,
      -1,
      22,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      12,
      -1,
      -1,
      -1,
      -1,
      -1,
      3,
      -1,
      6,
      7,
      26,
      27,
      -1,
      39,
      16,
      17,
      -1,
      34,
      20,
      21,
      22,
      23,
      17,
      34,
      35,
      36,
      37,
      19,
      16,
      32,
      33,
      18,
      24,
      25,
      28,
      29,
      30,
      31,
      26,
      27,
      8,
      9,
      10,
      11,
      0,
      1,
      2,
      3,
      12,
      13,
      14,
      15,
      4,
      5,
      6,
      7,
      511,
      25,
      -1,
      31,
      1,
      -1,
      640,
      670,
      668,
      510,
      38,
      -1,
      52,
      28,
      462,
      463,
      -1,
      19,
      513,
      -1,
      -1,
      44,
      -1,
      17,
      657,
      651,
      457,
      456,
      458,
      659,
      44,
      69,
      29,
      24,
      53,
      20,
      16,
      18,
      34,
      66,
      98,
      162,
      194,
      226,
      170,
      202,
      234,
      32,
      64,
      96,
      160,
      192,
      224,
      168,
      200,
      232,
      38,
      70,
      102,
      166,
      198,
      230,
      174,
      206,
      238,
      36,
      68,
      100,
      164,
      196,
      228,
      172,
      204,
      236,
      35,
      67,
      99,
      163,
      195,
      227,
      171,
      203,
      235,
      37,
      69,
      101,
      165,
      197,
      229,
      173,
      205,
      237,
      33,
      65,
      97,
      161,
      193,
      225,
      169,
      201,
      233,
      45,
      77,
      109,
      41,
      73,
      105,
      43,
      75,
      107,
      44,
      76,
      108,
      42,
      74,
      106,
      46,
      78,
      110,
      39,
      71,
      103,
      47,
      79,
      111,
      167,
      199,
      231,
      175,
      207,
      239,
      40,
      72,
      104,
      21,
      17,
      19,
      50,
      82,
      114,
      178,
      210,
      242,
      186,
      218,
      250,
      48,
      80,
      112,
      176,
      208,
      240,
      184,
      216,
      248,
      54,
      86,
      118,
      182,
      214,
      246,
      190,
      222,
      254,
      52,
      84,
      116,
      180,
      212,
      244,
      188,
      220,
      252,
      51,
      83,
      115,
      179,
      211,
      243,
      187,
      219,
      251,
      53,
      85,
      117,
      181,
      213,
      245,
      189,
      221,
      253,
      49,
      81,
      113,
      177,
      209,
      241,
      185,
      217,
      249,
      61,
      93,
      125,
      57,
      89,
      121,
      59,
      91,
      123,
      60,
      92,
      124,
      58,
      90,
      122,
      62,
      94,
      126,
      55,
      87,
      119,
      63,
      95,
      127,
      183,
      215,
      247,
      191,
      223,
      255,
      56,
      88,
      120,
      -1,
      0,
      74,
      42,
      452,
      455,
      453,
      454,
      10,
      58,
      57,
      -1,
      11,
      15,
      -1,
      5,
      6,
      17,
      18,
      19,
      20,
      16,
      4,
      22,
      13,
      60,
      8,
      3,
      -1,
      77,
      78,
      14,
      -1,
      -1,
      -1,
      -1,
      -1,
      664,
      -1,
      663,
      477,
      496,
      665,
      660,
      666,
      661,
      -1,
      662,
      12,
      -1,
      -1,
      59,
      7,
      21,
      -1,
      519,
      478,
      479,
      495,
      482,
      483,
      480,
      481,
      -1,
      -1,
      -1,
      35,
      38,
      39,
      55,
      -1,
      -1,
      -1,
      -1,
      40,
      -1,
      41,
      57,
      42,
      -1,
      43,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      65,
      32,
      75,
      47,
      45,
      46,
      68,
      31,
      26,
      518,
      459,
      460,
      494,
      -1,
      32,
      34,
      33,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      72,
      27,
      50,
      67,
      51,
      48,
      66,
      52,
      49,
      2,
      -1,
      -1,
      -1,
      0,
      628,
      629,
      631,
      -1,
      1,
      -1,
      631,
      630,
      -1,
      51,
      648,
      644,
      461,
      -1,
      64,
      33,
      76,
      509,
      -1,
      -1,
      512,
      42,
      -1,
      18,
      -1,
      655,
      -1,
      -1,
      43,
      -1,
      16,
      656,
      35,
      22,
      -1,
      515,
      449,
      517,
      498,
      450,
      489,
      490,
      448,
      492,
      491,
      516,
      497,
      451,
      488,
      37,
      24,
      36,
      23,
      503,
      467,
      504,
      468,
      505,
      469,
      45,
      11,
      -1,
      643,
      48,
      -1,
      13,
      -1,
      47,
      -1,
      15,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      653,
      652,
      506,
      470,
      507,
      471,
      508,
      472,
      500,
      464,
      501,
      465,
      502,
      466,
      46,
      10,
      -1,
      642,
      50,
      -1,
      12,
      -1,
      49,
      -1,
      14,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      1,
      -1,
      -1,
      -1,
      -1,
      486,
      487,
      484,
      34,
      5,
      -1,
      641,
      647,
      7,
      646,
      9,
      6,
      4,
      645,
      41,
      -1,
      645,
      8,
      -1,
      0,
      -1,
      43,
      514,
      -1,
      20,
      672,
      493,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      15,
      2,
      10,
      6,
      14,
      -1,
      4,
      5,
      0,
      9,
      18,
      7,
      12,
      -1,
      17,
      8,
      13,
      -1,
      16,
      1,
      3,
      11,
      485,
      -1,
      -1,
      61,
      34,
      37,
      35,
      -1,
      2,
      -1,
      649,
      71,
      30,
      25,
      -1,
      -1,
      63,
      36,
      38,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      474,
      475,
      476,
      473,
      79,
      55,
      73,
      41,
      62,
      39,
      40,
      26,
      -1,
      32,
      2,
      671,
      669,
      39,
      -1,
      53,
      29,
      30,
      27,
      -1,
      33,
      3,
      40,
      54,
      -1,
      81,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      658,
      70,
      28,
      23,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      650,
      499,
      -1,
      -1,
      -1,
      21,
   },
   {
      50,
      -1,
      82,
      57,
      89,
      49,
      81,
      52,
      61,
      93,
      62,
      94,
      64,
      96,
      63,
      95,
      60,
      92,
      58,
      90,
      -1,
      -1,
      -1,
      -1,
      55,
      87,
      53,
      85,
      51,
      83,
      48,
      80,
      56,
      88,
      54,
      86,
      59,
      91,
      113,
      114,
      12,
      13,
      15,
      14,
      38,
      128,
      129,
      130,
      131,
      0,
      1,
      2,
      3,
      9,
      34,
      35,
      36,
      37,
      11,
      8,
      32,
      33,
      10,
      24,
      25,
      28,
      29,
      31,
      30,
      39,
      132,
      133,
      134,
      135,
      4,
      5,
      6,
      7,
      -1,
      26,
      27,
      -1,
      -1,
      21,
      -1,
      85,
      32,
      96,
      149,
      128,
      192,
      0,
      64,
      9,
      73,
      41,
      105,
      137,
      201,
      62,
      179,
      -1,
      -1,
      -1,
      16,
      80,
      17,
      81,
      48,
      112,
      49,
      113,
      -1,
      -1,
      -1,
      126,
      61,
      36,
      100,
      132,
      196,
      4,
      68,
      29,
      25,
      27,
      28,
      24,
      26,
      35,
      99,
      131,
      195,
      3,
      67,
      19,
      83,
      6,
      70,
      51,
      115,
      38,
      102,
      40,
      104,
      147,
      211,
      134,
      198,
      136,
      200,
      8,
      72,
      18,
      82,
      5,
      69,
      50,
      114,
      37,
      101,
      39,
      103,
      146,
      210,
      133,
      197,
      135,
      199,
      7,
      71,
      12,
      76,
      44,
      108,
      20,
      10,
      74,
      42,
      106,
      138,
      202,
      63,
      178,
      -1,
      -1,
      -1,
      -1,
      55,
      119,
      56,
      120,
      177,
      255,
      54,
      118,
      254,
      59,
      57,
      164,
      165,
      60,
      166,
      167,
      58,
      162,
      163,
      34,
      98,
      130,
      194,
      2,
      66,
      -1,
      -1,
      -1,
      33,
      97,
      129,
      193,
      1,
      65,
      53,
      52,
      14,
      78,
      15,
      79,
      176,
      223,
      31,
      161,
      13,
      77,
      30,
      160,
      222,
      141,
      205,
      46,
      110,
      47,
      111,
      45,
      109,
      11,
      75,
      43,
      107,
      139,
      203,
      0,
      50,
      -1,
      82,
      57,
      89,
      49,
      81,
      -1,
      -1,
      61,
      93,
      62,
      94,
      64,
      96,
      63,
      95,
      60,
      92,
      58,
      90,
      -1,
      -1,
      55,
      87,
      53,
      85,
      51,
      83,
      48,
      80,
      56,
      88,
      54,
      86,
      59,
      91,
      12,
      13,
      15,
      14,
      9,
      34,
      35,
      36,
      37,
      11,
      8,
      32,
      33,
      10,
      24,
      25,
      28,
      29,
      31,
      30,
      26,
      27,
      50,
      -1,
      82,
      57,
      89,
      49,
      81,
      -1,
      52,
      61,
      93,
      62,
      94,
      64,
      96,
      63,
      95,
      60,
      92,
      58,
      90,
      -1,
      -1,
      -1,
      55,
      87,
      53,
      85,
      51,
      83,
      48,
      80,
      56,
      88,
      54,
      86,
      59,
      91,
      -1,
      12,
      22,
      13,
      15,
      14,
      9,
      34,
      35,
      36,
      37,
      11,
      -1,
      -1,
      8,
      32,
      33,
      10,
      24,
      25,
      28,
      23,
      29,
      31,
      30,
      26,
      27,
      -1,
      -1,
      17,
      -1,
      24,
      16,
      28,
      29,
      31,
      30,
      27,
      25,
      -1,
      -1,
      -1,
      22,
      20,
      18,
      15,
      23,
      21,
      26,
      231,
      -1,
      -1,
      230,
      64,
      69,
      70,
      86,
      85,
      72,
      77,
      78,
      94,
      93,
      73,
      89,
      76,
      92,
      79,
      95,
      88,
      65,
      81,
      68,
      84,
      71,
      87,
      80,
      96,
      14,
      0,
      1,
      4,
      5,
      2,
      3,
      128,
      32,
      37,
      38,
      54,
      53,
      40,
      45,
      46,
      62,
      61,
      106,
      107,
      111,
      110,
      41,
      57,
      42,
      43,
      171,
      59,
      187,
      170,
      58,
      186,
      44,
      60,
      47,
      63,
      56,
      104,
      105,
      109,
      108,
      33,
      49,
      34,
      35,
      163,
      51,
      179,
      162,
      50,
      178,
      36,
      52,
      39,
      55,
      48,
      8,
      9,
      11,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      52,
      44,
      -1,
      -1,
      2,
      0,
      -1,
      4,
      15,
      14,
      15,
      60,
      36,
      68,
      55,
      70,
      57,
      20,
      21,
      63,
      39,
      71,
      58,
      34,
      35,
      38,
      39,
      130,
      162,
      136,
      168,
      129,
      161,
      140,
      172,
      139,
      171,
      137,
      169,
      134,
      166,
      132,
      164,
      131,
      163,
      128,
      160,
      135,
      167,
      133,
      165,
      138,
      170,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      13,
      14,
      15,
      16,
      40,
      42,
      39,
      41,
      36,
      37,
      12,
      14,
      13,
      15,
      59,
      27,
      28,
      29,
      30,
      2,
      11,
      12,
      66,
      98,
      72,
      104,
      65,
      97,
      76,
      108,
      75,
      107,
      73,
      105,
      70,
      102,
      68,
      100,
      67,
      99,
      64,
      96,
      71,
      103,
      69,
      101,
      74,
      106,
      8,
      12,
      9,
      -1,
      10,
      11,
      -1,
      -1,
      -1,
      -1,
      -1,
      24,
      25,
      26,
      22,
      23,
      26,
      25,
      24,
      9,
      8,
      -1,
      -1,
      -1,
      4,
      5,
      7,
      6,
      -1,
      -1,
      33,
      5,
      6,
      2,
      -1,
      -1,
      0,
      6,
      18,
      -1,
      -1,
      3,
      9,
      -1,
      -1,
      2,
      8,
      -1,
      -1,
      5,
      11,
      -1,
      -1,
      1,
      7,
      19,
      -1,
      -1,
      4,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      3,
      9,
      6,
      12,
      5,
      11,
      8,
      14,
      4,
      10,
      7,
      13,
      31,
      10,
      11,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      40,
      41,
      32,
      -1,
      33,
      -1,
      21,
      -1,
      37,
      1,
      30,
      27,
      17,
      18,
      19,
      20,
      23,
      21,
      22,
      24,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      42,
      31,
      18,
      31,
      19,
      20,
      32,
      0,
      4,
      1,
      -1,
      2,
      3,
      -1,
      -1,
      -1,
      -1,
      46,
      47,
      48,
      49,
      30,
      31,
      32,
      33,
      -1,
      -1,
      8,
      9,
      -1,
      -1,
      37,
      36,
      -1,
      -1,
      6,
      7,
      -1,
      -1,
      3,
      4,
      53,
      0,
      48,
      49,
      46,
      47,
      73,
      -1,
      -1,
      54,
      53,
      38,
      -1,
      16,
      24,
      25,
      65,
      41,
      0,
      26,
      27,
      66,
      42,
      7,
      8,
      16,
      17,
      61,
      37,
      69,
      56,
      22,
      23,
      64,
      40,
      52,
      -1,
      51,
      50,
      -1,
      -1,
      -1,
      -1,
      44,
      45,
      34,
      -1,
      -1,
      -1,
      36,
      5,
      6,
      7,
      21,
      22,
      23,
      16,
      -1,
      -1,
      17,
      -1,
      -1,
      -1,
      -1,
      13,
      11,
      32,
      15,
      19,
      21,
      -1,
      26,
      25,
      -1,
      14,
      -1,
      16,
      17,
      18,
      -1,
      -1,
      3,
      1,
      -1,
      5,
      27,
      28,
      33,
      18,
      -1,
      -1,
      22,
      38,
      1,
      -1,
      -1,
      -1,
      -1,
      34,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      12,
      35,
      25,
      26,
      24,
      23,
      3,
      -1,
      9,
      10,
      28,
      29,
      67,
      43,
      18,
      19,
      62,
      38,
      12,
      13,
      15,
      14,
      9,
      34,
      35,
      36,
      37,
      11,
      8,
      32,
      33,
      10,
      24,
      25,
      28,
      29,
      31,
      30,
      26,
      27,
      8,
      9,
      10,
      11,
      0,
      1,
      2,
      3,
      12,
      13,
      14,
      15,
      4,
      5,
      6,
      7,
      877,
      -1,
      783,
      50,
      3,
      -1,
      356,
      781,
      895,
      839,
      -1,
      771,
      37,
      40,
      334,
      335,
      -1,
      27,
      881,
      -1,
      -1,
      -1,
      776,
      24,
      769,
      868,
      329,
      328,
      330,
      867,
      56,
      92,
      34,
      24,
      65,
      143,
      136,
      168,
      202,
      2,
      34,
      138,
      130,
      162,
      170,
      194,
      226,
      200,
      0,
      32,
      -1,
      128,
      160,
      -1,
      192,
      224,
      206,
      6,
      38,
      142,
      134,
      166,
      174,
      198,
      230,
      204,
      4,
      36,
      140,
      132,
      164,
      172,
      196,
      228,
      203,
      3,
      35,
      139,
      131,
      163,
      171,
      195,
      227,
      205,
      5,
      37,
      141,
      133,
      165,
      173,
      197,
      229,
      201,
      1,
      33,
      137,
      129,
      161,
      169,
      193,
      225,
      237,
      13,
      45,
      233,
      9,
      41,
      235,
      11,
      43,
      236,
      12,
      44,
      234,
      10,
      42,
      238,
      14,
      46,
      207,
      7,
      39,
      239,
      15,
      47,
      -1,
      135,
      167,
      -1,
      199,
      231,
      232,
      8,
      40,
      159,
      152,
      184,
      218,
      18,
      50,
      154,
      146,
      178,
      186,
      210,
      242,
      216,
      16,
      48,
      -1,
      144,
      176,
      -1,
      208,
      240,
      222,
      22,
      54,
      158,
      150,
      182,
      190,
      214,
      246,
      220,
      20,
      52,
      156,
      148,
      180,
      188,
      212,
      244,
      219,
      19,
      51,
      155,
      147,
      179,
      187,
      211,
      243,
      221,
      21,
      53,
      157,
      149,
      181,
      189,
      213,
      245,
      217,
      17,
      49,
      153,
      145,
      177,
      185,
      209,
      241,
      253,
      29,
      61,
      249,
      25,
      57,
      251,
      27,
      59,
      252,
      28,
      60,
      250,
      26,
      58,
      254,
      30,
      62,
      223,
      23,
      55,
      255,
      31,
      63,
      -1,
      151,
      183,
      -1,
      215,
      247,
      248,
      24,
      56,
      -1,
      1,
      97,
      54,
      324,
      327,
      325,
      326,
      10,
      81,
      80,
      -1,
      11,
      15,
      -1,
      5,
      6,
      17,
      18,
      19,
      20,
      16,
      4,
      22,
      13,
      83,
      8,
      3,
      -1,
      99,
      100,
      14,
      -1,
      -1,
      -1,
      -1,
      -1,
      875,
      -1,
      874,
      350,
      -1,
      786,
      872,
      787,
      873,
      47,
      -1,
      12,
      -1,
      -1,
      82,
      7,
      21,
      -1,
      863,
      351,
      352,
      -1,
      367,
      368,
      365,
      366,
      -1,
      -1,
      -1,
      19,
      20,
      21,
      2,
      -1,
      -1,
      -1,
      -1,
      22,
      -1,
      23,
      13,
      24,
      -1,
      25,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      88,
      37,
      -1,
      59,
      57,
      58,
      91,
      36,
      26,
      843,
      331,
      332,
      -1,
      320,
      32,
      34,
      33,
      56,
      45,
      54,
      43,
      6,
      55,
      44,
      95,
      32,
      62,
      90,
      63,
      60,
      89,
      64,
      61,
      2,
      -1,
      -1,
      -1,
      0,
      834,
      835,
      858,
      -1,
      1,
      -1,
      858,
      -1,
      -1,
      59,
      866,
      360,
      333,
      -1,
      87,
      39,
      -1,
      838,
      -1,
      -1,
      879,
      -1,
      788,
      26,
      -1,
      767,
      -1,
      -1,
      -1,
      775,
      22,
      768,
      -1,
      31,
      6,
      -1,
      321,
      862,
      885,
      322,
      375,
      -1,
      320,
      -1,
      -1,
      832,
      883,
      323,
      374,
      -1,
      33,
      -1,
      32,
      852,
      340,
      853,
      341,
      854,
      342,
      57,
      16,
      -1,
      359,
      -1,
      778,
      18,
      -1,
      -1,
      777,
      20,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      870,
      869,
      855,
      343,
      856,
      344,
      857,
      345,
      849,
      337,
      850,
      338,
      851,
      339,
      58,
      15,
      -1,
      358,
      -1,
      780,
      17,
      -1,
      -1,
      779,
      19,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      1,
      66,
      67,
      72,
      68,
      371,
      373,
      369,
      53,
      8,
      -1,
      357,
      364,
      10,
      362,
      12,
      9,
      7,
      363,
      -1,
      773,
      361,
      11,
      336,
      0,
      -1,
      55,
      882,
      -1,
      28,
      785,
      836,
      887,
      -1,
      -1,
      888,
      -1,
      27,
      15,
      2,
      10,
      6,
      14,
      60,
      4,
      5,
      0,
      9,
      18,
      7,
      12,
      -1,
      17,
      8,
      13,
      -1,
      16,
      1,
      3,
      11,
      370,
      -1,
      -1,
      84,
      42,
      47,
      43,
      -1,
      2,
      -1,
      864,
      94,
      35,
      25,
      -1,
      -1,
      86,
      46,
      49,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      347,
      348,
      349,
      346,
      98,
      -1,
      96,
      53,
      85,
      51,
      52,
      -1,
      784,
      51,
      4,
      782,
      886,
      -1,
      772,
      38,
      41,
      42,
      -1,
      793,
      52,
      5,
      -1,
      39,
      -1,
      101,
      104,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      372,
      93,
      33,
      23,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      865,
      837,
      30,
      376,
      -1,
      29,
   },
   {
      53,
      86,
      67,
      60,
      73,
      52,
      66,
      55,
      64,
      77,
      80,
      -1,
      82,
      -1,
      81,
      -1,
      63,
      76,
      61,
      74,
      -1,
      -1,
      -1,
      -1,
      58,
      71,
      56,
      69,
      54,
      68,
      51,
      65,
      59,
      72,
      57,
      70,
      62,
      75,
      43,
      44,
      20,
      21,
      22,
      23,
      38,
      8,
      9,
      10,
      11,
      0,
      1,
      2,
      3,
      17,
      31,
      34,
      32,
      35,
      19,
      16,
      30,
      33,
      18,
      24,
      36,
      26,
      27,
      28,
      29,
      39,
      12,
      13,
      14,
      15,
      4,
      5,
      6,
      7,
      -1,
      25,
      37,
      -1,
      -1,
      21,
      122,
      121,
      32,
      96,
      -1,
      -1,
      -1,
      0,
      64,
      9,
      73,
      41,
      105,
      -1,
      -1,
      62,
      179,
      173,
      -1,
      -1,
      16,
      80,
      17,
      81,
      48,
      112,
      49,
      113,
      -1,
      -1,
      -1,
      126,
      61,
      36,
      100,
      -1,
      -1,
      4,
      68,
      29,
      25,
      27,
      28,
      24,
      26,
      35,
      99,
      -1,
      -1,
      3,
      67,
      19,
      83,
      6,
      70,
      51,
      115,
      38,
      102,
      40,
      104,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      8,
      72,
      18,
      82,
      5,
      69,
      50,
      114,
      37,
      101,
      39,
      103,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      7,
      71,
      12,
      76,
      44,
      108,
      20,
      10,
      74,
      42,
      106,
      -1,
      -1,
      63,
      178,
      -1,
      -1,
      -1,
      -1,
      55,
      119,
      56,
      120,
      177,
      255,
      54,
      118,
      254,
      59,
      57,
      164,
      165,
      60,
      166,
      167,
      58,
      162,
      163,
      34,
      98,
      -1,
      -1,
      2,
      66,
      -1,
      -1,
      123,
      33,
      97,
      -1,
      -1,
      1,
      65,
      53,
      52,
      14,
      78,
      15,
      79,
      176,
      223,
      31,
      161,
      13,
      77,
      30,
      160,
      222,
      -1,
      -1,
      46,
      110,
      47,
      111,
      45,
      109,
      11,
      75,
      43,
      107,
      -1,
      -1,
      0,
      53,
      86,
      67,
      60,
      73,
      52,
      66,
      -1,
      -1,
      64,
      77,
      80,
      -1,
      82,
      -1,
      81,
      -1,
      63,
      76,
      61,
      74,
      -1,
      -1,
      58,
      71,
      56,
      69,
      54,
      68,
      51,
      65,
      59,
      72,
      57,
      70,
      62,
      75,
      20,
      21,
      22,
      23,
      17,
      31,
      34,
      32,
      35,
      19,
      16,
      30,
      33,
      18,
      24,
      36,
      26,
      27,
      28,
      29,
      25,
      37,
      53,
      86,
      67,
      60,
      73,
      52,
      66,
      -1,
      55,
      64,
      77,
      80,
      -1,
      82,
      -1,
      81,
      -1,
      63,
      76,
      61,
      74,
      -1,
      -1,
      -1,
      58,
      71,
      56,
      69,
      54,
      68,
      51,
      65,
      59,
      72,
      57,
      70,
      62,
      75,
      -1,
      20,
      40,
      21,
      22,
      23,
      17,
      31,
      34,
      32,
      35,
      19,
      -1,
      -1,
      16,
      30,
      33,
      18,
      24,
      36,
      26,
      41,
      27,
      28,
      29,
      25,
      37,
      -1,
      -1,
      12,
      -1,
      18,
      11,
      22,
      -1,
      -1,
      -1,
      21,
      19,
      -1,
      -1,
      -1,
      16,
      14,
      13,
      10,
      17,
      15,
      20,
      26,
      -1,
      -1,
      25,
      47,
      49,
      97,
      -1,
      -1,
      51,
      100,
      101,
      -1,
      -1,
      98,
      -1,
      99,
      -1,
      52,
      55,
      -1,
      96,
      -1,
      48,
      -1,
      50,
      54,
      53,
      56,
      23,
      0,
      1,
      4,
      5,
      2,
      3,
      24,
      27,
      30,
      66,
      72,
      40,
      32,
      35,
      69,
      75,
      45,
      -1,
      -1,
      -1,
      -1,
      67,
      73,
      33,
      68,
      84,
      74,
      86,
      58,
      43,
      60,
      34,
      44,
      36,
      46,
      42,
      -1,
      -1,
      -1,
      -1,
      64,
      70,
      28,
      65,
      95,
      71,
      85,
      57,
      38,
      59,
      29,
      39,
      31,
      41,
      37,
      6,
      7,
      9,
      8,
      1,
      0,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      21,
      6,
      73,
      64,
      2,
      0,
      -1,
      4,
      15,
      22,
      23,
      32,
      33,
      44,
      45,
      52,
      53,
      34,
      35,
      48,
      49,
      54,
      55,
      12,
      13,
      34,
      35,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      61,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      22,
      23,
      24,
      25,
      39,
      41,
      38,
      40,
      42,
      43,
      12,
      14,
      13,
      15,
      20,
      16,
      17,
      18,
      19,
      32,
      4,
      5,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      8,
      12,
      9,
      -1,
      10,
      11,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      20,
      39,
      42,
      41,
      40,
      38,
      37,
      -1,
      -1,
      -1,
      33,
      34,
      36,
      35,
      107,
      96,
      5,
      2,
      3,
      2,
      82,
      66,
      0,
      6,
      16,
      86,
      70,
      3,
      9,
      84,
      68,
      2,
      8,
      83,
      67,
      5,
      11,
      85,
      69,
      1,
      7,
      17,
      81,
      65,
      4,
      10,
      93,
      77,
      89,
      73,
      91,
      75,
      92,
      76,
      90,
      74,
      94,
      78,
      87,
      71,
      88,
      72,
      3,
      9,
      6,
      12,
      5,
      11,
      8,
      14,
      4,
      10,
      7,
      13,
      31,
      48,
      49,
      104,
      105,
      100,
      101,
      106,
      102,
      72,
      103,
      -1,
      -1,
      33,
      -1,
      -1,
      -1,
      57,
      7,
      18,
      48,
      -1,
      49,
      -1,
      -1,
      8,
      9,
      12,
      10,
      11,
      13,
      108,
      97,
      69,
      78,
      71,
      70,
      -1,
      -1,
      71,
      17,
      32,
      60,
      56,
      4,
      0,
      4,
      1,
      -1,
      2,
      3,
      -1,
      -1,
      -1,
      -1,
      14,
      15,
      16,
      17,
      8,
      9,
      10,
      11,
      76,
      67,
      20,
      21,
      -1,
      -1,
      -1,
      -1,
      75,
      66,
      18,
      19,
      -1,
      -1,
      0,
      1,
      -1,
      0,
      66,
      67,
      64,
      65,
      68,
      77,
      68,
      46,
      45,
      44,
      -1,
      16,
      28,
      29,
      38,
      39,
      0,
      30,
      31,
      40,
      41,
      30,
      31,
      24,
      25,
      34,
      35,
      46,
      47,
      36,
      37,
      50,
      51,
      52,
      53,
      51,
      50,
      -1,
      -1,
      -1,
      -1,
      26,
      27,
      74,
      -1,
      110,
      99,
      17,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      54,
      76,
      77,
      55,
      -1,
      -1,
      -1,
      -1,
      2,
      1,
      72,
      53,
      18,
      19,
      -1,
      15,
      14,
      19,
      3,
      -1,
      -1,
      -1,
      -1,
      74,
      65,
      3,
      1,
      -1,
      5,
      22,
      23,
      73,
      16,
      109,
      98,
      58,
      59,
      1,
      -1,
      -1,
      11,
      -1,
      10,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      9,
      8,
      26,
      27,
      25,
      24,
      52,
      -1,
      28,
      29,
      32,
      33,
      42,
      43,
      26,
      27,
      36,
      37,
      20,
      21,
      22,
      23,
      17,
      31,
      34,
      32,
      35,
      19,
      16,
      30,
      33,
      18,
      24,
      36,
      26,
      27,
      28,
      29,
      25,
      37,
      8,
      9,
      10,
      11,
      0,
      1,
      2,
      3,
      12,
      13,
      14,
      15,
      4,
      5,
      6,
      7,
      597,
      -1,
      768,
      50,
      3,
      -1,
      807,
      781,
      806,
      583,
      -1,
      771,
      37,
      32,
      534,
      535,
      866,
      27,
      599,
      -1,
      -1,
      -1,
      826,
      26,
      830,
      798,
      529,
      528,
      530,
      797,
      56,
      92,
      34,
      24,
      -1,
      125,
      126,
      127,
      2,
      18,
      34,
      50,
      66,
      82,
      58,
      74,
      90,
      0,
      16,
      32,
      -1,
      64,
      80,
      -1,
      72,
      88,
      6,
      22,
      38,
      54,
      70,
      86,
      62,
      78,
      94,
      4,
      20,
      36,
      52,
      68,
      84,
      60,
      76,
      92,
      3,
      19,
      35,
      51,
      67,
      83,
      59,
      75,
      91,
      5,
      21,
      37,
      53,
      69,
      85,
      61,
      77,
      93,
      1,
      17,
      33,
      49,
      65,
      81,
      57,
      73,
      89,
      13,
      29,
      45,
      9,
      25,
      41,
      11,
      27,
      43,
      12,
      28,
      44,
      10,
      26,
      42,
      14,
      30,
      46,
      7,
      23,
      39,
      15,
      31,
      47,
      -1,
      71,
      87,
      -1,
      79,
      95,
      8,
      24,
      40,
      253,
      254,
      255,
      130,
      146,
      162,
      178,
      194,
      210,
      186,
      202,
      218,
      128,
      144,
      160,
      -1,
      192,
      208,
      -1,
      200,
      216,
      134,
      150,
      166,
      182,
      198,
      214,
      190,
      206,
      222,
      132,
      148,
      164,
      180,
      196,
      212,
      188,
      204,
      220,
      131,
      147,
      163,
      179,
      195,
      211,
      187,
      203,
      219,
      133,
      149,
      165,
      181,
      197,
      213,
      189,
      205,
      221,
      129,
      145,
      161,
      177,
      193,
      209,
      185,
      201,
      217,
      141,
      157,
      173,
      137,
      153,
      169,
      139,
      155,
      171,
      140,
      156,
      172,
      138,
      154,
      170,
      142,
      158,
      174,
      135,
      151,
      167,
      143,
      159,
      175,
      -1,
      199,
      215,
      -1,
      207,
      223,
      136,
      152,
      168,
      605,
      1,
      97,
      54,
      524,
      527,
      525,
      526,
      10,
      81,
      80,
      -1,
      11,
      15,
      -1,
      5,
      6,
      17,
      18,
      19,
      20,
      16,
      4,
      22,
      13,
      83,
      8,
      3,
      106,
      99,
      100,
      14,
      -1,
      -1,
      -1,
      -1,
      774,
      804,
      775,
      803,
      550,
      -1,
      786,
      801,
      787,
      802,
      47,
      -1,
      12,
      -1,
      -1,
      82,
      7,
      21,
      107,
      596,
      551,
      552,
      -1,
      567,
      568,
      764,
      765,
      615,
      614,
      26,
      19,
      -1,
      -1,
      2,
      -1,
      -1,
      -1,
      -1,
      -1,
      22,
      23,
      -1,
      -1,
      24,
      25,
      4,
      16,
      18,
      9,
      13,
      12,
      1,
      0,
      2,
      17,
      10,
      11,
      8,
      7,
      3,
      5,
      6,
      88,
      37,
      -1,
      59,
      57,
      58,
      91,
      36,
      26,
      584,
      531,
      532,
      -1,
      521,
      32,
      34,
      33,
      56,
      45,
      54,
      43,
      6,
      55,
      44,
      95,
      32,
      62,
      90,
      63,
      60,
      89,
      64,
      61,
      -1,
      2,
      0,
      4,
      -1,
      -1,
      -1,
      -1,
      3,
      -1,
      1,
      -1,
      -1,
      5,
      59,
      796,
      811,
      533,
      -1,
      87,
      39,
      -1,
      582,
      -1,
      -1,
      598,
      -1,
      824,
      24,
      -1,
      828,
      -1,
      -1,
      -1,
      825,
      25,
      829,
      -1,
      -1,
      -1,
      -1,
      -1,
      595,
      602,
      522,
      767,
      -1,
      -1,
      -1,
      -1,
      577,
      601,
      523,
      766,
      -1,
      -1,
      -1,
      -1,
      588,
      540,
      589,
      541,
      590,
      542,
      57,
      16,
      -1,
      810,
      -1,
      778,
      18,
      -1,
      -1,
      777,
      20,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      608,
      606,
      612,
      610,
      -1,
      800,
      799,
      591,
      543,
      592,
      544,
      593,
      545,
      585,
      537,
      586,
      538,
      587,
      539,
      58,
      15,
      -1,
      809,
      -1,
      780,
      17,
      -1,
      -1,
      779,
      19,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      609,
      607,
      613,
      611,
      28,
      1,
      66,
      67,
      72,
      68,
      571,
      573,
      569,
      53,
      8,
      -1,
      808,
      814,
      10,
      813,
      12,
      9,
      7,
      812,
      -1,
      773,
      812,
      11,
      536,
      0,
      105,
      55,
      600,
      867,
      28,
      785,
      580,
      603,
      -1,
      103,
      604,
      -1,
      27,
      15,
      2,
      10,
      6,
      14,
      60,
      4,
      5,
      0,
      9,
      18,
      7,
      12,
      -1,
      17,
      8,
      13,
      -1,
      16,
      1,
      3,
      11,
      570,
      -1,
      -1,
      84,
      42,
      47,
      43,
      -1,
      2,
      -1,
      864,
      94,
      35,
      25,
      -1,
      -1,
      86,
      46,
      49,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      547,
      548,
      549,
      546,
      98,
      -1,
      96,
      53,
      85,
      51,
      52,
      -1,
      769,
      51,
      4,
      782,
      805,
      -1,
      772,
      38,
      33,
      34,
      -1,
      770,
      52,
      5,
      -1,
      39,
      102,
      101,
      104,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      815,
      93,
      33,
      23,
      67,
      66,
      65,
      -1,
      -1,
      64,
      -1,
      -1,
      69,
      68,
      -1,
      -1,
      865,
      581,
      30,
      576,
      868,
      29,
   },
   {
      53,
      86,
      67,
      60,
      73,
      52,
      66,
      55,
      64,
      77,
      -1,
      -1,
      82,
      -1,
      81,
      -1,
      63,
      76,
      61,
      74,
      90,
      89,
      -1,
      -1,
      58,
      71,
      56,
      69,
      54,
      68,
      51,
      65,
      59,
      72,
      57,
      70,
      62,
      75,
      -1,
      -1,
      20,
      21,
      22,
      23,
      38,
      8,
      9,
      10,
      11,
      0,
      1,
      2,
      3,
      17,
      31,
      34,
      32,
      35,
      19,
      16,
      30,
      33,
      18,
      24,
      36,
      26,
      27,
      28,
      29,
      39,
      12,
      13,
      14,
      15,
      4,
      5,
      6,
      7,
      -1,
      25,
      37,
      -1,
      -1,
      21,
      -1,
      121,
      32,
      96,
      -1,
      -1,
      -1,
      0,
      64,
      9,
      73,
      41,
      105,
      -1,
      -1,
      62,
      179,
      224,
      225,
      226,
      16,
      80,
      -1,
      -1,
      48,
      112,
      -1,
      -1,
      168,
      152,
      -1,
      126,
      61,
      36,
      100,
      -1,
      -1,
      4,
      68,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      35,
      99,
      -1,
      -1,
      3,
      67,
      19,
      83,
      6,
      70,
      51,
      115,
      38,
      102,
      40,
      104,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      8,
      72,
      18,
      82,
      5,
      69,
      50,
      114,
      37,
      101,
      39,
      103,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      7,
      71,
      12,
      76,
      44,
      108,
      20,
      10,
      74,
      42,
      106,
      -1,
      -1,
      -1,
      178,
      155,
      154,
      171,
      170,
      55,
      119,
      56,
      120,
      177,
      255,
      54,
      118,
      254,
      59,
      57,
      164,
      165,
      60,
      166,
      167,
      58,
      162,
      163,
      34,
      98,
      -1,
      -1,
      2,
      66,
      169,
      153,
      -1,
      33,
      97,
      -1,
      -1,
      1,
      65,
      53,
      -1,
      14,
      78,
      15,
      79,
      176,
      223,
      31,
      161,
      13,
      77,
      30,
      160,
      222,
      -1,
      -1,
      46,
      110,
      47,
      111,
      45,
      109,
      11,
      75,
      43,
      107,
      -1,
      -1,
      0,
      53,
      86,
      67,
      60,
      73,
      52,
      66,
      80,
      55,
      64,
      77,
      -1,
      -1,
      82,
      -1,
      81,
      -1,
      63,
      76,
      61,
      74,
      90,
      89,
      58,
      71,
      56,
      69,
      54,
      68,
      51,
      65,
      59,
      72,
      57,
      70,
      62,
      75,
      20,
      21,
      22,
      23,
      17,
      31,
      34,
      32,
      35,
      19,
      16,
      30,
      33,
      18,
      24,
      36,
      26,
      27,
      28,
      29,
      25,
      37,
      53,
      86,
      67,
      60,
      73,
      52,
      66,
      80,
      55,
      64,
      77,
      -1,
      -1,
      82,
      -1,
      81,
      -1,
      63,
      76,
      61,
      74,
      115,
      90,
      89,
      58,
      71,
      56,
      69,
      54,
      68,
      51,
      65,
      59,
      72,
      57,
      70,
      62,
      75,
      43,
      20,
      40,
      21,
      22,
      23,
      17,
      31,
      34,
      32,
      35,
      19,
      87,
      88,
      16,
      30,
      33,
      18,
      24,
      36,
      26,
      41,
      27,
      28,
      29,
      25,
      37,
      44,
      79,
      12,
      131,
      18,
      11,
      22,
      -1,
      133,
      132,
      21,
      19,
      135,
      134,
      -1,
      16,
      14,
      13,
      10,
      17,
      15,
      20,
      26,
      129,
      128,
      25,
      47,
      49,
      97,
      -1,
      -1,
      51,
      100,
      101,
      -1,
      -1,
      98,
      -1,
      99,
      -1,
      52,
      55,
      -1,
      96,
      -1,
      48,
      -1,
      50,
      54,
      53,
      56,
      23,
      0,
      1,
      4,
      5,
      2,
      3,
      24,
      27,
      30,
      66,
      72,
      40,
      32,
      35,
      69,
      75,
      45,
      -1,
      -1,
      -1,
      -1,
      67,
      73,
      33,
      68,
      84,
      74,
      86,
      58,
      43,
      60,
      34,
      44,
      36,
      46,
      42,
      -1,
      -1,
      -1,
      -1,
      64,
      70,
      28,
      65,
      95,
      71,
      85,
      57,
      38,
      59,
      29,
      39,
      31,
      41,
      37,
      6,
      7,
      9,
      8,
      1,
      0,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      21,
      6,
      73,
      64,
      2,
      0,
      83,
      4,
      15,
      22,
      23,
      32,
      33,
      44,
      45,
      52,
      53,
      34,
      35,
      48,
      49,
      54,
      55,
      12,
      13,
      34,
      35,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      61,
      81,
      82,
      21,
      78,
      79,
      20,
      22,
      23,
      24,
      25,
      39,
      41,
      38,
      40,
      42,
      43,
      12,
      14,
      13,
      15,
      20,
      16,
      17,
      18,
      19,
      32,
      4,
      5,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      16,
      20,
      17,
      21,
      18,
      19,
      24,
      26,
      25,
      27,
      39,
      -1,
      -1,
      -1,
      20,
      39,
      42,
      41,
      40,
      38,
      37,
      -1,
      -1,
      -1,
      33,
      34,
      36,
      35,
      107,
      96,
      5,
      2,
      3,
      2,
      82,
      66,
      0,
      6,
      16,
      86,
      70,
      3,
      9,
      84,
      68,
      2,
      8,
      83,
      67,
      5,
      11,
      85,
      69,
      1,
      7,
      17,
      81,
      65,
      4,
      10,
      93,
      77,
      89,
      73,
      91,
      75,
      92,
      76,
      90,
      74,
      94,
      78,
      87,
      71,
      88,
      72,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      31,
      48,
      49,
      104,
      105,
      100,
      101,
      106,
      102,
      72,
      103,
      -1,
      -1,
      33,
      -1,
      -1,
      -1,
      57,
      7,
      18,
      48,
      -1,
      49,
      -1,
      -1,
      8,
      9,
      12,
      10,
      11,
      13,
      108,
      97,
      69,
      78,
      71,
      70,
      80,
      -1,
      71,
      17,
      -1,
      60,
      56,
      4,
      0,
      4,
      1,
      5,
      2,
      3,
      8,
      10,
      9,
      11,
      14,
      15,
      16,
      17,
      8,
      9,
      10,
      11,
      76,
      67,
      20,
      21,
      82,
      80,
      -1,
      -1,
      75,
      66,
      18,
      19,
      81,
      79,
      0,
      1,
      -1,
      0,
      66,
      67,
      64,
      65,
      68,
      77,
      68,
      46,
      45,
      44,
      85,
      16,
      28,
      29,
      38,
      39,
      0,
      30,
      31,
      40,
      41,
      30,
      31,
      24,
      25,
      34,
      35,
      46,
      47,
      36,
      37,
      50,
      51,
      52,
      53,
      51,
      50,
      38,
      40,
      36,
      37,
      26,
      27,
      74,
      -1,
      110,
      99,
      17,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      54,
      76,
      77,
      55,
      -1,
      -1,
      -1,
      -1,
      2,
      1,
      72,
      53,
      18,
      19,
      -1,
      15,
      14,
      19,
      3,
      88,
      -1,
      -1,
      -1,
      74,
      65,
      3,
      1,
      84,
      5,
      -1,
      -1,
      73,
      16,
      109,
      98,
      58,
      59,
      1,
      67,
      70,
      11,
      68,
      10,
      71,
      64,
      72,
      66,
      65,
      73,
      9,
      8,
      -1,
      -1,
      -1,
      -1,
      52,
      87,
      28,
      29,
      32,
      33,
      42,
      43,
      26,
      27,
      36,
      37,
      20,
      21,
      22,
      23,
      17,
      31,
      34,
      32,
      35,
      19,
      16,
      30,
      33,
      18,
      24,
      36,
      26,
      27,
      28,
      29,
      25,
      37,
      8,
      9,
      10,
      11,
      0,
      1,
      2,
      3,
      12,
      13,
      14,
      15,
      4,
      5,
      6,
      7,
      597,
      -1,
      768,
      50,
      3,
      2,
      -1,
      781,
      806,
      583,
      -1,
      771,
      37,
      32,
      534,
      535,
      866,
      27,
      599,
      -1,
      -1,
      -1,
      826,
      26,
      830,
      798,
      529,
      528,
      530,
      797,
      56,
      92,
      34,
      24,
      -1,
      125,
      126,
      127,
      2,
      18,
      34,
      50,
      66,
      82,
      58,
      74,
      90,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      6,
      22,
      38,
      54,
      70,
      86,
      62,
      78,
      94,
      4,
      20,
      36,
      52,
      68,
      84,
      60,
      76,
      92,
      3,
      19,
      35,
      51,
      67,
      83,
      59,
      75,
      91,
      5,
      21,
      37,
      53,
      69,
      85,
      61,
      77,
      93,
      1,
      17,
      33,
      49,
      65,
      81,
      57,
      73,
      89,
      13,
      29,
      45,
      9,
      25,
      41,
      11,
      27,
      43,
      12,
      28,
      44,
      10,
      26,
      42,
      14,
      30,
      46,
      7,
      23,
      39,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      79,
      95,
      8,
      24,
      40,
      253,
      254,
      255,
      130,
      146,
      162,
      178,
      194,
      210,
      186,
      202,
      218,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      134,
      150,
      166,
      182,
      198,
      214,
      190,
      206,
      222,
      132,
      148,
      164,
      180,
      196,
      212,
      188,
      204,
      220,
      131,
      147,
      163,
      179,
      195,
      211,
      187,
      203,
      219,
      133,
      149,
      165,
      181,
      197,
      213,
      189,
      205,
      221,
      129,
      145,
      161,
      177,
      193,
      209,
      185,
      201,
      217,
      141,
      157,
      173,
      137,
      153,
      169,
      139,
      155,
      171,
      140,
      156,
      172,
      138,
      154,
      170,
      142,
      158,
      174,
      135,
      151,
      167,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      -1,
      207,
      223,
      136,
      152,
      168,
      605,
      1,
      97,
      54,
      524,
      527,
      525,
      526,
      10,
      81,
      80,
      109,
      11,
      15,
      108,
      5,
      6,
      17,
      18,
      19,
      20,
      16,
      4,
      22,
      13,
      83,
      8,
      3,
      106,
      99,
      100,
      14,
      874,
      111,
      110,
      873,
      774,
      804,
      775,
      803,
      550,
      -1,
      786,
      801,
      787,
      802,
      47,
      -1,
      12,
      876,
      875,
      82,
      7,
      21,
      107,
      596,
      551,
      552,
      -1,
      567,
      568,
      764,
      765,
      615,
      614,
      26,
      19,
      -1,
      -1,
      -1,
      39,
      37,
      36,
      38,
      -1,
      22,
      23,
      -1,
      -1,
      24,
      25,
      4,
      16,
      18,
      9,
      13,
      12,
      1,
      0,
      2,
      17,
      10,
      11,
      8,
      7,
      3,
      5,
      6,
      88,
      37,
      -1,
      59,
      57,
      58,
      91,
      36,
      26,
      584,
      531,
      532,
      -1,
      521,
      32,
      34,
      33,
      56,
      45,
      54,
      43,
      -1,
      55,
      44,
      95,
      32,
      62,
      90,
      63,
      60,
      89,
      64,
      61,
      -1,
      2,
      0,
      4,
      -1,
      -1,
      -1,
      -1,
      3,
      -1,
      1,
      -1,
      -1,
      5,
      59,
      796,
      811,
      533,
      -1,
      87,
      39,
      -1,
      582,
      -1,
      -1,
      598,
      -1,
      824,
      24,
      31,
      -1,
      -1,
      -1,
      -1,
      825,
      25,
      829,
      -1,
      -1,
      -1,
      -1,
      -1,
      595,
      602,
      522,
      767,
      -1,
      -1,
      -1,
      -1,
      577,
      601,
      523,
      766,
      -1,
      -1,
      -1,
      -1,
      556,
      554,
      589,
      541,
      590,
      542,
      49,
      22,
      14,
      -1,
      -1,
      778,
      18,
      -1,
      -1,
      777,
      20,
      560,
      558,
      872,
      870,
      834,
      623,
      621,
      619,
      617,
      612,
      610,
      -1,
      800,
      799,
      562,
      561,
      592,
      544,
      593,
      545,
      555,
      553,
      586,
      538,
      587,
      539,
      48,
      21,
      13,
      -1,
      -1,
      780,
      17,
      -1,
      -1,
      779,
      19,
      559,
      557,
      871,
      869,
      833,
      622,
      620,
      618,
      616,
      613,
      611,
      28,
      1,
      66,
      67,
      72,
      68,
      571,
      573,
      569,
      53,
      8,
      6,
      -1,
      814,
      10,
      813,
      12,
      9,
      7,
      812,
      -1,
      773,
      812,
      11,
      536,
      0,
      105,
      55,
      600,
      867,
      28,
      785,
      580,
      603,
      783,
      103,
      604,
      784,
      27,
      15,
      2,
      10,
      6,
      14,
      60,
      4,
      5,
      0,
      9,
      28,
      7,
      12,
      30,
      27,
      8,
      13,
      29,
      16,
      1,
      3,
      11,
      570,
      -1,
      -1,
      84,
      42,
      47,
      43,
      -1,
      2,
      -1,
      864,
      94,
      35,
      25,
      -1,
      -1,
      86,
      46,
      49,
      -1,
      641,
      640,
      643,
      642,
      645,
      644,
      647,
      646,
      649,
      648,
      547,
      548,
      549,
      546,
      98,
      -1,
      96,
      53,
      85,
      51,
      52,
      -1,
      769,
      51,
      4,
      782,
      805,
      -1,
      772,
      38,
      33,
      34,
      -1,
      770,
      52,
      5,
      -1,
      39,
      102,
      101,
      104,
      83,
      82,
      81,
      90,
      89,
      80,
      88,
      87,
      85,
      84,
      86,
      815,
      93,
      33,
      23,
      67,
      66,
      65,
      73,
      72,
      64,
      71,
      70,
      69,
      68,
      74,
      -1,
      865,
      581,
      30,
      576,
      868,
      29,
   },
   std::bitset<1631>("00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000001111111111111111111111111100000000000000000000000000000111111111111111111111111110000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000011111111111111111111000000000000000000000000000001111111111111111111111111111111111111100000000000000000000001111111111111111111111111111111111111000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000011111111111111111111111111111111111111"),
   {
      "buffer_atomic_add",
      "buffer_atomic_add_f32",
      "buffer_atomic_add_x2",
      "buffer_atomic_and",
      "buffer_atomic_and_x2",
      "buffer_atomic_cmpswap",
      "buffer_atomic_cmpswap_x2",
      "buffer_atomic_csub",
      "buffer_atomic_dec",
      "buffer_atomic_dec_x2",
      "buffer_atomic_fcmpswap",
      "buffer_atomic_fcmpswap_x2",
      "buffer_atomic_fmax",
      "buffer_atomic_fmax_x2",
      "buffer_atomic_fmin",
      "buffer_atomic_fmin_x2",
      "buffer_atomic_inc",
      "buffer_atomic_inc_x2",
      "buffer_atomic_or",
      "buffer_atomic_or_x2",
      "buffer_atomic_pk_add_bf16",
      "buffer_atomic_pk_add_f16",
      "buffer_atomic_rsub",
      "buffer_atomic_rsub_x2",
      "buffer_atomic_smax",
      "buffer_atomic_smax_x2",
      "buffer_atomic_smin",
      "buffer_atomic_smin_x2",
      "buffer_atomic_sub",
      "buffer_atomic_sub_x2",
      "buffer_atomic_swap",
      "buffer_atomic_swap_x2",
      "buffer_atomic_umax",
      "buffer_atomic_umax_x2",
      "buffer_atomic_umin",
      "buffer_atomic_umin_x2",
      "buffer_atomic_xor",
      "buffer_atomic_xor_x2",
      "buffer_gl0_inv",
      "buffer_gl1_inv",
      "buffer_load_dword",
      "buffer_load_dwordx2",
      "buffer_load_dwordx3",
      "buffer_load_dwordx4",
      "buffer_load_format_d16_hi_x",
      "buffer_load_format_d16_x",
      "buffer_load_format_d16_xy",
      "buffer_load_format_d16_xyz",
      "buffer_load_format_d16_xyzw",
      "buffer_load_format_x",
      "buffer_load_format_xy",
      "buffer_load_format_xyz",
      "buffer_load_format_xyzw",
      "buffer_load_sbyte",
      "buffer_load_sbyte_d16",
      "buffer_load_sbyte_d16_hi",
      "buffer_load_short_d16",
      "buffer_load_short_d16_hi",
      "buffer_load_sshort",
      "buffer_load_ubyte",
      "buffer_load_ubyte_d16",
      "buffer_load_ubyte_d16_hi",
      "buffer_load_ushort",
      "buffer_store_byte",
      "buffer_store_byte_d16_hi",
      "buffer_store_dword",
      "buffer_store_dwordx2",
      "buffer_store_dwordx3",
      "buffer_store_dwordx4",
      "buffer_store_format_d16_hi_x",
      "buffer_store_format_d16_x",
      "buffer_store_format_d16_xy",
      "buffer_store_format_d16_xyz",
      "buffer_store_format_d16_xyzw",
      "buffer_store_format_x",
      "buffer_store_format_xy",
      "buffer_store_format_xyz",
      "buffer_store_format_xyzw",
      "buffer_store_lds_dword",
      "buffer_store_short",
      "buffer_store_short_d16_hi",
      "buffer_wbinvl1",
      "buffer_wbinvl1_vol",
      "ds_add_f32",
      "ds_add_gs_reg_rtn",
      "ds_add_rtn_f32",
      "ds_add_rtn_u32",
      "ds_add_rtn_u64",
      "ds_add_src2_f32",
      "ds_add_src2_u32",
      "ds_add_src2_u64",
      "ds_add_u32",
      "ds_add_u64",
      "ds_and_b32",
      "ds_and_b64",
      "ds_and_rtn_b32",
      "ds_and_rtn_b64",
      "ds_and_src2_b32",
      "ds_and_src2_b64",
      "ds_append",
      "ds_bpermute_b32",
      "ds_bvh_stack_push4_pop1_rtn_b32",
      "ds_bvh_stack_push8_pop1_rtn_b32",
      "ds_bvh_stack_push8_pop2_rtn_b64",
      "ds_cmpst_b32",
      "ds_cmpst_b64",
      "ds_cmpst_f32",
      "ds_cmpst_f64",
      "ds_cmpst_rtn_b32",
      "ds_cmpst_rtn_b64",
      "ds_cmpst_rtn_f32",
      "ds_cmpst_rtn_f64",
      "ds_cond_sub_rtn",
      "ds_cond_sub_u32",
      "ds_condxchg32_rtn_b128",
      "ds_condxchg32_rtn_b64",
      "ds_consume",
      "ds_dec_rtn_u32",
      "ds_dec_rtn_u64",
      "ds_dec_src2_u32",
      "ds_dec_src2_u64",
      "ds_dec_u32",
      "ds_dec_u64",
      "ds_gws_barrier",
      "ds_gws_init",
      "ds_gws_sema_br",
      "ds_gws_sema_p",
      "ds_gws_sema_release_all",
      "ds_gws_sema_v",
      "ds_inc_rtn_u32",
      "ds_inc_rtn_u64",
      "ds_inc_src2_u32",
      "ds_inc_src2_u64",
      "ds_inc_u32",
      "ds_inc_u64",
      "ds_max_f32",
      "ds_max_f64",
      "ds_max_i32",
      "ds_max_i64",
      "ds_max_rtn_f32",
      "ds_max_rtn_f64",
      "ds_max_rtn_i32",
      "ds_max_rtn_i64",
      "ds_max_rtn_u32",
      "ds_max_rtn_u64",
      "ds_max_src2_f32",
      "ds_max_src2_f64",
      "ds_max_src2_i32",
      "ds_max_src2_i64",
      "ds_max_src2_u32",
      "ds_max_src2_u64",
      "ds_max_u32",
      "ds_max_u64",
      "ds_min_f32",
      "ds_min_f64",
      "ds_min_i32",
      "ds_min_i64",
      "ds_min_rtn_f32",
      "ds_min_rtn_f64",
      "ds_min_rtn_i32",
      "ds_min_rtn_i64",
      "ds_min_rtn_u32",
      "ds_min_rtn_u64",
      "ds_min_src2_f32",
      "ds_min_src2_f64",
      "ds_min_src2_i32",
      "ds_min_src2_i64",
      "ds_min_src2_u32",
      "ds_min_src2_u64",
      "ds_min_u32",
      "ds_min_u64",
      "ds_mskor_b32",
      "ds_mskor_b64",
      "ds_mskor_rtn_b32",
      "ds_mskor_rtn_b64",
      "ds_nop",
      "ds_or_b32",
      "ds_or_b64",
      "ds_or_rtn_b32",
      "ds_or_rtn_b64",
      "ds_or_src2_b32",
      "ds_or_src2_b64",
      "ds_ordered_count",
      "ds_permute_b32",
      "ds_pk_add_bf16",
      "ds_pk_add_f16",
      "ds_pk_add_rtn_bf16",
      "ds_pk_add_rtn_f16",
      "ds_read2_b32",
      "ds_read2_b64",
      "ds_read2st64_b32",
      "ds_read2st64_b64",
      "ds_read_addtid_b32",
      "ds_read_b128",
      "ds_read_b32",
      "ds_read_b64",
      "ds_read_b96",
      "ds_read_i16",
      "ds_read_i8",
      "ds_read_i8_d16",
      "ds_read_i8_d16_hi",
      "ds_read_u16",
      "ds_read_u16_d16",
      "ds_read_u16_d16_hi",
      "ds_read_u8",
      "ds_read_u8_d16",
      "ds_read_u8_d16_hi",
      "ds_rsub_rtn_u32",
      "ds_rsub_rtn_u64",
      "ds_rsub_src2_u32",
      "ds_rsub_src2_u64",
      "ds_rsub_u32",
      "ds_rsub_u64",
      "ds_sub_clamp_rtn_u32",
      "ds_sub_clamp_u32",
      "ds_sub_gs_reg_rtn",
      "ds_sub_rtn_u32",
      "ds_sub_rtn_u64",
      "ds_sub_src2_u32",
      "ds_sub_src2_u64",
      "ds_sub_u32",
      "ds_sub_u64",
      "ds_swizzle_b32",
      "ds_wrap_rtn_b32",
      "ds_write2_b32",
      "ds_write2_b64",
      "ds_write2st64_b32",
      "ds_write2st64_b64",
      "ds_write_addtid_b32",
      "ds_write_b128",
      "ds_write_b16",
      "ds_write_b16_d16_hi",
      "ds_write_b32",
      "ds_write_b64",
      "ds_write_b8",
      "ds_write_b8_d16_hi",
      "ds_write_b96",
      "ds_write_src2_b32",
      "ds_write_src2_b64",
      "ds_wrxchg2_rtn_b32",
      "ds_wrxchg2_rtn_b64",
      "ds_wrxchg2st64_rtn_b32",
      "ds_wrxchg2st64_rtn_b64",
      "ds_wrxchg_rtn_b32",
      "ds_wrxchg_rtn_b64",
      "ds_xor_b32",
      "ds_xor_b64",
      "ds_xor_rtn_b32",
      "ds_xor_rtn_b64",
      "ds_xor_src2_b32",
      "ds_xor_src2_b64",
      "exp",
      "flat_atomic_add",
      "flat_atomic_add_f32",
      "flat_atomic_add_x2",
      "flat_atomic_and",
      "flat_atomic_and_x2",
      "flat_atomic_cmpswap",
      "flat_atomic_cmpswap_x2",
      "flat_atomic_cond_sub_u32",
      "flat_atomic_csub_u32",
      "flat_atomic_dec",
      "flat_atomic_dec_x2",
      "flat_atomic_fcmpswap",
      "flat_atomic_fcmpswap_x2",
      "flat_atomic_fmax",
      "flat_atomic_fmax_x2",
      "flat_atomic_fmin",
      "flat_atomic_fmin_x2",
      "flat_atomic_inc",
      "flat_atomic_inc_x2",
      "flat_atomic_or",
      "flat_atomic_or_x2",
      "flat_atomic_pk_add_bf16",
      "flat_atomic_pk_add_f16",
      "flat_atomic_smax",
      "flat_atomic_smax_x2",
      "flat_atomic_smin",
      "flat_atomic_smin_x2",
      "flat_atomic_sub",
      "flat_atomic_sub_x2",
      "flat_atomic_swap",
      "flat_atomic_swap_x2",
      "flat_atomic_umax",
      "flat_atomic_umax_x2",
      "flat_atomic_umin",
      "flat_atomic_umin_x2",
      "flat_atomic_xor",
      "flat_atomic_xor_x2",
      "flat_load_dword",
      "flat_load_dwordx2",
      "flat_load_dwordx3",
      "flat_load_dwordx4",
      "flat_load_sbyte",
      "flat_load_sbyte_d16",
      "flat_load_sbyte_d16_hi",
      "flat_load_short_d16",
      "flat_load_short_d16_hi",
      "flat_load_sshort",
      "flat_load_ubyte",
      "flat_load_ubyte_d16",
      "flat_load_ubyte_d16_hi",
      "flat_load_ushort",
      "flat_store_byte",
      "flat_store_byte_d16_hi",
      "flat_store_dword",
      "flat_store_dwordx2",
      "flat_store_dwordx3",
      "flat_store_dwordx4",
      "flat_store_short",
      "flat_store_short_d16_hi",
      "global_atomic_add",
      "global_atomic_add_f32",
      "global_atomic_add_x2",
      "global_atomic_and",
      "global_atomic_and_x2",
      "global_atomic_cmpswap",
      "global_atomic_cmpswap_x2",
      "global_atomic_cond_sub_u32",
      "global_atomic_csub",
      "global_atomic_dec",
      "global_atomic_dec_x2",
      "global_atomic_fcmpswap",
      "global_atomic_fcmpswap_x2",
      "global_atomic_fmax",
      "global_atomic_fmax_x2",
      "global_atomic_fmin",
      "global_atomic_fmin_x2",
      "global_atomic_inc",
      "global_atomic_inc_x2",
      "global_atomic_or",
      "global_atomic_or_x2",
      "global_atomic_ordered_add_b64",
      "global_atomic_pk_add_bf16",
      "global_atomic_pk_add_f16",
      "global_atomic_smax",
      "global_atomic_smax_x2",
      "global_atomic_smin",
      "global_atomic_smin_x2",
      "global_atomic_sub",
      "global_atomic_sub_x2",
      "global_atomic_swap",
      "global_atomic_swap_x2",
      "global_atomic_umax",
      "global_atomic_umax_x2",
      "global_atomic_umin",
      "global_atomic_umin_x2",
      "global_atomic_xor",
      "global_atomic_xor_x2",
      "global_inv",
      "global_load_dword",
      "global_load_dword_addtid",
      "global_load_dwordx2",
      "global_load_dwordx3",
      "global_load_dwordx4",
      "global_load_sbyte",
      "global_load_sbyte_d16",
      "global_load_sbyte_d16_hi",
      "global_load_short_d16",
      "global_load_short_d16_hi",
      "global_load_sshort",
      "global_load_tr_b128",
      "global_load_tr_b64",
      "global_load_ubyte",
      "global_load_ubyte_d16",
      "global_load_ubyte_d16_hi",
      "global_load_ushort",
      "global_store_byte",
      "global_store_byte_d16_hi",
      "global_store_dword",
      "global_store_dword_addtid",
      "global_store_dwordx2",
      "global_store_dwordx3",
      "global_store_dwordx4",
      "global_store_short",
      "global_store_short_d16_hi",
      "global_wb",
      "global_wbinv",
      "image_atomic_add",
      "image_atomic_add_flt",
      "image_atomic_and",
      "image_atomic_cmpswap",
      "image_atomic_dec",
      "image_atomic_fcmpswap",
      "image_atomic_fmax",
      "image_atomic_fmin",
      "image_atomic_inc",
      "image_atomic_or",
      "image_atomic_pk_add_bf16",
      "image_atomic_pk_add_f16",
      "image_atomic_rsub",
      "image_atomic_smax",
      "image_atomic_smin",
      "image_atomic_sub",
      "image_atomic_swap",
      "image_atomic_umax",
      "image_atomic_umin",
      "image_atomic_xor",
      "image_bvh64_intersect_ray",
      "image_bvh8_intersect_ray",
      "image_bvh_dual_intersect_ray",
      "image_bvh_intersect_ray",
      "image_gather4",
      "image_gather4_b",
      "image_gather4_b_cl",
      "image_gather4_b_cl_o",
      "image_gather4_b_o",
      "image_gather4_c",
      "image_gather4_c_b",
      "image_gather4_c_b_cl",
      "image_gather4_c_b_cl_o",
      "image_gather4_c_b_o",
      "image_gather4_c_cl",
      "image_gather4_c_cl_o",
      "image_gather4_c_l",
      "image_gather4_c_l_o",
      "image_gather4_c_lz",
      "image_gather4_c_lz_o",
      "image_gather4_c_o",
      "image_gather4_cl",
      "image_gather4_cl_o",
      "image_gather4_l",
      "image_gather4_l_o",
      "image_gather4_lz",
      "image_gather4_lz_o",
      "image_gather4_o",
      "image_get_lod",
      "image_get_resinfo",
      "image_load",
      "image_load_mip",
      "image_load_mip_pck",
      "image_load_mip_pck_sgn",
      "image_load_pck",
      "image_load_pck_sgn",
      "image_msaa_load",
      "image_sample",
      "image_sample_b",
      "image_sample_b_cl",
      "image_sample_b_cl_o",
      "image_sample_b_o",
      "image_sample_c",
      "image_sample_c_b",
      "image_sample_c_b_cl",
      "image_sample_c_b_cl_o",
      "image_sample_c_b_o",
      "image_sample_c_cd",
      "image_sample_c_cd_cl",
      "image_sample_c_cd_cl_o",
      "image_sample_c_cd_o",
      "image_sample_c_cl",
      "image_sample_c_cl_o",
      "image_sample_c_d",
      "image_sample_c_d_cl",
      "image_sample_c_d_cl_g16",
      "image_sample_c_d_cl_o",
      "image_sample_c_d_cl_o_g16",
      "image_sample_c_d_g16",
      "image_sample_c_d_o",
      "image_sample_c_d_o_g16",
      "image_sample_c_l",
      "image_sample_c_l_o",
      "image_sample_c_lz",
      "image_sample_c_lz_o",
      "image_sample_c_o",
      "image_sample_cd",
      "image_sample_cd_cl",
      "image_sample_cd_cl_o",
      "image_sample_cd_o",
      "image_sample_cl",
      "image_sample_cl_o",
      "image_sample_d",
      "image_sample_d_cl",
      "image_sample_d_cl_g16",
      "image_sample_d_cl_o",
      "image_sample_d_cl_o_g16",
      "image_sample_d_g16",
      "image_sample_d_o",
      "image_sample_d_o_g16",
      "image_sample_l",
      "image_sample_l_o",
      "image_sample_lz",
      "image_sample_lz_o",
      "image_sample_o",
      "image_store",
      "image_store_mip",
      "image_store_mip_pck",
      "image_store_pck",
      "lds_direct_load",
      "lds_param_load",
      "p_as_uniform",
      "p_barrier",
      "p_boolean_phi",
      "p_bpermute_permlane",
      "p_bpermute_readlane",
      "p_bpermute_shared_vgpr",
      "p_branch",
      "p_call",
      "p_callee_stack_ptr",
      "p_cbranch",
      "p_cbranch_nz",
      "p_cbranch_z",
      "p_constaddr",
      "p_constaddr_addlo",
      "p_constaddr_getpc",
      "p_create_vector",
      "p_debug_info",
      "p_demote_to_helper",
      "p_discard_if",
      "p_dual_src_export_gfx11",
      "p_elect",
      "p_end_linear_vgpr",
      "p_end_with_regs",
      "p_end_wqm",
      "p_exclusive_scan",
      "p_exit_early_if_not",
      "p_extract",
      "p_extract_vector",
      "p_inclusive_scan",
      "p_init_scratch",
      "p_insert",
      "p_interp_gfx11",
      "p_is_helper",
      "p_jump_to_epilog",
      "p_linear_phi",
      "p_load_symbol",
      "p_logical_end",
      "p_logical_start",
      "p_parallelcopy",
      "p_permlane64_shared_vgpr",
      "p_phi",
      "p_pops_gfx9_add_exiting_wave_id",
      "p_pops_gfx9_ordered_section_done",
      "p_pops_gfx9_overlapped_wave_wait_done",
      "p_reduce",
      "p_reload",
      "p_reload_preserved",
      "p_resume_shader_address",
      "p_resumeaddr_addlo",
      "p_resumeaddr_getpc",
      "p_return",
      "p_s_cvt_f16_f32_rtne",
      "p_shader_cycles_hi_lo_hi",
      "p_spill",
      "p_split_vector",
      "p_start_linear_vgpr",
      "p_startpgm",
      "p_unit_test",
      "p_v_cvt_f16_f32_rtne",
      "p_v_cvt_f16_f32_rtni",
      "p_v_cvt_f16_f32_rtpi",
      "p_v_cvt_pk_fp8_f32_ovfl",
      "p_v_fma_mixhi_f16_rtz",
      "p_v_fma_mixlo_f16_rtz",
      "s_abs_i32",
      "s_absdiff_i32",
      "s_add_f16",
      "s_add_f32",
      "s_add_i32",
      "s_add_u32",
      "s_add_u64",
      "s_addc_u32",
      "s_addk_i32",
      "s_and_b32",
      "s_and_b64",
      "s_and_saveexec_b32",
      "s_and_saveexec_b64",
      "s_andn1_saveexec_b32",
      "s_andn1_saveexec_b64",
      "s_andn1_wrexec_b32",
      "s_andn1_wrexec_b64",
      "s_andn2_b32",
      "s_andn2_b64",
      "s_andn2_saveexec_b32",
      "s_andn2_saveexec_b64",
      "s_andn2_wrexec_b32",
      "s_andn2_wrexec_b64",
      "s_ashr_i32",
      "s_ashr_i64",
      "s_atc_probe",
      "s_atc_probe_buffer",
      "s_atomic_add",
      "s_atomic_add_x2",
      "s_atomic_and",
      "s_atomic_and_x2",
      "s_atomic_cmpswap",
      "s_atomic_cmpswap_x2",
      "s_atomic_dec",
      "s_atomic_dec_x2",
      "s_atomic_inc",
      "s_atomic_inc_x2",
      "s_atomic_or",
      "s_atomic_or_x2",
      "s_atomic_smax",
      "s_atomic_smax_x2",
      "s_atomic_smin",
      "s_atomic_smin_x2",
      "s_atomic_sub",
      "s_atomic_sub_x2",
      "s_atomic_swap",
      "s_atomic_swap_x2",
      "s_atomic_umax",
      "s_atomic_umax_x2",
      "s_atomic_umin",
      "s_atomic_umin_x2",
      "s_atomic_xor",
      "s_atomic_xor_x2",
      "s_barrier",
      "s_barrier_init",
      "s_barrier_join",
      "s_barrier_leave",
      "s_barrier_signal",
      "s_barrier_signal_isfirst",
      "s_barrier_wait",
      "s_bcnt0_i32_b32",
      "s_bcnt0_i32_b64",
      "s_bcnt1_i32_b32",
      "s_bcnt1_i32_b64",
      "s_bfe_i32",
      "s_bfe_i64",
      "s_bfe_u32",
      "s_bfe_u64",
      "s_bfm_b32",
      "s_bfm_b64",
      "s_bitcmp0_b32",
      "s_bitcmp0_b64",
      "s_bitcmp1_b32",
      "s_bitcmp1_b64",
      "s_bitreplicate_b64_b32",
      "s_bitset0_b32",
      "s_bitset0_b64",
      "s_bitset1_b32",
      "s_bitset1_b64",
      "s_branch",
      "s_brev_b32",
      "s_brev_b64",
      "s_buffer_atomic_add",
      "s_buffer_atomic_add_x2",
      "s_buffer_atomic_and",
      "s_buffer_atomic_and_x2",
      "s_buffer_atomic_cmpswap",
      "s_buffer_atomic_cmpswap_x2",
      "s_buffer_atomic_dec",
      "s_buffer_atomic_dec_x2",
      "s_buffer_atomic_inc",
      "s_buffer_atomic_inc_x2",
      "s_buffer_atomic_or",
      "s_buffer_atomic_or_x2",
      "s_buffer_atomic_smax",
      "s_buffer_atomic_smax_x2",
      "s_buffer_atomic_smin",
      "s_buffer_atomic_smin_x2",
      "s_buffer_atomic_sub",
      "s_buffer_atomic_sub_x2",
      "s_buffer_atomic_swap",
      "s_buffer_atomic_swap_x2",
      "s_buffer_atomic_umax",
      "s_buffer_atomic_umax_x2",
      "s_buffer_atomic_umin",
      "s_buffer_atomic_umin_x2",
      "s_buffer_atomic_xor",
      "s_buffer_atomic_xor_x2",
      "s_buffer_load_dword",
      "s_buffer_load_dwordx16",
      "s_buffer_load_dwordx2",
      "s_buffer_load_dwordx3",
      "s_buffer_load_dwordx4",
      "s_buffer_load_dwordx8",
      "s_buffer_load_sbyte",
      "s_buffer_load_sshort",
      "s_buffer_load_ubyte",
      "s_buffer_load_ushort",
      "s_buffer_prefetch_data",
      "s_buffer_store_dword",
      "s_buffer_store_dwordx2",
      "s_buffer_store_dwordx4",
      "s_call_b64",
      "s_cbranch_cdbgsys",
      "s_cbranch_cdbgsys_and_user",
      "s_cbranch_cdbgsys_or_user",
      "s_cbranch_cdbguser",
      "s_cbranch_execnz",
      "s_cbranch_execz",
      "s_cbranch_g_fork",
      "s_cbranch_i_fork",
      "s_cbranch_join",
      "s_cbranch_scc0",
      "s_cbranch_scc1",
      "s_cbranch_vccnz",
      "s_cbranch_vccz",
      "s_ceil_f16",
      "s_ceil_f32",
      "s_clause",
      "s_cmov_b32",
      "s_cmov_b64",
      "s_cmovk_i32",
      "s_cmp_eq_f16",
      "s_cmp_eq_f32",
      "s_cmp_eq_i32",
      "s_cmp_eq_u32",
      "s_cmp_eq_u64",
      "s_cmp_ge_f16",
      "s_cmp_ge_f32",
      "s_cmp_ge_i32",
      "s_cmp_ge_u32",
      "s_cmp_gt_f16",
      "s_cmp_gt_f32",
      "s_cmp_gt_i32",
      "s_cmp_gt_u32",
      "s_cmp_le_f16",
      "s_cmp_le_f32",
      "s_cmp_le_i32",
      "s_cmp_le_u32",
      "s_cmp_lg_f16",
      "s_cmp_lg_f32",
      "s_cmp_lg_i32",
      "s_cmp_lg_u32",
      "s_cmp_lg_u64",
      "s_cmp_lt_f16",
      "s_cmp_lt_f32",
      "s_cmp_lt_i32",
      "s_cmp_lt_u32",
      "s_cmp_neq_f16",
      "s_cmp_neq_f32",
      "s_cmp_nge_f16",
      "s_cmp_nge_f32",
      "s_cmp_ngt_f16",
      "s_cmp_ngt_f32",
      "s_cmp_nle_f16",
      "s_cmp_nle_f32",
      "s_cmp_nlg_f16",
      "s_cmp_nlg_f32",
      "s_cmp_nlt_f16",
      "s_cmp_nlt_f32",
      "s_cmp_o_f16",
      "s_cmp_o_f32",
      "s_cmp_u_f16",
      "s_cmp_u_f32",
      "s_cmpk_eq_i32",
      "s_cmpk_eq_u32",
      "s_cmpk_ge_i32",
      "s_cmpk_ge_u32",
      "s_cmpk_gt_i32",
      "s_cmpk_gt_u32",
      "s_cmpk_le_i32",
      "s_cmpk_le_u32",
      "s_cmpk_lg_i32",
      "s_cmpk_lg_u32",
      "s_cmpk_lt_i32",
      "s_cmpk_lt_u32",
      "s_code_end",
      "s_cselect_b32",
      "s_cselect_b64",
      "s_cvt_f16_f32",
      "s_cvt_f32_f16",
      "s_cvt_f32_i32",
      "s_cvt_f32_u32",
      "s_cvt_hi_f32_f16",
      "s_cvt_i32_f32",
      "s_cvt_pk_rtz_f16_f32",
      "s_cvt_u32_f32",
      "s_dcache_discard",
      "s_dcache_discard_x2",
      "s_dcache_inv",
      "s_dcache_inv_vol",
      "s_dcache_wb",
      "s_dcache_wb_vol",
      "s_decperflevel",
      "s_delay_alu",
      "s_denorm_mode",
      "s_endpgm",
      "s_endpgm_ordered_ps_done",
      "s_endpgm_saved",
      "s_ff0_i32_b32",
      "s_ff0_i32_b64",
      "s_ff1_i32_b32",
      "s_ff1_i32_b64",
      "s_flbit_i32",
      "s_flbit_i32_b32",
      "s_flbit_i32_b64",
      "s_flbit_i32_i64",
      "s_floor_f16",
      "s_floor_f32",
      "s_fmaak_f32",
      "s_fmac_f16",
      "s_fmac_f32",
      "s_fmamk_f32",
      "s_get_barrier_state",
      "s_get_waveid_in_workgroup",
      "s_getpc_b64",
      "s_getreg_b32",
      "s_gl1_inv",
      "s_icache_inv",
      "s_incperflevel",
      "s_inst_prefetch",
      "s_load_dword",
      "s_load_dwordx16",
      "s_load_dwordx2",
      "s_load_dwordx3",
      "s_load_dwordx4",
      "s_load_dwordx8",
      "s_load_sbyte",
      "s_load_sshort",
      "s_load_ubyte",
      "s_load_ushort",
      "s_lshl1_add_u32",
      "s_lshl2_add_u32",
      "s_lshl3_add_u32",
      "s_lshl4_add_u32",
      "s_lshl_b32",
      "s_lshl_b64",
      "s_lshr_b32",
      "s_lshr_b64",
      "s_max_f16",
      "s_max_f32",
      "s_max_i32",
      "s_max_u32",
      "s_maximum_f16",
      "s_maximum_f32",
      "s_memrealtime",
      "s_memtime",
      "s_min_f16",
      "s_min_f32",
      "s_min_i32",
      "s_min_u32",
      "s_minimum_f16",
      "s_minimum_f32",
      "s_mov_b32",
      "s_mov_b64",
      "s_mov_fed_b32",
      "s_movk_i32",
      "s_movreld_b32",
      "s_movreld_b64",
      "s_movrels_b32",
      "s_movrels_b64",
      "s_movrelsd_2_b32",
      "s_mul_f16",
      "s_mul_f32",
      "s_mul_hi_i32",
      "s_mul_hi_u32",
      "s_mul_i32",
      "s_mul_u64",
      "s_mulk_i32",
      "s_nand_b32",
      "s_nand_b64",
      "s_nand_saveexec_b32",
      "s_nand_saveexec_b64",
      "s_nop",
      "s_nor_b32",
      "s_nor_b64",
      "s_nor_saveexec_b32",
      "s_nor_saveexec_b64",
      "s_not_b32",
      "s_not_b64",
      "s_or_b32",
      "s_or_b64",
      "s_or_saveexec_b32",
      "s_or_saveexec_b64",
      "s_orn1_saveexec_b32",
      "s_orn1_saveexec_b64",
      "s_orn2_b32",
      "s_orn2_b64",
      "s_orn2_saveexec_b32",
      "s_orn2_saveexec_b64",
      "s_pack_hh_b32_b16",
      "s_pack_hl_b32_b16",
      "s_pack_lh_b32_b16",
      "s_pack_ll_b32_b16",
      "s_prefetch_data",
      "s_prefetch_data_pc_rel",
      "s_prefetch_inst",
      "s_prefetch_inst_pc_rel",
      "s_quadmask_b32",
      "s_quadmask_b64",
      "s_rfe_b64",
      "s_rfe_restore_b64",
      "s_rndne_f16",
      "s_rndne_f32",
      "s_round_mode",
      "s_scratch_load_dword",
      "s_scratch_load_dwordx2",
      "s_scratch_load_dwordx4",
      "s_scratch_store_dword",
      "s_scratch_store_dwordx2",
      "s_scratch_store_dwordx4",
      "s_sendmsg",
      "s_sendmsg_rtn_b32",
      "s_sendmsg_rtn_b64",
      "s_sendmsghalt",
      "s_set_gpr_idx_idx",
      "s_set_gpr_idx_mode",
      "s_set_gpr_idx_off",
      "s_set_gpr_idx_on",
      "s_sethalt",
      "s_setkill",
      "s_setpc_b64",
      "s_setprio",
      "s_setreg_b32",
      "s_setreg_imm32_b32",
      "s_setvskip",
      "s_sext_i32_i16",
      "s_sext_i32_i8",
      "s_singleuse_vdst",
      "s_sleep",
      "s_sleep_var",
      "s_store_dword",
      "s_store_dwordx2",
      "s_store_dwordx4",
      "s_sub_f16",
      "s_sub_f32",
      "s_sub_i32",
      "s_sub_u32",
      "s_sub_u64",
      "s_subb_u32",
      "s_subvector_loop_begin",
      "s_subvector_loop_end",
      "s_swappc_b64",
      "s_trap",
      "s_trunc_f16",
      "s_trunc_f32",
      "s_ttracedata",
      "s_ttracedata_imm",
      "s_version",
      "s_wait_bvhcnt",
      "s_wait_dscnt",
      "s_wait_event",
      "s_wait_expcnt",
      "s_wait_idle",
      "s_wait_kmcnt",
      "s_wait_loadcnt",
      "s_wait_loadcnt_dscnt",
      "s_wait_samplecnt",
      "s_wait_storecnt",
      "s_wait_storecnt_dscnt",
      "s_waitcnt",
      "s_waitcnt_depctr",
      "s_waitcnt_expcnt",
      "s_waitcnt_lgkmcnt",
      "s_waitcnt_vmcnt",
      "s_waitcnt_vscnt",
      "s_wakeup",
      "s_wakeup_barrier",
      "s_wqm_b32",
      "s_wqm_b64",
      "s_xnor_b32",
      "s_xnor_b64",
      "s_xnor_saveexec_b32",
      "s_xnor_saveexec_b64",
      "s_xor_b32",
      "s_xor_b64",
      "s_xor_saveexec_b32",
      "s_xor_saveexec_b64",
      "scratch_load_dword",
      "scratch_load_dwordx2",
      "scratch_load_dwordx3",
      "scratch_load_dwordx4",
      "scratch_load_sbyte",
      "scratch_load_sbyte_d16",
      "scratch_load_sbyte_d16_hi",
      "scratch_load_short_d16",
      "scratch_load_short_d16_hi",
      "scratch_load_sshort",
      "scratch_load_ubyte",
      "scratch_load_ubyte_d16",
      "scratch_load_ubyte_d16_hi",
      "scratch_load_ushort",
      "scratch_store_byte",
      "scratch_store_byte_d16_hi",
      "scratch_store_dword",
      "scratch_store_dwordx2",
      "scratch_store_dwordx3",
      "scratch_store_dwordx4",
      "scratch_store_short",
      "scratch_store_short_d16_hi",
      "tbuffer_load_format_d16_x",
      "tbuffer_load_format_d16_xy",
      "tbuffer_load_format_d16_xyz",
      "tbuffer_load_format_d16_xyzw",
      "tbuffer_load_format_x",
      "tbuffer_load_format_xy",
      "tbuffer_load_format_xyz",
      "tbuffer_load_format_xyzw",
      "tbuffer_store_format_d16_x",
      "tbuffer_store_format_d16_xy",
      "tbuffer_store_format_d16_xyz",
      "tbuffer_store_format_d16_xyzw",
      "tbuffer_store_format_x",
      "tbuffer_store_format_xy",
      "tbuffer_store_format_xyz",
      "tbuffer_store_format_xyzw",
      "v_add3_u32",
      "v_add_co_u32",
      "v_add_co_u32_e64",
      "v_add_f16",
      "v_add_f32",
      "v_add_f64",
      "v_add_f64_e64",
      "v_add_i16",
      "v_add_i32",
      "v_add_lshl_u32",
      "v_add_u16",
      "v_add_u16_e64",
      "v_add_u32",
      "v_addc_co_u32",
      "v_alignbit_b32",
      "v_alignbyte_b32",
      "v_and_b16",
      "v_and_b32",
      "v_and_or_b32",
      "v_ashr_i32",
      "v_ashr_i64",
      "v_ashrrev_i16",
      "v_ashrrev_i16_e64",
      "v_ashrrev_i32",
      "v_ashrrev_i64",
      "v_bcnt_u32_b32",
      "v_bfe_i32",
      "v_bfe_u32",
      "v_bfi_b32",
      "v_bfm_b32",
      "v_bfrev_b32",
      "v_ceil_f16",
      "v_ceil_f32",
      "v_ceil_f64",
      "v_clrexcp",
      "v_cmp_class_f16",
      "v_cmp_class_f32",
      "v_cmp_class_f64",
      "v_cmp_eq_f16",
      "v_cmp_eq_f32",
      "v_cmp_eq_f64",
      "v_cmp_eq_i16",
      "v_cmp_eq_i32",
      "v_cmp_eq_i64",
      "v_cmp_eq_u16",
      "v_cmp_eq_u32",
      "v_cmp_eq_u64",
      "v_cmp_f_f16",
      "v_cmp_f_f32",
      "v_cmp_f_f64",
      "v_cmp_f_i16",
      "v_cmp_f_i32",
      "v_cmp_f_i64",
      "v_cmp_f_u16",
      "v_cmp_f_u32",
      "v_cmp_f_u64",
      "v_cmp_ge_f16",
      "v_cmp_ge_f32",
      "v_cmp_ge_f64",
      "v_cmp_ge_i16",
      "v_cmp_ge_i32",
      "v_cmp_ge_i64",
      "v_cmp_ge_u16",
      "v_cmp_ge_u32",
      "v_cmp_ge_u64",
      "v_cmp_gt_f16",
      "v_cmp_gt_f32",
      "v_cmp_gt_f64",
      "v_cmp_gt_i16",
      "v_cmp_gt_i32",
      "v_cmp_gt_i64",
      "v_cmp_gt_u16",
      "v_cmp_gt_u32",
      "v_cmp_gt_u64",
      "v_cmp_le_f16",
      "v_cmp_le_f32",
      "v_cmp_le_f64",
      "v_cmp_le_i16",
      "v_cmp_le_i32",
      "v_cmp_le_i64",
      "v_cmp_le_u16",
      "v_cmp_le_u32",
      "v_cmp_le_u64",
      "v_cmp_lg_f16",
      "v_cmp_lg_f32",
      "v_cmp_lg_f64",
      "v_cmp_lg_i16",
      "v_cmp_lg_i32",
      "v_cmp_lg_i64",
      "v_cmp_lg_u16",
      "v_cmp_lg_u32",
      "v_cmp_lg_u64",
      "v_cmp_lt_f16",
      "v_cmp_lt_f32",
      "v_cmp_lt_f64",
      "v_cmp_lt_i16",
      "v_cmp_lt_i32",
      "v_cmp_lt_i64",
      "v_cmp_lt_u16",
      "v_cmp_lt_u32",
      "v_cmp_lt_u64",
      "v_cmp_neq_f16",
      "v_cmp_neq_f32",
      "v_cmp_neq_f64",
      "v_cmp_nge_f16",
      "v_cmp_nge_f32",
      "v_cmp_nge_f64",
      "v_cmp_ngt_f16",
      "v_cmp_ngt_f32",
      "v_cmp_ngt_f64",
      "v_cmp_nle_f16",
      "v_cmp_nle_f32",
      "v_cmp_nle_f64",
      "v_cmp_nlg_f16",
      "v_cmp_nlg_f32",
      "v_cmp_nlg_f64",
      "v_cmp_nlt_f16",
      "v_cmp_nlt_f32",
      "v_cmp_nlt_f64",
      "v_cmp_o_f16",
      "v_cmp_o_f32",
      "v_cmp_o_f64",
      "v_cmp_tru_f16",
      "v_cmp_tru_f32",
      "v_cmp_tru_f64",
      "v_cmp_tru_i16",
      "v_cmp_tru_i32",
      "v_cmp_tru_i64",
      "v_cmp_tru_u16",
      "v_cmp_tru_u32",
      "v_cmp_tru_u64",
      "v_cmp_u_f16",
      "v_cmp_u_f32",
      "v_cmp_u_f64",
      "v_cmpx_class_f16",
      "v_cmpx_class_f32",
      "v_cmpx_class_f64",
      "v_cmpx_eq_f16",
      "v_cmpx_eq_f32",
      "v_cmpx_eq_f64",
      "v_cmpx_eq_i16",
      "v_cmpx_eq_i32",
      "v_cmpx_eq_i64",
      "v_cmpx_eq_u16",
      "v_cmpx_eq_u32",
      "v_cmpx_eq_u64",
      "v_cmpx_f_f16",
      "v_cmpx_f_f32",
      "v_cmpx_f_f64",
      "v_cmpx_f_i16",
      "v_cmpx_f_i32",
      "v_cmpx_f_i64",
      "v_cmpx_f_u16",
      "v_cmpx_f_u32",
      "v_cmpx_f_u64",
      "v_cmpx_ge_f16",
      "v_cmpx_ge_f32",
      "v_cmpx_ge_f64",
      "v_cmpx_ge_i16",
      "v_cmpx_ge_i32",
      "v_cmpx_ge_i64",
      "v_cmpx_ge_u16",
      "v_cmpx_ge_u32",
      "v_cmpx_ge_u64",
      "v_cmpx_gt_f16",
      "v_cmpx_gt_f32",
      "v_cmpx_gt_f64",
      "v_cmpx_gt_i16",
      "v_cmpx_gt_i32",
      "v_cmpx_gt_i64",
      "v_cmpx_gt_u16",
      "v_cmpx_gt_u32",
      "v_cmpx_gt_u64",
      "v_cmpx_le_f16",
      "v_cmpx_le_f32",
      "v_cmpx_le_f64",
      "v_cmpx_le_i16",
      "v_cmpx_le_i32",
      "v_cmpx_le_i64",
      "v_cmpx_le_u16",
      "v_cmpx_le_u32",
      "v_cmpx_le_u64",
      "v_cmpx_lg_f16",
      "v_cmpx_lg_f32",
      "v_cmpx_lg_f64",
      "v_cmpx_lg_i16",
      "v_cmpx_lg_i32",
      "v_cmpx_lg_i64",
      "v_cmpx_lg_u16",
      "v_cmpx_lg_u32",
      "v_cmpx_lg_u64",
      "v_cmpx_lt_f16",
      "v_cmpx_lt_f32",
      "v_cmpx_lt_f64",
      "v_cmpx_lt_i16",
      "v_cmpx_lt_i32",
      "v_cmpx_lt_i64",
      "v_cmpx_lt_u16",
      "v_cmpx_lt_u32",
      "v_cmpx_lt_u64",
      "v_cmpx_neq_f16",
      "v_cmpx_neq_f32",
      "v_cmpx_neq_f64",
      "v_cmpx_nge_f16",
      "v_cmpx_nge_f32",
      "v_cmpx_nge_f64",
      "v_cmpx_ngt_f16",
      "v_cmpx_ngt_f32",
      "v_cmpx_ngt_f64",
      "v_cmpx_nle_f16",
      "v_cmpx_nle_f32",
      "v_cmpx_nle_f64",
      "v_cmpx_nlg_f16",
      "v_cmpx_nlg_f32",
      "v_cmpx_nlg_f64",
      "v_cmpx_nlt_f16",
      "v_cmpx_nlt_f32",
      "v_cmpx_nlt_f64",
      "v_cmpx_o_f16",
      "v_cmpx_o_f32",
      "v_cmpx_o_f64",
      "v_cmpx_tru_f16",
      "v_cmpx_tru_f32",
      "v_cmpx_tru_f64",
      "v_cmpx_tru_i16",
      "v_cmpx_tru_i32",
      "v_cmpx_tru_i64",
      "v_cmpx_tru_u16",
      "v_cmpx_tru_u32",
      "v_cmpx_tru_u64",
      "v_cmpx_u_f16",
      "v_cmpx_u_f32",
      "v_cmpx_u_f64",
      "v_cndmask_b16",
      "v_cndmask_b32",
      "v_cos_f16",
      "v_cos_f32",
      "v_cubeid_f32",
      "v_cubema_f32",
      "v_cubesc_f32",
      "v_cubetc_f32",
      "v_cvt_f16_f32",
      "v_cvt_f16_i16",
      "v_cvt_f16_u16",
      "v_cvt_f32_bf8",
      "v_cvt_f32_f16",
      "v_cvt_f32_f64",
      "v_cvt_f32_fp8",
      "v_cvt_f32_i32",
      "v_cvt_f32_u32",
      "v_cvt_f32_ubyte0",
      "v_cvt_f32_ubyte1",
      "v_cvt_f32_ubyte2",
      "v_cvt_f32_ubyte3",
      "v_cvt_f64_f32",
      "v_cvt_f64_i32",
      "v_cvt_f64_u32",
      "v_cvt_flr_i32_f32",
      "v_cvt_i16_f16",
      "v_cvt_i32_f32",
      "v_cvt_i32_f64",
      "v_cvt_i32_i16",
      "v_cvt_norm_i16_f16",
      "v_cvt_norm_u16_f16",
      "v_cvt_off_f32_i4",
      "v_cvt_pk_bf8_f32",
      "v_cvt_pk_f32_bf8",
      "v_cvt_pk_f32_fp8",
      "v_cvt_pk_fp8_f32",
      "v_cvt_pk_i16_f32",
      "v_cvt_pk_i16_i32",
      "v_cvt_pk_u16_f32",
      "v_cvt_pk_u16_u32",
      "v_cvt_pk_u8_f32",
      "v_cvt_pkaccum_u8_f32",
      "v_cvt_pknorm_i16_f16",
      "v_cvt_pknorm_i16_f32",
      "v_cvt_pknorm_u16_f16",
      "v_cvt_pknorm_u16_f32",
      "v_cvt_pkrtz_f16_f32",
      "v_cvt_pkrtz_f16_f32_e64",
      "v_cvt_rpi_i32_f32",
      "v_cvt_sr_bf8_f32",
      "v_cvt_sr_fp8_f32",
      "v_cvt_u16_f16",
      "v_cvt_u32_f32",
      "v_cvt_u32_f64",
      "v_cvt_u32_u16",
      "v_div_fixup_f16",
      "v_div_fixup_f32",
      "v_div_fixup_f64",
      "v_div_fixup_legacy_f16",
      "v_div_fmas_f32",
      "v_div_fmas_f64",
      "v_div_scale_f32",
      "v_div_scale_f64",
      "v_dot2_bf16_bf16",
      "v_dot2_f16_f16",
      "v_dot2_f32_bf16",
      "v_dot2_f32_f16",
      "v_dot2_i32_i16",
      "v_dot2_u32_u16",
      "v_dot2c_f32_f16",
      "v_dot4_f32_bf8_bf8",
      "v_dot4_f32_bf8_fp8",
      "v_dot4_f32_fp8_bf8",
      "v_dot4_f32_fp8_fp8",
      "v_dot4_i32_i8",
      "v_dot4_i32_iu8",
      "v_dot4_u32_u8",
      "v_dot4c_i32_i8",
      "v_dot8_i32_i4",
      "v_dot8_i32_iu4",
      "v_dot8_u32_u4",
      "v_dual_add_f32",
      "v_dual_add_nc_u32",
      "v_dual_and_b32",
      "v_dual_cndmask_b32",
      "v_dual_dot2acc_f32_bf16",
      "v_dual_dot2acc_f32_f16",
      "v_dual_fmaak_f32",
      "v_dual_fmac_f32",
      "v_dual_fmamk_f32",
      "v_dual_lshlrev_b32",
      "v_dual_max_f32",
      "v_dual_min_f32",
      "v_dual_mov_b32",
      "v_dual_mul_dx9_zero_f32",
      "v_dual_mul_f32",
      "v_dual_sub_f32",
      "v_dual_subrev_f32",
      "v_exp_f16",
      "v_exp_f32",
      "v_exp_legacy_f32",
      "v_ffbh_i32",
      "v_ffbh_u32",
      "v_ffbl_b32",
      "v_floor_f16",
      "v_floor_f32",
      "v_floor_f64",
      "v_fma_f16",
      "v_fma_f32",
      "v_fma_f64",
      "v_fma_legacy_f16",
      "v_fma_legacy_f32",
      "v_fma_mix_f32",
      "v_fma_mixhi_f16",
      "v_fma_mixlo_f16",
      "v_fmaak_f16",
      "v_fmaak_f32",
      "v_fmac_f16",
      "v_fmac_f32",
      "v_fmac_legacy_f32",
      "v_fmamk_f16",
      "v_fmamk_f32",
      "v_fract_f16",
      "v_fract_f32",
      "v_fract_f64",
      "v_frexp_exp_i16_f16",
      "v_frexp_exp_i32_f32",
      "v_frexp_exp_i32_f64",
      "v_frexp_mant_f16",
      "v_frexp_mant_f32",
      "v_frexp_mant_f64",
      "v_interp_mov_f32",
      "v_interp_p10_f16_f32_inreg",
      "v_interp_p10_f32_inreg",
      "v_interp_p10_rtz_f16_f32_inreg",
      "v_interp_p1_f32",
      "v_interp_p1ll_f16",
      "v_interp_p1lv_f16",
      "v_interp_p2_f16",
      "v_interp_p2_f16_f32_inreg",
      "v_interp_p2_f32",
      "v_interp_p2_f32_inreg",
      "v_interp_p2_hi_f16",
      "v_interp_p2_legacy_f16",
      "v_interp_p2_rtz_f16_f32_inreg",
      "v_ldexp_f16",
      "v_ldexp_f32",
      "v_ldexp_f64",
      "v_lerp_u8",
      "v_log_clamp_f32",
      "v_log_f16",
      "v_log_f32",
      "v_log_legacy_f32",
      "v_lshl_add_u32",
      "v_lshl_b32",
      "v_lshl_b64",
      "v_lshl_or_b32",
      "v_lshlrev_b16",
      "v_lshlrev_b16_e64",
      "v_lshlrev_b32",
      "v_lshlrev_b64",
      "v_lshlrev_b64_e64",
      "v_lshr_b32",
      "v_lshr_b64",
      "v_lshrrev_b16",
      "v_lshrrev_b16_e64",
      "v_lshrrev_b32",
      "v_lshrrev_b64",
      "v_mac_f16",
      "v_mac_f32",
      "v_mac_legacy_f32",
      "v_mad_f16",
      "v_mad_f32",
      "v_mad_i16",
      "v_mad_i32_i16",
      "v_mad_i32_i24",
      "v_mad_i64_i32",
      "v_mad_legacy_f16",
      "v_mad_legacy_f32",
      "v_mad_legacy_i16",
      "v_mad_legacy_u16",
      "v_mad_u16",
      "v_mad_u32_u16",
      "v_mad_u32_u24",
      "v_mad_u64_u32",
      "v_madak_f16",
      "v_madak_f32",
      "v_madmk_f16",
      "v_madmk_f32",
      "v_max3_f16",
      "v_max3_f32",
      "v_max3_i16",
      "v_max3_i32",
      "v_max3_u16",
      "v_max3_u32",
      "v_max_f16",
      "v_max_f32",
      "v_max_f64",
      "v_max_f64_e64",
      "v_max_i16",
      "v_max_i16_e64",
      "v_max_i32",
      "v_max_legacy_f32",
      "v_max_u16",
      "v_max_u16_e64",
      "v_max_u32",
      "v_maximum3_f16",
      "v_maximum3_f32",
      "v_maximum_f16",
      "v_maximum_f32",
      "v_maximum_f64",
      "v_maximumminimum_f16",
      "v_maximumminimum_f32",
      "v_maxmin_f16",
      "v_maxmin_f32",
      "v_maxmin_i32",
      "v_maxmin_u32",
      "v_mbcnt_hi_u32_b32",
      "v_mbcnt_hi_u32_b32_e64",
      "v_mbcnt_lo_u32_b32",
      "v_med3_f16",
      "v_med3_f32",
      "v_med3_i16",
      "v_med3_i32",
      "v_med3_u16",
      "v_med3_u32",
      "v_min3_f16",
      "v_min3_f32",
      "v_min3_i16",
      "v_min3_i32",
      "v_min3_u16",
      "v_min3_u32",
      "v_min_f16",
      "v_min_f32",
      "v_min_f64",
      "v_min_f64_e64",
      "v_min_i16",
      "v_min_i16_e64",
      "v_min_i32",
      "v_min_legacy_f32",
      "v_min_u16",
      "v_min_u16_e64",
      "v_min_u32",
      "v_minimum3_f16",
      "v_minimum3_f32",
      "v_minimum_f16",
      "v_minimum_f32",
      "v_minimum_f64",
      "v_minimummaximum_f16",
      "v_minimummaximum_f32",
      "v_minmax_f16",
      "v_minmax_f32",
      "v_minmax_i32",
      "v_minmax_u32",
      "v_mov_b16",
      "v_mov_b32",
      "v_movreld_b32",
      "v_movrels_b32",
      "v_movrelsd_2_b32",
      "v_movrelsd_b32",
      "v_mqsad_pk_u16_u8",
      "v_mqsad_u32_u8",
      "v_msad_u8",
      "v_mul_f16",
      "v_mul_f32",
      "v_mul_f64",
      "v_mul_f64_e64",
      "v_mul_hi_i32",
      "v_mul_hi_i32_i24",
      "v_mul_hi_u32",
      "v_mul_hi_u32_u24",
      "v_mul_i32_i24",
      "v_mul_legacy_f32",
      "v_mul_lo_i32",
      "v_mul_lo_u16",
      "v_mul_lo_u16_e64",
      "v_mul_lo_u32",
      "v_mul_u32_u24",
      "v_mullit_f32",
      "v_nop",
      "v_not_b16",
      "v_not_b32",
      "v_or3_b32",
      "v_or_b16",
      "v_or_b32",
      "v_pack_b32_f16",
      "v_perm_b32",
      "v_permlane16_b32",
      "v_permlane16_var_b32",
      "v_permlane64_b32",
      "v_permlanex16_b32",
      "v_permlanex16_var_b32",
      "v_pipeflush",
      "v_pk_add_f16",
      "v_pk_add_i16",
      "v_pk_add_u16",
      "v_pk_ashrrev_i16",
      "v_pk_fma_f16",
      "v_pk_fmac_f16",
      "v_pk_lshlrev_b16",
      "v_pk_lshrrev_b16",
      "v_pk_mad_i16",
      "v_pk_mad_u16",
      "v_pk_max_f16",
      "v_pk_max_i16",
      "v_pk_max_u16",
      "v_pk_maximum_f16",
      "v_pk_min_f16",
      "v_pk_min_i16",
      "v_pk_min_u16",
      "v_pk_minimum_f16",
      "v_pk_mul_f16",
      "v_pk_mul_lo_u16",
      "v_pk_sub_i16",
      "v_pk_sub_u16",
      "v_qsad_pk_u16_u8",
      "v_rcp_clamp_f32",
      "v_rcp_clamp_f64",
      "v_rcp_f16",
      "v_rcp_f32",
      "v_rcp_f64",
      "v_rcp_iflag_f32",
      "v_rcp_legacy_f32",
      "v_readfirstlane_b32",
      "v_readlane_b32",
      "v_readlane_b32_e64",
      "v_rndne_f16",
      "v_rndne_f32",
      "v_rndne_f64",
      "v_rsq_clamp_f32",
      "v_rsq_clamp_f64",
      "v_rsq_f16",
      "v_rsq_f32",
      "v_rsq_f64",
      "v_rsq_legacy_f32",
      "v_s_exp_f16",
      "v_s_exp_f32",
      "v_s_log_f16",
      "v_s_log_f32",
      "v_s_rcp_f16",
      "v_s_rcp_f32",
      "v_s_rsq_f16",
      "v_s_rsq_f32",
      "v_s_sqrt_f16",
      "v_s_sqrt_f32",
      "v_sad_hi_u8",
      "v_sad_u16",
      "v_sad_u32",
      "v_sad_u8",
      "v_sat_pk_u8_i16",
      "v_screen_partition_4se_b32",
      "v_sin_f16",
      "v_sin_f32",
      "v_sqrt_f16",
      "v_sqrt_f32",
      "v_sqrt_f64",
      "v_sub_co_u32",
      "v_sub_co_u32_e64",
      "v_sub_f16",
      "v_sub_f32",
      "v_sub_i16",
      "v_sub_i32",
      "v_sub_u16",
      "v_sub_u16_e64",
      "v_sub_u32",
      "v_subb_co_u32",
      "v_subbrev_co_u32",
      "v_subrev_co_u32",
      "v_subrev_co_u32_e64",
      "v_subrev_f16",
      "v_subrev_f32",
      "v_subrev_u16",
      "v_subrev_u32",
      "v_swap_b16",
      "v_swap_b32",
      "v_swaprel_b32",
      "v_swmmac_bf16_16x16x32_bf16",
      "v_swmmac_f16_16x16x32_f16",
      "v_swmmac_f32_16x16x32_bf16",
      "v_swmmac_f32_16x16x32_bf8_bf8",
      "v_swmmac_f32_16x16x32_bf8_fp8",
      "v_swmmac_f32_16x16x32_f16",
      "v_swmmac_f32_16x16x32_fp8_bf8",
      "v_swmmac_f32_16x16x32_fp8_fp8",
      "v_swmmac_i32_16x16x32_iu4",
      "v_swmmac_i32_16x16x32_iu8",
      "v_swmmac_i32_16x16x64_iu4",
      "v_trig_preop_f64",
      "v_trunc_f16",
      "v_trunc_f32",
      "v_trunc_f64",
      "v_wmma_bf16_16x16x16_bf16",
      "v_wmma_f16_16x16x16_f16",
      "v_wmma_f32_16x16x16_bf16",
      "v_wmma_f32_16x16x16_bf8_bf8",
      "v_wmma_f32_16x16x16_bf8_fp8",
      "v_wmma_f32_16x16x16_f16",
      "v_wmma_f32_16x16x16_fp8_bf8",
      "v_wmma_f32_16x16x16_fp8_fp8",
      "v_wmma_i32_16x16x16_iu4",
      "v_wmma_i32_16x16x16_iu8",
      "v_wmma_i32_16x16x32_iu4",
      "v_writelane_b32",
      "v_writelane_b32_e64",
      "v_xad_u32",
      "v_xnor_b32",
      "v_xor3_b32",
      "v_xor_b16",
      "v_xor_b32",
   },
   {
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::MUBUF,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::DS,
      aco::Format::EXP,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::FLAT,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::GLOBAL,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::MIMG,
      aco::Format::LDSDIR,
      aco::Format::LDSDIR,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO_BARRIER,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO_BRANCH,
      aco::Format::PSEUDO_CALL,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO_BRANCH,
      aco::Format::PSEUDO_BRANCH,
      aco::Format::PSEUDO_BRANCH,
      aco::Format::PSEUDO,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO_REDUCTION,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO_REDUCTION,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::SOP1,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO_REDUCTION,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::PSEUDO,
      aco::Format::SOP1,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::PSEUDO,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOPK,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SOPK,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOP2,
      aco::Format::SOPK,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPK,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPC,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPP,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SMEM,
      aco::Format::SOP1,
      aco::Format::SOPK,
      aco::Format::SMEM,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPK,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOPK,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPC,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPC,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SMEM,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPK,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPP,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPK,
      aco::Format::SOPP,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SOP2,
      aco::Format::SOP2,
      aco::Format::SOP1,
      aco::Format::SOP1,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::SCRATCH,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::MTBUF,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOPC,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP2,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP2,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOPD,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VINTRP,
      aco::Format::VINTERP_INREG,
      aco::Format::VINTERP_INREG,
      aco::Format::VINTERP_INREG,
      aco::Format::VINTRP,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VINTERP_INREG,
      aco::Format::VINTRP,
      aco::Format::VINTERP_INREG,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VINTERP_INREG,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP2,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP2,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP1,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP3P,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
      aco::Format::VOP3,
      aco::Format::VOP3,
      aco::Format::VOP2,
   },
   {
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::ds,
      instr_class::exp,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::ds,
      instr_class::ds,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::salu,
      instr_class::salu,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::salu,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::salu,
      instr_class::salu,
      instr_class::other,
      instr_class::sfpu,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::other,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::barrier,
      instr_class::barrier,
      instr_class::barrier,
      instr_class::barrier,
      instr_class::barrier,
      instr_class::barrier,
      instr_class::barrier,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::branch,
      instr_class::salu,
      instr_class::salu,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::salu,
      instr_class::waitcnt,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::barrier,
      instr_class::smem,
      instr_class::salu,
      instr_class::salu,
      instr_class::smem,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::smem,
      instr_class::smem,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::salu,
      instr_class::salu,
      instr_class::branch,
      instr_class::branch,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::sendmsg,
      instr_class::salu,
      instr_class::salu,
      instr_class::sendmsg,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::branch,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::waitcnt,
      instr_class::salu,
      instr_class::salu,
      instr_class::smem,
      instr_class::smem,
      instr_class::smem,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::branch,
      instr_class::branch,
      instr_class::branch,
      instr_class::other,
      instr_class::sfpu,
      instr_class::sfpu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::salu,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::waitcnt,
      instr_class::salu,
      instr_class::barrier,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::salu,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::vmem,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double_add,
      instr_class::valu_double_add,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double_convert,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double_convert,
      instr_class::valu_double_convert,
      instr_class::valu_double_convert,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double_convert,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double_convert,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu_fma,
      instr_class::valu_double,
      instr_class::valu_fma,
      instr_class::valu_fma,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu64,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double_add,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double_add,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_quarter_rate32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double_add,
      instr_class::valu_double,
      instr_class::valu_quarter_rate32,
      instr_class::valu32,
      instr_class::valu_quarter_rate32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_quarter_rate32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_quarter_rate32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_transcendental32,
      instr_class::valu_double_transcendental,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu_double_transcendental,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::valu_transcendental32,
      instr_class::valu_double_transcendental,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu_double_transcendental,
      instr_class::valu_transcendental32,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu_pseudo_scalar_trans,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu_transcendental32,
      instr_class::valu_double_transcendental,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::valu_double,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu_double,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::wmma,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
      instr_class::valu32,
   },
   {
      { // buffer_atomic_add
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_add_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_add_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_and
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_and_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_cmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_cmpswap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_csub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_dec
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_dec_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_fcmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_fcmpswap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_fmax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_fmax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_fmin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_fmin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_inc
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_inc_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_or
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_or_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_pk_add_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_pk_add_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_rsub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_rsub_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_smax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_smax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_smin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_smin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_sub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_sub_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_swap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_swap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_umax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_umax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_umin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_umin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_xor
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_atomic_xor_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_gl0_inv
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_gl1_inv
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_format_d16_hi_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_format_d16_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_format_d16_xy
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_format_d16_xyz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_format_d16_xyzw
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_format_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_format_xy
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_format_xyz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_format_xyzw
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_sbyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_sbyte_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_sbyte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_short_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_short_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_sshort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_ubyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_ubyte_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_ubyte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_load_ushort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_byte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_byte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_format_d16_hi_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_format_d16_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_format_d16_xy
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_format_d16_xyz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_format_d16_xyzw
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_format_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_format_xy
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_format_xyz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_format_xyzw
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_lds_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_short
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_store_short_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_wbinvl1
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // buffer_wbinvl1_vol
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_gs_reg_rtn
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_rtn_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_rtn_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_rtn_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_src2_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_src2_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_src2_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_add_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_and_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_and_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_and_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_and_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_and_src2_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_and_src2_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_append
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_bpermute_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_bvh_stack_push4_pop1_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_bvh_stack_push8_pop1_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_bvh_stack_push8_pop2_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cmpst_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cmpst_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cmpst_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cmpst_f64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cmpst_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cmpst_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cmpst_rtn_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cmpst_rtn_f64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cond_sub_rtn
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_cond_sub_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_condxchg32_rtn_b128
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_condxchg32_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_consume
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_dec_rtn_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_dec_rtn_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_dec_src2_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_dec_src2_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_dec_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_dec_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_gws_barrier
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_gws_init
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_gws_sema_br
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_gws_sema_p
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_gws_sema_release_all
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_gws_sema_v
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_inc_rtn_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_inc_rtn_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_inc_src2_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_inc_src2_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_inc_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_inc_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_f64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_i32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_i64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_rtn_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_rtn_f64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_rtn_i32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_rtn_i64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_rtn_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_rtn_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_src2_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_src2_f64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_src2_i32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_src2_i64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_src2_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_src2_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_max_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_f64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_i32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_i64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_rtn_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_rtn_f64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_rtn_i32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_rtn_i64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_rtn_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_rtn_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_src2_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_src2_f64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_src2_i32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_src2_i64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_src2_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_src2_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_min_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_mskor_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_mskor_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_mskor_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_mskor_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_nop
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_or_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_or_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_or_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_or_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_or_src2_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_or_src2_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_ordered_count
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_permute_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_pk_add_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_pk_add_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_pk_add_rtn_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_pk_add_rtn_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read2_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read2_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read2st64_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read2st64_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_addtid_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_b128
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_b96
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_i16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_i8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_i8_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_i8_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_u16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_u16_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_u16_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_u8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_u8_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_read_u8_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_rsub_rtn_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_rsub_rtn_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_rsub_src2_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_rsub_src2_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_rsub_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_rsub_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_sub_clamp_rtn_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_sub_clamp_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_sub_gs_reg_rtn
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_sub_rtn_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_sub_rtn_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_sub_src2_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_sub_src2_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_sub_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_sub_u64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_swizzle_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_wrap_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write2_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write2_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write2st64_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write2st64_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_addtid_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_b128
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_b16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_b16_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_b8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_b8_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_b96
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_src2_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_write_src2_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_wrxchg2_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_wrxchg2_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_wrxchg2st64_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_wrxchg2st64_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_wrxchg_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_wrxchg_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_xor_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_xor_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_xor_rtn_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_xor_rtn_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_xor_src2_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // ds_xor_src2_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // exp
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_add
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_add_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_add_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_and
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_and_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_cmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_cmpswap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_cond_sub_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_csub_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_dec
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_dec_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_fcmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_fcmpswap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_fmax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_fmax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_fmin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_fmin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_inc
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_inc_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_or
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_or_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_pk_add_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_pk_add_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_smax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_smax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_smin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_smin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_sub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_sub_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_swap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_swap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_umax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_umax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_umin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_umin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_xor
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_atomic_xor_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_sbyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_sbyte_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_sbyte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_short_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_short_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_sshort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_ubyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_ubyte_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_ubyte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_load_ushort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_store_byte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_store_byte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_store_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_store_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_store_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_store_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_store_short
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // flat_store_short_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_add
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_add_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_add_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_and
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_and_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_cmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_cmpswap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_cond_sub_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_csub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_dec
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_dec_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_fcmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_fcmpswap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_fmax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_fmax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_fmin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_fmin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_inc
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_inc_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_or
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_or_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_ordered_add_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_pk_add_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_pk_add_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_smax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_smax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_smin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_smin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_sub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_sub_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_swap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_swap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_umax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_umax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_umin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_umin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_xor
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_atomic_xor_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_inv
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_dword_addtid
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_sbyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_sbyte_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_sbyte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_short_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_short_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_sshort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_tr_b128
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_tr_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_ubyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_ubyte_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_ubyte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_load_ushort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_store_byte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_store_byte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_store_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_store_dword_addtid
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_store_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_store_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_store_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_store_short
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_store_short_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_wb
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // global_wbinv
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_add
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_add_flt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_and
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_cmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_dec
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_fcmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_fmax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_fmin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_inc
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_or
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_pk_add_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_pk_add_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_rsub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_smax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_smin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_sub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_swap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_umax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_umin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_atomic_xor
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_bvh64_intersect_ray
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_bvh8_intersect_ray
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_bvh_dual_intersect_ray
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_bvh_intersect_ray
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_b
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_b_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_b_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_b_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_b
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_b_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_b_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_b_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_l
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_l_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_lz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_lz_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_c_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_l
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_l_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_lz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_lz_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_gather4_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_get_lod
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_get_resinfo
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_load
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_load_mip
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_load_mip_pck
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_load_mip_pck_sgn
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_load_pck
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_load_pck_sgn
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_msaa_load
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_b
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_b_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_b_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_b_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_b
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_b_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_b_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_b_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_cd
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_cd_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_cd_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_cd_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_d
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_d_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_d_cl_g16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_d_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_d_cl_o_g16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_d_g16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_d_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_d_o_g16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_l
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_l_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_lz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_lz_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_c_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_cd
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_cd_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_cd_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_cd_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_d
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_d_cl
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_d_cl_g16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_d_cl_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_d_cl_o_g16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_d_g16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_d_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_d_o_g16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_l
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_l_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_lz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_lz_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_sample_o
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_store
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_store_mip
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_store_mip_pck
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // image_store_pck
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // lds_direct_load
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // lds_param_load
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_as_uniform
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_barrier
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_boolean_phi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_bpermute_permlane
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_bpermute_readlane
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_bpermute_shared_vgpr
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_branch
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_call
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_callee_stack_ptr
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_cbranch
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_cbranch_nz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_cbranch_z
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_constaddr
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_constaddr_addlo
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // p_constaddr_getpc
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // p_create_vector
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_debug_info
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_demote_to_helper
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_discard_if
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_dual_src_export_gfx11
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_elect
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_end_linear_vgpr
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_end_with_regs
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_end_wqm
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_exclusive_scan
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_exit_early_if_not
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_extract
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_extract_vector
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_inclusive_scan
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_init_scratch
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_insert
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_interp_gfx11
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_is_helper
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_jump_to_epilog
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_linear_phi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_load_symbol
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // p_logical_end
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_logical_start
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_parallelcopy
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_permlane64_shared_vgpr
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_phi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_pops_gfx9_add_exiting_wave_id
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_pops_gfx9_ordered_section_done
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_pops_gfx9_overlapped_wave_wait_done
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_reduce
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_reload
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_reload_preserved
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_resume_shader_address
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_resumeaddr_addlo
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // p_resumeaddr_getpc
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // p_return
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_s_cvt_f16_f32_rtne
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // p_shader_cycles_hi_lo_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_spill
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_split_vector
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_start_linear_vgpr
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_startpgm
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_unit_test
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // p_v_cvt_f16_f32_rtne
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // p_v_cvt_f16_f32_rtni
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // p_v_cvt_f16_f32_rtpi
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // p_v_cvt_pk_fp8_f32_ovfl
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               8,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // p_v_fma_mixhi_f16_rtz
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // p_v_fma_mixlo_f16_rtz
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_abs_i32
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_absdiff_i32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_add_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_add_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_add_i32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_add_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_add_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_addc_u32
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_scc,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_addk_i32
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_and_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_and_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_and_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_and_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_andn1_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_andn1_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_andn1_wrexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_andn1_wrexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_andn2_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_andn2_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_andn2_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_andn2_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_andn2_wrexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_andn2_wrexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_ashr_i32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_ashr_i64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_atc_probe
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atc_probe_buffer
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_add
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_add_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_and
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_and_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_cmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_cmpswap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_dec
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_dec_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_inc
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_inc_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_or
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_or_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_smax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_smax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_smin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_smin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_sub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_sub_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_swap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_swap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_umax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_umax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_umin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_umin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_xor
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_atomic_xor_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_barrier
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_barrier_init
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_barrier_join
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_barrier_leave
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_barrier_signal
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_barrier_signal_isfirst
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_barrier_wait
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_bcnt0_i32_b32
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_bcnt0_i32_b64
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_bcnt1_i32_b32
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_bcnt1_i32_b64
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_bfe_i32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_bfe_i64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_bfe_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_bfe_u64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_bfm_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_bfm_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_bitcmp0_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_bitcmp0_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_bitcmp1_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_bitcmp1_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_bitreplicate_b64_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_bitset0_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_bitset0_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_bitset1_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_bitset1_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_branch
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_brev_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_brev_b64
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_buffer_atomic_add
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_add_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_and
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_and_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_cmpswap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_cmpswap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_dec
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_dec_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_inc
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_inc_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_or
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_or_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_smax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_smax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_smin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_smin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_sub
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_sub_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_swap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_swap_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_umax
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_umax_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_umin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_umin_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_xor
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_atomic_xor_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_dwordx16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_dwordx8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_sbyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_sshort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_ubyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_load_ushort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_prefetch_data
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_store_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_store_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_buffer_store_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_call_b64
         0,
         1,
         0,
         0,
         {
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
         },
         {
               not_fixed,
         },
      },
      { // s_cbranch_cdbgsys
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_cdbgsys_and_user
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_cdbgsys_or_user
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_cdbguser
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_execnz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_execz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_g_fork
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_i_fork
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_join
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_scc0
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_scc1
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_vccnz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cbranch_vccz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_ceil_f16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_ceil_f32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_clause
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cmov_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_scc,
         },
         {
               not_fixed,
         },
      },
      { // s_cmov_b64
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_scc,
         },
         {
               not_fixed,
         },
      },
      { // s_cmovk_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_scc,
         },
         {
               not_fixed,
         },
      },
      { // s_cmp_eq_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_eq_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_eq_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_eq_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_eq_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_ge_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_ge_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_ge_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_ge_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_gt_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_gt_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_gt_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_gt_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_le_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_le_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_le_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_le_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_lg_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_lg_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_lg_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_lg_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_lg_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_lt_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_lt_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_lt_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_lt_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_neq_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_neq_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_nge_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_nge_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_ngt_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_ngt_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_nle_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_nle_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_nlg_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_nlg_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_nlt_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_nlt_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_o_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_o_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_u_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmp_u_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_eq_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_eq_u32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_ge_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_ge_u32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_gt_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_gt_u32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_le_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_le_u32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_lg_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_lg_u32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_lt_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_cmpk_lt_u32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               fixed_scc,
         },
      },
      { // s_code_end
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_cselect_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_scc,
         },
         {
               not_fixed,
         },
      },
      { // s_cselect_b64
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_scc,
         },
         {
               not_fixed,
         },
      },
      { // s_cvt_f16_f32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_cvt_f32_f16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_cvt_f32_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_cvt_f32_u32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_cvt_hi_f32_f16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_cvt_i32_f32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_cvt_pk_rtz_f16_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_cvt_u32_f32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_dcache_discard
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_dcache_discard_x2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_dcache_inv
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_dcache_inv_vol
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_dcache_wb
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_dcache_wb_vol
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_decperflevel
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_delay_alu
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_denorm_mode
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_endpgm
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_endpgm_ordered_ps_done
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_endpgm_saved
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_ff0_i32_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_ff0_i32_b64
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_ff1_i32_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_ff1_i32_b64
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_flbit_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_flbit_i32_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_flbit_i32_b64
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_flbit_i32_i64
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_floor_f16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_floor_f32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_fmaak_f32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // s_fmac_f16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_fmac_f32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_fmamk_f32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // s_get_barrier_state
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_get_waveid_in_workgroup
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_getpc_b64
         0,
         1,
         0,
         0,
         {
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
         },
         {
               not_fixed,
         },
      },
      { // s_getreg_b32
         0,
         1,
         0,
         0,
         {
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
      },
      { // s_gl1_inv
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_icache_inv
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_incperflevel
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_inst_prefetch
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_dwordx16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_dwordx8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_sbyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_sshort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_ubyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_load_ushort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_lshl1_add_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_lshl2_add_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_lshl3_add_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_lshl4_add_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_lshl_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_lshl_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_lshr_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_lshr_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_max_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_max_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_max_i32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_max_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_maximum_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_maximum_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_memrealtime
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_memtime
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_min_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_min_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_min_i32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_min_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_minimum_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_minimum_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_mov_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_mov_b64
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_mov_fed_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_movk_i32
         0,
         1,
         0,
         0,
         {
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
      },
      { // s_movreld_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // s_movreld_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // s_movrels_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // s_movrels_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // s_movrelsd_2_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // s_mul_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_mul_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_mul_hi_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_mul_hi_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_mul_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_mul_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_mulk_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_nand_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_nand_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_nand_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_nand_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_nop
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_nor_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_nor_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_nor_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_nor_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_not_b32
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_not_b64
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_or_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_or_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_or_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_or_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_orn1_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_orn1_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_orn2_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_orn2_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_orn2_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_orn2_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_pack_hh_b32_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_pack_hl_b32_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_pack_lh_b32_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_pack_ll_b32_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_prefetch_data
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_prefetch_data_pc_rel
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_prefetch_inst
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_prefetch_inst_pc_rel
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_quadmask_b32
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_quadmask_b64
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_rfe_b64
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_rfe_restore_b64
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_rndne_f16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_rndne_f32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_round_mode
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_scratch_load_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_scratch_load_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_scratch_load_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_scratch_store_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_scratch_store_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_scratch_store_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_sendmsg
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_sendmsg_rtn_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_sendmsg_rtn_b64
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_sendmsghalt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_set_gpr_idx_idx
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               fixed_m0,
         },
      },
      { // s_set_gpr_idx_mode
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               fixed_m0,
         },
         {
               fixed_m0,
         },
      },
      { // s_set_gpr_idx_off
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_set_gpr_idx_on
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_m0,
         },
      },
      { // s_sethalt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_setkill
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_setpc_b64
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_setprio
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_setreg_b32
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_setreg_imm32_b32
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
         },
         {
               fixed_imm,
         },
         {
         },
      },
      { // s_setvskip
         2,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
               not_fixed,
         },
         {
         },
      },
      { // s_sext_i32_i16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_sext_i32_i8
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_singleuse_vdst
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_sleep
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_sleep_var
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_store_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_store_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_store_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_sub_f16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_sub_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_sub_i32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_sub_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_sub_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_subb_u32
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_scc,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_subvector_loop_begin
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_subvector_loop_end
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_swappc_b64
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_trap
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_trunc_f16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_trunc_f32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // s_ttracedata
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               fixed_m0,
         },
         {
         },
      },
      { // s_ttracedata_imm
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_version
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_bvhcnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_dscnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_event
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_expcnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_idle
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_kmcnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_loadcnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_loadcnt_dscnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_samplecnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_storecnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wait_storecnt_dscnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_waitcnt
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_waitcnt_depctr
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_waitcnt_expcnt
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_waitcnt_lgkmcnt
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_waitcnt_vmcnt
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_waitcnt_vscnt
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_wakeup
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // s_wakeup_barrier
         1,
         0,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
         },
         {
               not_fixed,
         },
         {
         },
      },
      { // s_wqm_b32
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_wqm_b64
         1,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_xnor_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_xnor_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_xnor_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_xnor_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_xor_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_xor_b64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               fixed_scc,
         },
      },
      { // s_xor_saveexec_b32
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // s_xor_saveexec_b64
         2,
         3,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               fixed_exec,
         },
         {
               not_fixed,
               fixed_scc,
               fixed_exec,
         },
      },
      { // scratch_load_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_sbyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_sbyte_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_sbyte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_short_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_short_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_sshort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_ubyte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_ubyte_d16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_ubyte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_load_ushort
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_store_byte
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_store_byte_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_store_dword
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_store_dwordx2
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_store_dwordx3
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_store_dwordx4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_store_short
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // scratch_store_short_d16_hi
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_load_format_d16_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_load_format_d16_xy
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_load_format_d16_xyz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_load_format_d16_xyzw
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_load_format_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_load_format_xy
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_load_format_xyz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_load_format_xyzw
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_store_format_d16_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_store_format_d16_xy
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_store_format_d16_xyz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_store_format_d16_xyzw
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_store_format_x
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_store_format_xy
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_store_format_xyz
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // tbuffer_store_format_xyzw
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_add3_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_co_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_add_co_u32_e64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_add_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_f64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_f64_e64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_lshl_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_u16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_add_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_addc_co_u32
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_alignbit_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_alignbyte_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_and_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_and_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_and_or_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ashr_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ashr_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ashrrev_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ashrrev_i16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ashrrev_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ashrrev_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_bcnt_u32_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_bfe_i32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_bfe_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_bfi_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_bfm_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_bfrev_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ceil_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ceil_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ceil_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_clrexcp
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_cmp_class_f16
         2,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_class_f32
         2,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_class_f64
         2,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_eq_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_eq_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_eq_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_eq_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_eq_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_eq_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_eq_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_eq_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_eq_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_f_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_f_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_f_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_f_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_f_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_f_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_f_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_f_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_f_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ge_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ge_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ge_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ge_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ge_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ge_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ge_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ge_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ge_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_gt_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_gt_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_gt_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_gt_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_gt_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_gt_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_gt_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_gt_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_gt_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_le_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_le_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_le_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_le_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_le_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_le_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_le_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_le_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_le_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lg_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lg_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lg_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lg_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lg_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lg_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lg_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lg_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lg_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lt_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lt_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lt_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lt_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lt_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lt_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lt_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lt_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_lt_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_neq_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_neq_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_neq_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nge_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nge_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nge_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ngt_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ngt_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_ngt_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nle_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nle_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nle_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nlg_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nlg_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nlg_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nlt_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nlt_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_nlt_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_o_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_o_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_o_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_tru_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_tru_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_tru_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_tru_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_tru_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_tru_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_tru_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_tru_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_tru_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_u_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_u_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmp_u_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cmpx_class_f16
         2,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_class_f32
         2,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_class_f64
         2,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_eq_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_eq_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_eq_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_eq_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_eq_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_eq_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_eq_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_eq_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_eq_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_f_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_f_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_f_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_f_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_f_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_f_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_f_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_f_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_f_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ge_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ge_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ge_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ge_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ge_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ge_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ge_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ge_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ge_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_gt_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_gt_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_gt_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_gt_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_gt_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_gt_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_gt_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_gt_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_gt_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_le_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_le_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_le_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_le_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_le_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_le_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_le_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_le_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_le_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lg_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lg_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lg_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lg_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lg_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lg_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lg_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lg_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lg_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lt_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lt_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lt_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lt_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lt_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lt_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lt_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lt_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_lt_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_neq_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_neq_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_neq_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nge_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nge_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nge_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ngt_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ngt_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_ngt_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nle_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nle_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nle_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nlg_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nlg_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nlg_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nlt_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nlt_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_nlt_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_o_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_o_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_o_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_tru_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_tru_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_tru_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_tru_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_tru_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_tru_i64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_tru_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_tru_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_tru_u64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_u_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_u_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cmpx_u_f64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               fixed_exec,
         },
      },
      { // v_cndmask_b16
         3,
         1,
         3,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cndmask_b32
         3,
         1,
         3,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cos_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cos_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cubeid_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cubema_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cubesc_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cubetc_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f16_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f16_i16
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f16_u16
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_bf8
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_bfloat,
               1,
               8,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_fp8
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               8,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_i32
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_u32
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_ubyte0
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_ubyte1
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_ubyte2
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f32_ubyte3
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f64_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f64_i32
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_f64_u32
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_flr_i32_f32
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_i16_f16
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_i32_f32
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_i32_f64
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_i32_i16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_norm_i16_f16
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_norm_u16_f16
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_off_f32_i4
         1,
         1,
         0,
         1,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pk_bf8_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_bfloat,
               2,
               8,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pk_f32_bf8
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_bfloat,
               2,
               8,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pk_f32_fp8
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               2,
               8,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pk_fp8_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               8,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pk_i16_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pk_i16_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pk_u16_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pk_u16_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pk_u8_f32
         3,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pkaccum_u8_f32
         3,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pknorm_i16_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pknorm_i16_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pknorm_u16_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pknorm_u16_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pkrtz_f16_f32
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_pkrtz_f16_f32_e64
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_rpi_i32_f32
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_sr_bf8_f32
         2,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_bfloat,
               1,
               8,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_sr_fp8_f32
         2,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               8,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_u16_f16
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_u32_f32
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_u32_f64
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_cvt_u32_u16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_div_fixup_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_div_fixup_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_div_fixup_f64
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_div_fixup_legacy_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_div_fmas_f32
         4,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_div_fmas_f64
         4,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_div_scale_f32
         3,
         2,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_div_scale_f64
         3,
         2,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_dot2_bf16_bf16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_bfloat,
               2,
               16,
            },
            {
               aco_base_type_bfloat,
               2,
               16,
            },
            {
               aco_base_type_bfloat,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_bfloat,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot2_f16_f16
         3,
         1,
         7,
         0,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot2_f32_bf16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_bfloat,
               2,
               16,
            },
            {
               aco_base_type_bfloat,
               2,
               16,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot2_f32_f16
         3,
         1,
         7,
         0,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot2_i32_i16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot2_u32_u16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot2c_f32_f16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot4_f32_bf8_bf8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_bfloat,
               4,
               8,
            },
            {
               aco_base_type_bfloat,
               4,
               8,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot4_f32_bf8_fp8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_bfloat,
               4,
               8,
            },
            {
               aco_base_type_float,
               4,
               8,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot4_f32_fp8_bf8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               4,
               8,
            },
            {
               aco_base_type_bfloat,
               4,
               8,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot4_f32_fp8_fp8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               4,
               8,
            },
            {
               aco_base_type_float,
               4,
               8,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot4_i32_i8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot4_i32_iu8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot4_u32_u8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot4c_i32_i8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot8_i32_i4
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot8_i32_iu4
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dot8_u32_u4
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_dual_add_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_add_nc_u32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_and_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_cndmask_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_dot2acc_f32_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_dot2acc_f32_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_fmaak_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_fmac_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_fmamk_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_lshlrev_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_max_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_min_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_mov_b32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_mul_dx9_zero_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_mul_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_sub_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_dual_subrev_f32
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_exp_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_exp_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_exp_legacy_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ffbh_i32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ffbh_u32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ffbl_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_floor_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_floor_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_floor_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fma_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fma_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fma_f64
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fma_legacy_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fma_legacy_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fma_mix_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fma_mixhi_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fma_mixlo_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fmaak_f16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // v_fmaak_f32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // v_fmac_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fmac_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fmac_legacy_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fmamk_f16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // v_fmamk_f32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // v_fract_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fract_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_fract_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_frexp_exp_i16_f16
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_frexp_exp_i32_f32
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_frexp_exp_i32_f64
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_frexp_mant_f16
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_frexp_mant_f32
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_frexp_mant_f64
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_mov_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p10_f16_f32_inreg
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p10_f32_inreg
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p10_rtz_f16_f32_inreg
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p1_f32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p1ll_f16
         2,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p1lv_f16
         3,
         1,
         5,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p2_f16
         3,
         1,
         5,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               fixed_m0,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p2_f16_f32_inreg
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p2_f32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p2_f32_inreg
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p2_hi_f16
         3,
         1,
         5,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               fixed_m0,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p2_legacy_f16
         3,
         1,
         5,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               fixed_m0,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_interp_p2_rtz_f16_f32_inreg
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ldexp_f16
         2,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ldexp_f32
         2,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_ldexp_f64
         2,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lerp_u8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_log_clamp_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_log_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_log_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_log_legacy_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshl_add_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshl_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshl_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshl_or_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshlrev_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshlrev_b16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshlrev_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshlrev_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshlrev_b64_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshr_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshr_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshrrev_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshrrev_b16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshrrev_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_lshrrev_b64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mac_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mac_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mac_legacy_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_i16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_i32_i16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_i32_i24
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_i64_i32
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_int,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_int,
               1,
               64,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_mad_legacy_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_legacy_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_legacy_i16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_legacy_u16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_u16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_u32_u16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_u32_u24
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mad_u64_u32
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_madak_f16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // v_madak_f32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // v_madmk_f16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // v_madmk_f32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_none,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_imm,
         },
         {
               not_fixed,
         },
      },
      { // v_max3_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max3_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max3_i16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max3_i32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max3_u16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max3_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_f64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_f64_e64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_i16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_legacy_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_u16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_max_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maximum3_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maximum3_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maximum_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maximum_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maximum_f64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maximumminimum_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maximumminimum_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maxmin_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maxmin_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maxmin_i32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_maxmin_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mbcnt_hi_u32_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mbcnt_hi_u32_b32_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mbcnt_lo_u32_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_med3_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_med3_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_med3_i16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_med3_i32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_med3_u16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_med3_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min3_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min3_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min3_i16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min3_i32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min3_u16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min3_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_f64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_f64_e64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_i16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_legacy_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_u16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_min_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minimum3_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minimum3_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minimum_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minimum_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minimum_f64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minimummaximum_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minimummaximum_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minmax_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minmax_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minmax_i32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_minmax_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mov_b16
         1,
         1,
         1,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mov_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_movreld_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // v_movrels_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // v_movrelsd_2_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // v_movrelsd_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
         },
      },
      { // v_mqsad_pk_u16_u8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mqsad_u32_u8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               4,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               4,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_msad_u8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_f64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_f64_e64
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_hi_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_hi_i32_i24
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_hi_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_hi_u32_u24
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_i32_i24
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_legacy_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_lo_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_lo_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_lo_u16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_lo_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mul_u32_u24
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_mullit_f32
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_nop
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_not_b16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_not_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_or3_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_or_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_or_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pack_b32_f16
         2,
         1,
         3,
         0,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_perm_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_permlane16_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_permlane16_var_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_permlane64_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_permlanex16_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_permlanex16_var_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pipeflush
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_pk_add_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_add_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_add_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_ashrrev_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_fma_f16
         3,
         1,
         7,
         1,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_fmac_f16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_lshlrev_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_lshrrev_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_mad_i16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_mad_u16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_max_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_max_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_max_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_maximum_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_min_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_min_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_min_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_minimum_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_mul_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               2,
               16,
            },
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_mul_lo_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_sub_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_pk_sub_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               2,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_qsad_pk_u16_u8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rcp_clamp_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rcp_clamp_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rcp_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rcp_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rcp_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rcp_iflag_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rcp_legacy_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_readfirstlane_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_readlane_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_readlane_b32_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rndne_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rndne_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rndne_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rsq_clamp_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rsq_clamp_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rsq_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rsq_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rsq_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_rsq_legacy_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_exp_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_exp_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_log_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_log_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_rcp_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_rcp_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_rsq_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_rsq_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_sqrt_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_s_sqrt_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sad_hi_u8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sad_u16
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sad_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sad_u8
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sat_pk_u8_i16
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_screen_partition_4se_b32
         1,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sin_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sin_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sqrt_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sqrt_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sqrt_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sub_co_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_sub_co_u32_e64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_sub_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sub_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sub_i16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sub_i32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sub_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sub_u16_e64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_sub_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_subb_co_u32
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_subbrev_co_u32
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_subrev_co_u32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_subrev_co_u32_e64
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_lanemask,
               1,
               1,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_subrev_f16
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_subrev_f32
         2,
         1,
         3,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_subrev_u16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_subrev_u32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_swap_b16
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_swap_b32
         2,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_swaprel_b32
         3,
         2,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               fixed_m0,
         },
         {
               not_fixed,
               not_fixed,
         },
      },
      { // v_swmmac_bf16_16x16x32_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_f16_16x16x32_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_f32_16x16x32_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_f32_16x16x32_bf8_bf8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_f32_16x16x32_bf8_fp8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_f32_16x16x32_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_f32_16x16x32_fp8_bf8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_f32_16x16x32_fp8_fp8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_i32_16x16x32_iu4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_i32_16x16x32_iu8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_swmmac_i32_16x16x64_iu4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_trig_preop_f64
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_trunc_f16
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               16,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_trunc_f32
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               32,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_trunc_f64
         1,
         1,
         1,
         1,
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
            {
               aco_base_type_float,
               1,
               64,
            },
         },
         {
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_wmma_bf16_16x16x16_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_f16_16x16x16_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_f32_16x16x16_bf16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_f32_16x16x16_bf8_bf8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_f32_16x16x16_bf8_fp8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_f32_16x16x16_f16
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_f32_16x16x16_fp8_bf8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_f32_16x16x16_fp8_fp8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_i32_16x16x16_iu4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_i32_16x16x16_iu8
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_wmma_i32_16x16x32_iu4
         0,
         0,
         0,
         0,
         {
         },
         {
         },
         {
         },
         {
         },
      },
      { // v_writelane_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_writelane_b32_e64
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_xad_u32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_xnor_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_xor3_b32
         3,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_xor_b16
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               16,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
      { // v_xor_b32
         2,
         1,
         0,
         0,
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
            {
               aco_base_type_uint,
               1,
               32,
            },
         },
         {
               not_fixed,
               not_fixed,
         },
         {
               not_fixed,
         },
      },
   },
};

}

